//
// DIVERSE graphics interface -- Performer ---- version 2.0.5
// released: Sat May 18 10:54:26 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//
#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pr/pfTexture.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>

const char* mapTexture = "2Dmap.rgb";
const float scale = 22000.0;
#define MAXPOINTS 10000

//  Key to toggle map
#define TOGGLE_KEY XK_E

//  Key to select a point
#define SELECT_KEY XK_S

// example generic dtkAugment DSO

class SelectPoints : public dtkAugment {
public:

  SelectPoints(dpf *app) ;
  ~SelectPoints(void) ;
  int preConfig(void);
  int postConfig(void);
  int preFrame(void);
  int postFrame(void);

private:

  dtkColor color;
  int count;
  dpf* App;
  pfDCS* mapDCS;
  pfDCS* markerDCS;
  pfDCS* youAreHereDCS;
  //  Shared mem for intersection point
  dtkSharedMem* shm;

  //  Shared mem to write the two selected points
  dtkSharedMem* selectShm;

  //  File to read the path from
  char* file;

  //  Show the mini map
  void addMiniMap();

  //  Construct a geode containing to hold the path
  void makeGeode (pfGroup* group, pfVec3*& vertices, pfGeode*& geode);

  //  Check to see if the plot file has been modified.  If it has been, read
  //  the new path from the file
  void checkModified();

  //  Read in a path from the playback file and populate vertices
  void readPath (pfVec3* vertices, float scale, int constantZ);

  //  True if the map is in front of the user, false if it is to the right
  //  Used to switch the map
  int inFront;

  //  Return true if the given point is on the mini-map
  int goodPoint(float* point);

  //  Check the keyboard for key presses
  void checkKey();

  //  Set the mini map to be in front of or to the right of the user
  void setFront();
  void setRight();

  //  Add a line to show the user's selection
  void addSelectLine();

  //  Points that the user has selected (in the coordinate system of the
  //  mini map)
  pfVec3* selectPoints;

  //  Number of points selected so far
  int numSelected;

  //  Markers for the selected points
  pfDCS* selectDCS[2];

  //  Vertices of the path on the large map
  pfVec3* largeVertices;

  //  Geode for the path plot on the large map
  pfGeode* largeGeode;

  //  Vertices of the path on the mini map
  pfVec3* miniVertices;

  //  Geode for the path plot on the mini map
  pfGeode* miniGeode;

  //  Number of vertices in the path (same for large and mini maps)
  int* numVertices;

  //  Key code to toggle map
  KeyCode toggle_key;

  //  Key code to select points
  KeyCode select_key;

  //  Keyboard to read from
  dtkInXKeyboard *keyboard;

  //  Dequeuer for the keyboard
  dtkDequeuer *dq;

  //  Stat record for the plot file
  struct stat fileStat;

  //  Modified time of the plot file when it was last read in
  time_t lastModified;
} ;

// create object and set description.
// be sure to validate it!
SelectPoints::SelectPoints(dpf *app) :
  dtkAugment("selectPoints"), color(stdout)
{
  setDescription("generic dtkAugment DSO") ;
  App = app;
  count = 0;
  
  //  Get the file name
  file = getenv("DPF_PLAYBACK_FILE");
  
  largeVertices = NULL;
  miniVertices = NULL;
  largeGeode = NULL;
  miniGeode = NULL;
  numVertices = NULL;
  keyboard = NULL;
  dq = NULL;
  selectPoints = NULL;

  lastModified = 0;
  numSelected = 0;

  validate();
} ;

// destructor
SelectPoints::~SelectPoints(void) {
}

// called only once by dpf, after pfInit() and dpfDisplay:preConfig()
// but before pfConfig()
int SelectPoints::preConfig(void) {
  return DTK_REMOVE_CALLBACK ;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int SelectPoints::postConfig(void) {

    //  Get the keyboard
  keyboard = (dtkInXKeyboard *) App->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard) 
    return DTK_CALLBACK_ERROR;
  keyboard->queue();

  //  Get the X key codes for toggling
  if(!(toggle_key = keyboard->getKeyCode(TOGGLE_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                              " to get x key code for TOGGLE_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }

  
  //  Get the X key codes for selecting
  if(!(select_key = keyboard->getKeyCode(SELECT_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                              " to get x key code for SELECT_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }

  //  Start the dequeuer for the keyboard
  dq = new dtkDequeuer(App->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;

  //  Get the shared memory for the point that the user has selected
  shm = new dtkSharedMem (3 * sizeof(float), "localPoint");
  if (!shm || shm->isInvalid() || shm->getSize() < (3*sizeof(float))) {
    printf ("could not get shared mem\n");
  }

  //  Get the shared memory to write the two points the user selects
  selectShm = new dtkSharedMem (4 * sizeof(float), "select");
  if (!selectShm || selectShm->isInvalid() || selectShm->getSize() < (4*sizeof(float))) {
    printf ("could not get shared mem\n");
  }

  pfFilePath(".:./ftirwin_roads:"DPF_DATA_DIR);

  //  Load a marker to show our position on the full map
  pfNode* marker = pfdLoadFile("red_arrow.pfb");
  App->display()->addGhostNode(marker);
  markerDCS = new pfDCS();
  markerDCS->addChild(marker);
  markerDCS->setRot(0, 180, 0);
  markerDCS->setTrans(1000, 1000, 1500);
  markerDCS->setScale(1000);
  App->display()->world()->addChild(markerDCS);

  //  Add the mini map, starts out in front
  addMiniMap();

  //  Plot the path on the large map and mini map
  makeGeode(mapDCS, miniVertices, miniGeode);
  makeGeode(App->display()->world(), largeVertices, largeGeode);

  checkModified();

  //  Load a smaller marker ("You are here") to show the navigation position
  //  on the mini map
  pfNode* sphere = pfdLoadFile("fov.pfb");
  App->display()->addGhostNode(sphere);
  youAreHereDCS = new pfDCS();
  youAreHereDCS->setScale(0.003);
  youAreHereDCS->addChild(sphere);
  youAreHereDCS->setRot(-45, 90, 0);
  mapDCS->addChild(youAreHereDCS);

  addSelectLine();

  //  Increase the navigation response to a good level
  App->display()->currentNav->response(2048);

  return DTK_CONTINUE ; 
}

//  Create a line segment to show where the user is selecting
void SelectPoints::addSelectLine() {

  void* arena = pfGetSharedArena();

  //  2 vertices for the selected points
  selectPoints = (pfVec3 *) pfMalloc (2 * sizeof(pfVec3), arena);

  //  Initialize the vertices to be behind the map
  selectPoints[0].set(0.5, 0.5, -1);
  selectPoints[1].set(0.5, 0.5, -1);

  //  Length of the vertex list
  int* len = static_cast<int*> (pfMalloc(sizeof(int), arena));
  *len = 2;

  //  Create the geo set
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType(PFGS_LINESTRIPS);
  gSet->setNumPrims(1);  // one line segment
  gSet->setPrimLengths(len);
  gSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, selectPoints, NULL);

  //  Disable lighting
  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gSet->setGState(gState);

  //  Set the color to blue
  pfVec4* color = static_cast<pfVec4*> (pfMalloc(sizeof(pfVec4), arena));
  color->set(0.3, 0.3, 1, 0);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  pfGeode* geode = new pfGeode();
  geode->addGSet(gSet);

  mapDCS->addChild(geode);
}

//  Add a mini-map to the scenegraph
void SelectPoints::addMiniMap() {

  //  Load the 2D map of Ft. Irwin in front of the user
  void* arena = pfGetSharedArena();

  //  Create a geoset with one polygon
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType (PFGS_QUADS);
  gSet->setNumPrims(1);

  pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

  //  Create a polygon 1 unit wide and 1.25 units high
  float w = 1;
  float h = 1.25;
  verts[0].set(0, 0, 0);
  verts[1].set(w, 0, 0);
  verts[2].set(w, h, 0);
  verts[3].set(0, h, 0);
  gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

  //  Set the texture to cover the entire polygon
  pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
  textcoords[0].set (0,0);
  textcoords[1].set (1, 0);
  textcoords[2].set (1, 1);
  textcoords[3].set (0, 1);

  gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

  pfGeode* geode = new pfGeode();
  geode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);

  pfTexture* tex = new pfTexture();

  int ret = tex->loadFile(mapTexture);
  if (ret == 0) {
    printf ("could not load file\n");
    return;
  }

  gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gState->setAttr(PFSTATE_TEXTURE, tex);
  
  gSet->setGState(gState);
  
  //  Set it to be in front of the user, in scene coordinates
  mapDCS = new pfDCS();
  setFront();
  inFront = 1;
  mapDCS->addChild(geode);

  //  Create markers for the selected points (blue dots)
  for (int c=0; c<2; c++) {
    selectDCS[c] = new pfDCS();

    pfNode* node = pfdLoadFile("blueSphere.pfb");
    App->display()->addGhostNode(node);
    if (node)
      selectDCS[c]->addChild(node);
    else
      break;

    selectDCS[c]->setScale(0.03);

    //  Put the markers behind the map so they are not seen
    selectDCS[c]->setTrans(0.5, 0.5, -1);
    
    mapDCS->addChild(selectDCS[c]);
  }


  App->display()->scene()->addChild(mapDCS);
  
  /*
  //  Create a sphere
  gSet = pfdNewSphere(200, arena);
  geode = new pfGeode();

  pfVec4* color = (pfVec4*) pfMalloc(sizeof(pfVec4), arena);
  color->set(0, 0, 1, 1);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  geode->addGSet(gSet);
  //gSet->setGState(gState);
  pfdStoreFile(geode, "blueSphere.pfb");
  */
}

//  Return true if the given point is on the mini-map
int SelectPoints::goodPoint(float* point) {

    //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
    //  Otherwise discard it
    return (fabs(point[2]) < 0.001 && point[0] >= 0 && point[0] <= 1.0 &&
        point[1] >= 0 && point[1] <= 1.25);
}

//  Check to see if the 'e' key was pressed.  If so, toggle the mini-map
//  between being in front of the user and to the right of the user
void SelectPoints::checkKey() {
  
  //  Check shared memory to get the selection point
  float point[3];
  shm->read(point);

  //  Update the selected line
  if (numSelected == 1) {

    if (goodPoint(point)) {
      selectPoints[1].set(point[0], point[1], 0.02);
    }
    else {
      selectPoints[1].set(0.5, 0.5, -1);
    }
  }

  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    if (pressed && keyCode == toggle_key) {

      if (inFront) {
        inFront = 0;
        setRight();
      }
      else {
        inFront = 1;
        setFront();
      }

    }

    if (pressed && keyCode == select_key) {

      //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
      //  Otherwise discard it
      if (goodPoint(point)) {
       
        //  If we have already selected two points, discard them
        if (numSelected == 2) {
          numSelected = 0;
          selectDCS[0]->setTrans(0.5, 0.5, -1);
          selectDCS[1]->setTrans(0.5, 0.5, -1);
          selectPoints[0].set(0.5, 0.5, -1);
          selectPoints[1].set(0.5, 0.5, -1);
        }

        selectDCS[numSelected]->setTrans(point[0], point[1], 0);
        selectPoints[numSelected].set(point[0], point[1], 0.02);
        numSelected++;

        if (numSelected == 2) {
          //  After a valid selection, write them to shared memory
          float select[4];
          select[0] = selectPoints[0][0] * scale;
          select[1] = selectPoints[0][1] * scale;
          select[2] = selectPoints[1][0] * scale;
          select[3] = selectPoints[1][1] * scale;
          selectShm->write(select, 4*sizeof(float));
        }
      }

    }
    
  }
}

//  Set the map to be in front of the user
void SelectPoints::setFront() {
  mapDCS->setRot(0, 90, 0);
  mapDCS->setTrans(-1.2, 2.5, 0);
}

//  Set the map to be to the right of the user (right cave wall)
void SelectPoints::setRight() {
  mapDCS->setRot(-90, 90, 0);
  mapDCS->setTrans(1.5, .5, -.5);
}

//  Create a geode for the path plot.  This creates a segment list with
//  zero vertices and adds it to group.
void SelectPoints::makeGeode (pfGroup* group, pfVec3*& vertices, pfGeode*& geode) {
  void* arena = pfGetSharedArena();

  //  Create the vertex list
  vertices = static_cast<pfVec3*> (pfMalloc(MAXPOINTS * sizeof(pfVec3), arena));

  //  Length of the vertex list
  if (numVertices == NULL)
    numVertices = static_cast<int*> (pfMalloc(sizeof(int), arena));
  *numVertices = 0;

  //  Create the geo set
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType(PFGS_LINESTRIPS);
  gSet->setNumPrims(1);
  gSet->setPrimLengths(numVertices);
  gSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, vertices, NULL);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gSet->setGState(gState);

  //  Set the color to green
  pfVec4* color = static_cast<pfVec4*> (pfMalloc(sizeof(pfVec4), arena));
  color->set(0.3, 1, 0.3, 0);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  geode = new pfGeode();
  geode->addGSet(gSet);

  group->addChild(geode);
}

//  Check to see if the plot file has been modified.  If it has been, read
//  the new path from the file
void SelectPoints::checkModified() {
  if (file == NULL) {
    printf ("Environment variable DPF_PLAYBACK_FILE is not set\n");
    return;
  }

  stat(file, &fileStat);

  if (fileStat.st_mtime > lastModified) {
    lastModified = fileStat.st_mtime;

    readPath(miniVertices, 1/scale, 1);
    readPath(largeVertices, 1, 0);

    //  Update the bounding volumes of the geodes and geosets
    miniGeode->getGSet(0)->setBound(NULL, PFBOUND_DYNAMIC);
    miniGeode->setBound(NULL, PFBOUND_DYNAMIC);
    largeGeode->getGSet(0)->setBound(NULL, PFBOUND_DYNAMIC);
    largeGeode->setBound(NULL, PFBOUND_DYNAMIC);
  }

}

//  Read a path from the playback file and populate vertices.  If constantZ is
//  non-zero, use a constant Z value instead of the Z values from the file
void SelectPoints::readPath (pfVec3* vertices, float scale, int constantZ) {

  //  Read in the map points
  int numPoints = 0;

  if (file == NULL) {
    printf ("Environment variable DPF_PLAYBACK_FILE is not set\n");
    return;
  }

  FILE* in = fopen(file, "r");
  if (in == NULL) {
    printf ("could not load file %s\n", file);
    return;
  }
  float t,x,y,z,h,p,r;
  while (fscanf (in, "%f %f %f %f %f %f %f", &t, &x, &y, &z, &h, &p, &r) >= 7) {
    vertices[numPoints][0] = x*scale;
    vertices[numPoints][1] = y*scale;

    if (constantZ)
      vertices[numPoints][2] = 0.02;
    else
      vertices[numPoints][2] = z*scale;

    numPoints++;

    if (numPoints >= MAXPOINTS) {
      printf ("Exceeded %i points in the playback file, ignoring the rest\n", numPoints);
      break;
    }
  }

  *numVertices = numPoints;
}

// called every display loop, as a pfFrame callback, before pfDraw()
// this return value causes it to only be called once
int SelectPoints::preFrame(void) {
  //  Check shared memory to get the selection point
  float point[3];
  shm->read(point);

  //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
  //  Otherwise discard it
  if (goodPoint(point)) {
    markerDCS->setTrans(point[0]*scale, point[1]*scale, 1500);
  }

  //  Plot the current navigation point on the mini-map
  
  //  Convert 0,0,0 from scene to world units
  pfCoord coord;
  coord.xyz.set(0,0,0);
  coord.hpr.set(0,0,0);
  App->display()->dpfToWorld(&coord);

  youAreHereDCS->setRot(-45+coord.hpr[0], 90, 0);
  youAreHereDCS->setTrans(coord.xyz[0]/scale, coord.xyz[1]/scale, 0.02);

  checkKey();

  checkModified();

  return DTK_CONTINUE ; 
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int SelectPoints::postFrame(void) {
  return DTK_CONTINUE ; 
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new SelectPoints(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
