#include <dpf.h>
#include <Performer/pr/pfTexture.h>
#include <dtk/dtkDSO_loader.h>
#include <vector>
#include <iostream>
#include <cmath>

using namespace std;

#define GRIDX 3 
#define GRIDY 3
#define RADIUS 2000
#define HEIGHT 2000 
#define SPEED 100;

class choppers : public dtkAugment
{
  public:
    choppers(dpf *app);
    
    int postConfig(void);
    int postFrame(void);

  private:

    dpf* app;
    dtkTime Timer;

    pfNode* chopperModel;
    pfNode* bladesModel;

    vector<pfDCS*> helies;
    double speed;
    double circumference;
    double loopTime;
    double degreesPerSec;
    long double lastTime;

    pfMatrix position;
    pfCoord coord;
    pfDCS* rotDCS;
    pfDCS* Blades;
};

// Constructor

choppers::choppers(dpf *app_in) : dtkAugment("choppers")
{
  setDescription("make a bunch of choppers fly around in circles");
  app = app_in;
  speed = SPEED;
  lastTime =0;
  circumference = 2*M_PI*RADIUS;
  loopTime = circumference/speed;
  degreesPerSec = 360/loopTime;
  cout<<"desgrees per sec = "<<degreesPerSec<<endl;

  validate();
}

int choppers::postConfig(void)
{
  pfFilePath(".:./models:../models:"DPF_DATA_DIR);

  chopperModel = pfdLoadFile("oh58d-kiowa-rotary_des.pfb");
  app->display()->addGhostNode(chopperModel);
//  bladesModel = pfdLoadFile("blades.pfb");
//  app->display()->addGhostNode(bladesModel);
  
  int X = 22000;
  int Y = 27500-2000;

  double distApartX = X/GRIDX;
  double distApartY = Y/GRIDY;
  double distX = distApartX/2;
  double distY = distApartY/2;
/*
  Blades = new pfDCS();
  Blades->addChild(bladesModel);
  Blades->setRot(0,0,0);
  Blades->setTrans(0,60,30);
*/

  rotDCS = new pfDCS();
  rotDCS->setRot(0,0,0);
  rotDCS->setTrans(0,0,0);
  pfDCS* newChopper = new pfDCS();
  newChopper->setRot(0,0,-10);
  newChopper->setTrans(RADIUS,0,0);
  newChopper->addChild(chopperModel);
//  newChopper->addChild(Blades);
  newChopper->setScale(.4);
  rotDCS->addChild(newChopper);

  for(int x=0;x<GRIDX;x++){
    for(int y=0;y<GRIDY;y++){
      pfDCS* transDCS = new pfDCS();
      transDCS->setRot(0,0,0);
      transDCS->setTrans(distX,distY,HEIGHT);
      transDCS->addChild(rotDCS);
      app->display()->world()->addChild(transDCS);

      distY += distApartY;
    }
    distY = distApartY;
    distX += distApartX;
  }

  return 0;
}

int choppers::postFrame(void)
{
  //postFrame pretty much just moves the helies around in a circle.
  //x = r * cos( (time * 2pi)/looptime )
  //y = r * sin( (time * 2pi)/looptime )
  //z will remain constant

  long double currTime = Timer.get();
  long double diffTime = currTime - lastTime;
  float degreeChange = (float) (degreesPerSec * diffTime);
  
  rotDCS->getMat(position);
  position.getOrthoCoord(&coord);
  rotDCS->setRot(coord.hpr[0]+degreeChange,0,0);
/*
  Blades->getMat(position);
  position.getOrthoCoord(&coord);
  Blades->setRot(coord.hpr[0]+degreeChange*100,0,0);
*/  
  lastTime = currTime;

  return 0;
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new choppers (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
