#include <stdio.h>
#include <dpf.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <string>
using namespace std;
#include "Messenger.h"
#include "dclient.h"

// example generic dtkAugment DSO modified to be a messeage sender to the 
// collaboration console
// Notice how simple it is to take an example program and modify it

class Generic : public dtkAugment {
public:

  Generic(dpf *app) ;
  ~Generic(void) ;
  //During the post frame step handle the navigation
  int postFrame(void);

private:

  //The application to play with
  dpf* App;
  //The message sender that sends the data to the fltk gui
  Messenger* Message;
  int NumUsers;
  dclient* Client;
  int Send;
  dtkTime* Time;
  double ElaspedTime;
};

// create object and set description.
// be sure to validate it!
Generic::Generic(dpf *app) : dtkAugment("CollaborationDSO")
{
	ElaspedTime = 0;
	Send = 0;
	NumUsers = 0;
	validate();
	App = app;
	Message = new Messenger("Collab-Commands");
	Message->setFrom("Server");
	Message->setTo("CollabGui");
	Message->setMessage("Update-Users");
	Time = new dtkTime;
	Time->get();
} 

// destructor
Generic::~Generic(void) 
{
 	delete Message;
}



//Every hundred frames output the user list to the collaborative gui
int Generic::postFrame(void) 
{
	ElaspedTime = ElaspedTime + Time->get();
	if (ElaspedTime > 5000)
	{
		ElaspedTime = 0;
		Client =  (dclient*) App->check("dclient");
		if (!Client)
			printf("Could not get the client pointer\n");
		 vector<duser*> NewUsers = Client->getUsers();
		 Message->addInt(NewUsers.size());
		 for (unsigned int i=0; i < NewUsers.size(); i++)
		 {
			 string UserName = NewUsers[i]->getName();
			 Message->addString(UserName);
		 }
		 Message->SendMessage();
	}
	return DTK_CONTINUE;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new Generic(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
