// This is the client to connect to the dedicated server

#include "dclient.h"
#include <dtk/dtkDSO_loader.h>

dclient::dclient(dpf *app_in) :
  dtkAugment("dclient")
{
	//Features of the old collaborative console
	Tethered = false;
	SharedView = false;
	SharedPerson = "";
  cout<<"dclient starting"<<endl;
  setDescription("connects to the server");
  app = app_in;
  mapSCS = NULL;
  Time = new dtkTime();
  validate();
}

dclient::~dclient(void)
{
  messenger->setTo("server");
  messenger->setMessage("disconnect");
  messenger->SendMessage();

  for(unsigned int i=0;i<users.size();i++){
    delete users[i];
  }

  if(positionShm != NULL)
    delete positionShm;

  if(parser) delete parser;

  if(messenger) delete messenger;
}

int dclient::preConfig(void)
{
  //Load the radar

  dtkAugment* radar = app->check("radar");
  if(radar == NULL){
    printf ("loading the radar\n");
    if(app->load("radar") == -1){
        dtkMsg.add(DTKMSG_ERROR,
          "dclient preConfig unable to load radar %s",
          __FILE__);
    }
  }       
  
  return DTK_CONTINUE;
}

int dclient::postConfig(void)
{
  cout<<"in postConfig"<<endl;
  
  char buff[64];
  unsigned int size = 64;
  gethostname(buff,size);
  compName = buff;
  ClientManager = new dtkClient;
  parser = new MessageParser("genMessage");
  messenger = new Messenger("genMessage");

  //DPF_SHARED_SERVER is the env varible.

  host = getenv("DPF_SHARED_SERVER");
  if(host == NULL){
    cout<<"no host name specified"<<endl;
    return -1;
  }

  if(messenger->connectRemoteSharedMem( host) == -1){
    cout<<compName.c_str()<<" could not connect to "<<&host<<endl;
  }

  string avatarname = getenv("DPF_AVATAR_NAME");
  messenger->setFrom(compName);
  messenger->setTo("server");
  messenger->setMessage("addme");
  messenger->addString(avatarname);

  messenger->SendMessage();

  usleep(300000); //Sleep for 1/3 of a second while server creates shm for it.

//  lastFrameTime = Timer.delta();
//  lastPingTime = -20;

  string tempname = "user-" + compName;

  positionShm = new dtkSharedMem (6*sizeof(float),tempname.c_str());
  if(!positionShm || positionShm->isInvalid() || positionShm->getSize() < 6*sizeof(float))
  {
    cout<<"Could not get shared mem for "<<tempname.c_str()<<endl;
  }

  return DTK_CONTINUE;
}

int dclient::postFrame(void)
{
/*  long double currTime = Timer.get();

  if(currTime - lastPingTime > 10){ //Check in w/the server every 10 sec.
    lastPingTime = currTime;
    messenger->setTo("server");
    messenger->setMessage("checkin");
    messenger->SendMessage();
  }
*/
//  if(currTime - lastFrameTime > .05 ){  //update every 1/20 seconds.
//    lastFrameTime = currTime;
    //  Convert 0,0,0 from scene to world units
    if (Highlight)
    {
	    ElapsedTime = ElapsedTime  + Time->get();
	    if (ElapsedTime > 2000)
	    {
		    ElapsedTime = 0;
		    Highlight = false;
		    int pos = SearchUsers(PersonToHighlight);
		    if (pos == -1)
			    return DTK_CONTINUE;
		    users[pos]->Highlight(false, app);
	    }
    }



	
    if (!Tethered && !SharedView)
    {
	    coord.xyz.set(0,0,0);
	    coord.hpr.set(0,0,0);
	    app->display()->worldToDpf(&coord);
	    //Access the position with coord.xyz and coord.hpr.
	    positionShm->write(&coord,6*sizeof(float));
    }
    if(Tethered && !SharedView)
    {
	    // Array for getting other users position
	    float Data[6];
	    for (int i=0; i<5; i++)
		    Data[i] = 0;
	    //Find out which user they are
	    int pos = SearchUsers(TetheredTo);
	    if (pos == -1)
	    {
		    Tethered = false;
		    printf("Could not find user %s\n", TetheredTo.c_str());
		    return DTK_CONTINUE;
	    }
	    const double Distance=80.0;
	    //Get their position into a temp array
	     users[pos]->getPosition(Data);
	     //Put the position into a pfCoord (only way for it 2 work)
	     pfCoord MoreTemp;
	     //Get a pfCoord of where I am currently at
	    pfCoord MyPos;
	    //My position
	    dtkNavList* NavList = (dtkNavList*)app->check("dtkNavList", DTKNAVLIST_TYPE);
	    MyPos.xyz[0]  = NavList->current()->location.x;
	    MyPos.xyz[1]  = NavList->current()->location.y;
	    MyPos.xyz[2]  = NavList->current()->location.z;
	    MyPos.hpr[0]  = NavList->current()->location.h;
	    MyPos.hpr[1]  = NavList->current()->location.p;
	    MyPos.hpr[2]  = NavList->current()->location.r;
	    //Position of the avatar I'm tethered to
	    MoreTemp.xyz[0] = Data[0] ;
	    MoreTemp.xyz[1] = Data[1] ;
	    MoreTemp.xyz[2] = Data[2] ;
	    MoreTemp.hpr[0] = Data[3] ;
	    MoreTemp.hpr[1] = Data[4] ;
	    MoreTemp.hpr[2] = Data[5] ;
	    //Are any of these distances negative
	    bool Neg[3]; //xyz array for negatives
	    if (MoreTemp.xyz[0] < 0)
		    Neg[0] = true;
	    if (MoreTemp.xyz[1] < 0)
		    Neg[1] = true;
	    if (MoreTemp.xyz[2] < 0)
		    Neg[2] = true;
	    double Dist[3]; //xyz array for the distances


	    //Computer the distance between the two points
	    Dist[0] = MoreTemp.xyz[0] - MyPos.xyz[0];
	    Dist[1] = MoreTemp.xyz[1] - MyPos.xyz[1];
	    Dist[2] = MoreTemp.xyz[2] - MyPos.xyz[2];

	    //Abs it since abs doesn't like doubles
	    if (Dist[0] < 0)
		    Dist[0] = Dist[0] * -1;
	    if (Dist[1] < 0)
		    Dist[1] = Dist[1] * -1;
	    if (Dist[2] < 0)
		    Dist[2] = Dist[2] * -1;
	    //Go through each point (x,y,z) and see if we are within the boundary
	    //of the sphere, if we are not then go to the boundary of the sphere
	    for (int i=0; i<3; i++)
	    {
		    if (Dist[i] > Distance)
		    {
			    printf("Different in %d\n", i);
		    //They are negative
			    if (Neg[i])
			    {
				    //What side of the sphere am I on and where should I be?
				    if (MyPos.xyz[i] < MoreTemp.xyz[i])
				    MyPos.xyz[i] = -1*Distance + MoreTemp.xyz[i];
				    else
				    MyPos.xyz[i] = Distance +  MoreTemp.xyz[i];
			    }
			    else
			    {
				    //What side of the sphere am I on and where should I be?
				    if (MyPos.xyz[i] < MoreTemp.xyz[i])
				    MyPos.xyz[i] = MoreTemp.xyz[i] - Distance;
				    else
				    MyPos.xyz[i] = Distance +  MoreTemp.xyz[i];
			    }
			   
		    }
	    }
	    app->display()->worldToDpf(MyPos);
	    //Set where I need to be in the world
	    for (int i=0; i<3; i++)
	    {
		NavList->current()->location.d[i] = MyPos.xyz[i];
		NavList->current()->location.d[i+3] = MyPos.hpr[i];
	    }
	    //app->display()->currentNav->location=MyPos;
	    //Write my location out so everyone can tell where I am at
	    positionShm->write(&MyPos,6*sizeof(float));
    }
    if(SharedView && !Tethered)
    {
	    // Array for getting other users position
	    float Data[6];
	    for (int i=0; i<5; i++)
		    Data[i] = 0;
	    int pos = SearchUsers(SharedPerson);
	    if (pos == -1)
	    {
		    printf("Could not share view with person %s\n", SharedPerson.c_str());
		    return DTK_CONTINUE;
	    }
	    //Find out which user they are
	     users[pos]->getPosition(Data);
	     //Put the position into a pfCoord (only way for it 2 work)
	     pfCoord MyPos;
	     MyPos.xyz[0] = Data[0];
	     MyPos.xyz[1] = Data[1];
	     MyPos.xyz[2] = Data[2];
	     MyPos.hpr[0] = Data[3];
	     MyPos.hpr[1] = Data[4];
	     MyPos.hpr[2] = Data[5];
	    dtkNavList* NavList = (dtkNavList*)app->check("dtkNavList", DTKNAVLIST_TYPE);
	    app->display()->worldToDpf(MyPos);
	     //Put myself in their eyes
	    for (int i=0; i<3; i++)
	    {
		    NavList->current()->location.d[i] = MyPos.xyz[i];
		    NavList->current()->location.d[i+3] = MyPos.hpr[i];
	    }
	    //The old way of setting the navigation
	    //app->display()->currentNav->location=MyPos;
    }
  
//  }

  CheckMessages();
 
  return DTK_CONTINUE; 
}

// Not so Deep Copy
vector<duser*> dclient::getUsers()
{
  /*
  vector<duser*> pass;
  duser* newuser;
  for(int i=0;i<users.size();i++){
    newuser = new duser(users[i]->getName(),users[i]->getAvatar());
    pass.push_back(newuser);
  }
  return pass;
  */
  return users;
}

int dclient::CheckMessages()
{
//  string* doh = NULL;
//  usleep(1000);
//  string doh2 = *doh;
  if(parser->NextMessage() == -1) return -1;

  if(parser->getFrom() == compName) return -1;

  if(parser->getTo() != compName && parser->getTo() != "everyone"){
   cout<<"Message not for me"<<endl;
   return -1;
  }
  
  string avatarname;
  string message = parser->getMessage();
  string temp;
  unsigned int avatarcounter =0;

  //Check to see what message I got
  switch(message[0]){
    case 'n': //newuser
      temp = parser->getString(); //temp will be equal to the new user name
      avatarname = parser->getString();
      addNewUser(temp,avatarname);
      break;
    case 'u': //usernames
      if(message == "usernames"){
        temp = parser->getString();
        while(temp != ""){
          addNewUser(temp,"");
          temp = parser->getString();
        }
      }
      else if(message == "useravatars"){
        temp = parser->getString();
        while(temp!=""){
          if(avatarcounter < users.size()){
            cout<<"setting avatar "<<temp.c_str()<<endl;
            users[avatarcounter]->setAvatar(temp);
            avatarcounter++;
            temp = parser->getString();
          }
        }
      }
      break;
    case 'd':
      temp = parser->getString();
      disconnect(temp);
      break;

    //Jump to a specific users position
    case 'j':
    {
	    temp = parser->getString();
	    printf("Moving to user-%s's position\n", temp.c_str());
	    int pos = SearchUsers(temp);
	    if (pos == -1)
	    {
		    printf("User %s not found\n", temp.c_str());
		    return -1;
	    }
	    float Data[6];
	    for (int i=0; i<5; i++)
		    Data[i] = 0;
	    users[pos]->getPosition(Data);
	    pfCoord MoreTemp;
	    MoreTemp.xyz[0] = Data[0] - 100;
	    MoreTemp.xyz[1] = Data[1];
	    MoreTemp.xyz[2] = Data[2];
	    MoreTemp.hpr[0] = Data[3];
	    MoreTemp.hpr[1] = Data[4];
	    MoreTemp.hpr[2] = Data[5];
	    dtkNavList* NavList = (dtkNavList*)app->check("dtkNavList", DTKNAVLIST_TYPE);
	    app->display()->worldToDpf(MoreTemp);
	     //Put myself in their eyes
	    for (int i=0; i<3; i++)
	    {
		    NavList->current()->location.d[i] = MoreTemp.xyz[i];
		    NavList->current()->location.d[i+3] = MoreTemp.hpr[i];
	    }
	    //app->display()->currentNav->location=MoreTemp;
	    break;
    }
    //Tether to a specific user
    case 't':
    {
	    if (SharedView)
		    return 0;
	    TetheredTo = parser->getString();
	    Tethered = true;
	    break;
    }
    //Free the user from being tethered or from shared view
    case 'f':
    {
	    //Un-tether them (relativley simple)
	    if (message == "free-tether")
	    {
		    Tethered = false;
		    return 0;
	    }
	    //Free them from shared view and return them to their original position
	    //before having shared view
	    if (message == "free-sharedview")
	    {
		SharedView = false;
		int pos = SearchUsers(SharedPerson);
		if (pos == -1)
		{
			printf("Could not unshare user %s's view\n", SharedPerson.c_str());
			return 0;
		}
		SharedPerson="";
		users[pos]->hideAvatar(false, app);
		    dtkNavList* NavList = (dtkNavList*)app->check("dtkNavList", DTKNAVLIST_TYPE);
	    app->display()->worldToDpf(oldLocation);
		     //Put myself in their eyes
		    for (int i=0; i<3; i++)
		    {
			    NavList->current()->location.d[i] = oldLocation.xyz[i];
			    NavList->current()->location.d[i+3] = oldLocation.hpr[i];
		    }
		    
		//	app->display()->currentNav->location = oldLocation;
		return 0;
	    }
	    break;
    }
    //Shared view for the user
    case 's':
    {
	    if (Tethered)
		    return 0;
	    string TempString = parser->getString();
	    int pos = SearchUsers(TempString);
	    if (pos == -1)
	    {
		    printf("Could not find user %s\n", TempString.c_str());
		    return 0;
	    }
	    if (SharedPerson == "")
	    {
		    dtkNavList* NavList = (dtkNavList*)app->check("dtkNavList", DTKNAVLIST_TYPE);
		    app->display()->worldToDpf(oldLocation);
		     //Put myself in their eyes
		    for (int i=0; i<3; i++)
		    {
			    NavList->current()->location.d[i] = oldLocation.xyz[i];
			    NavList->current()->location.d[i+3] = oldLocation.hpr[i];
		    }
		    //oldLocation = app->display()->currentNav->location;
	    }
	    SharedPerson=TempString;
	    users[pos]->hideAvatar(true, app);
	    SharedView = true;
	    return 0;
	    break;
    }
    case 'h':
    {
	    PersonToHighlight = parser->getString();
	    int pos = SearchUsers(PersonToHighlight);
	    if (pos == -1)
	    {
		    printf("Could not find user %s\n", PersonToHighlight.c_str());
		    return 0;
	    }
	    ElapsedTime =0;
	    users[pos]->Highlight(true, app);
	    Highlight = true;
	    Time->delta();
	break;
    }


    default:
      //cout<<compName<<": I did not recognize the command "<<message<<endl;
      break;
  }

  return 0;
}

int dclient::disconnect(string user)
{
  int pos = SearchUsers(user);

  if(pos == -1) return -1; //user not found;

  if(users[pos] == NULL) return -1;
  cout<<"dropping user "<<users[pos]->getName().c_str()<<endl;
  users[pos]->removeDCS(app,mapSCS);
    
  delete users[pos];
  users.erase(users.begin()+pos);
  cout<<"Leaving disconnect"<<endl;
  return 0;
}

int dclient::addNewUser(string newguy,string avatar)
{
  int pos=-1; //position in vectors

  pos = SearchUsers(newguy);

  if(pos != -1){
    if(users[pos]->getName() == newguy){
     cout<<"Already have user: "<<users[pos]->getName().c_str()<<endl;
     return -1;
    }
  }

  duser* newUser = new duser(newguy,avatar);
  users.push_back(newUser);

  //Connect Remote Shared Mem
  
  newguy = "user-" + newguy;

  if(ClientManager->connectRemoteSharedMem(newguy.c_str(),host) == -1){
      cout<<"Could not connect shm for the newguy"<<endl;
      return -1;
  }
  
  return 0;
}

///////////////////////////////////////////
//    Returns where in the clientNameList vector
//    a client is or the closet location
///////////////////////////////////////////

int dclient::SearchUsers(string user)
{
  for(unsigned int i=0;i<users.size();i++)
    if(users[i]->getName() == user) return i;

  return -1;
/*
  int size = clientNameList.size();
  int low = 0;
  int high = size;
  string temp;
  int mid = (high+low)/2;

  for(mid = (high+low)/2;low<=high;){
    temp = clientNameList[mid];
    if(temp == user)
      return mid;
    if(user > temp){
        low = mid+1;
    }
    else
      high = mid-1;
  }

  return mid;  
*/
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new dclient(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
