
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Scroll.H>
#include <FL/Fl_Button.H>

Fl_Scroll *Listing=(Fl_Scroll*)0;
Fl_Button *Prev=(Fl_Button *)0;
Fl_Button *Next=(Fl_Button *)0;
Fl_Button *Close=(Fl_Button *)0;
Fl_Button* lastButton;
int dsoCount;
class dsow;
dsow* ptrDsow;

class dsow : public dtkAugment {

	public:
  		dsow(dpf *app);
  		~dsow(void);

  		int postConfig(void);
  		int postFrame(void);

	private:

  		KeyCode xkc_p;

	  	dpf *app;
  		dpfDisplay *display;

	  	Fl_Window *guiWin;
  
  		friend void update(dpfNav*);
		friend void dso_callback(Fl_Button*, dpfNav*);
  		friend void switch_callback(Fl_Button*, void*);
  		friend void prev_callback(Fl_Button*, void*);
  		friend void next_callback(Fl_Button*, void*);
		friend void close_callback(Fl_Button*, dsow*);
};

Fl_Window* make_window(dsow*);

// constructor
dsow::dsow(dpf *app_in) :
  dtkAugment("dsow") {

	setDescription("Allows users to switch Navigation DSO's via graphical interface.\n");
  	app = app_in;
  	display = NULL;
  	guiWin = NULL;
  	dsoCount = 0;

  	// dtkAugment::dtkAugment() will not validate the object
  	validate();
}

// destructor
dsow::~dsow(void) {
	
	if(guiWin){
		delete guiWin;
		guiWin = NULL;
	}
}

// This tries to find a reset location and scale
int dsow::postConfig(void) {

  	display = app->display();
  	if(!display) return DTK_CALLBACK_ERROR;

	guiWin = make_window(this);
	if(!guiWin) return DTK_CALLBACK_ERROR;
	
	// initialized the list of dso's
	dpfNav *tempNav = display->firstNav(),
	       *d = NULL;
	Fl_Button *b = NULL;
	int voffset = 20;

	// loops through all loaded dso,and adds them to the package	
	while (tempNav!= NULL ){

		if(tempNav != NULL) {

			b = new Fl_Button(20,voffset,260,25, tempNav->getName());	//245,25
			if(b) {
				b->labeltype(FL_EMBOSSED_LABEL);
				b->callback((Fl_Callback*)dso_callback, tempNav);
				Listing->add(b);
				lastButton = b;
				b = NULL;
			}
		}
		
		d = display->nextNav(tempNav);
		tempNav = d;
		d = NULL;
	
		dsoCount++;
		voffset += 30;
	}

	// changes colors for the current Nav DSO button
	lastButton->color(137);
	lastButton->labeltype(FL_NORMAL_LABEL);
	lastButton->labelcolor(49);
	Listing->redraw();

	tempNav = d = NULL;
	ptrDsow = this;

  	// just call dsow::preFrame() once at the start.
  	return DTK_REMOVE_CALLBACK;
}

int dsow::postFrame(void) {
	
	Fl::check();
	return DTK_CONTINUE;
}

/***********************************************
		Callbacks 
 **********************************************/

// changes the colors of buttons and switches the navigation
// to the selected NAV dso
void dso_callback(Fl_Button* currButton, dpfNav* newNav){

	// if the last button pointer is not null
	if(lastButton) {

		// turn the last button back to origional colors
		lastButton->labeltype(FL_NORMAL_LABEL);
		lastButton->labelcolor(0);
		lastButton->color(48);
		lastButton->labeltype(FL_EMBOSSED_LABEL);
		lastButton->redraw();
	
		// highlight current button
		currButton->labeltype(FL_NORMAL_LABEL);
		currButton->color(137);
		currButton->labelcolor(49);
		currButton->redraw();

		// update to the new Nav DSO
		if(newNav) update(newNav);
						
		// assign currnet button to last button
		lastButton = currButton;
	} else
		printf("lastButton = NULL\n");
}
// changes the colors of the buttons, and switches to the prev
// NAV dso in teh list
void prev_callback(Fl_Button*, void*) {

	int index = Listing->find(lastButton);

	// loop through list mod number of nav dso's
	index = (index-1) % dsoCount;
	if(index < 0) index = dsoCount-1;
	
	
	Fl_Button* prevButton = (Fl_Button*) Listing->child(index);

	// change to color characteristics of buttons to show which is selected
	if(prevButton){
			
		lastButton->labeltype(FL_NORMAL_LABEL);
	        lastButton->labelcolor(0);
		lastButton->color(48);
		lastButton->labeltype(FL_EMBOSSED_LABEL);
		lastButton->redraw();
			
		// highlight the prev  button
		prevButton->labeltype(FL_NORMAL_LABEL);
		prevButton->color(137);
		prevButton->labelcolor(49);
		prevButton->redraw();

		dpfNav* newNav = (dpfNav*) prevButton->user_data();
		if(newNav) update(newNav);

		lastButton = prevButton;
	}
}

// changes the colors of the buttons, and switches to the next
// NAV dso in the list
void next_callback(Fl_Button*, void* ) {
	
	int index = Listing->find(lastButton);
	
	index = (index+1) % dsoCount;

	Fl_Button* nextButton = (Fl_Button*) Listing->child(index);
	if(nextButton){

		lastButton->labeltype(FL_NORMAL_LABEL);
		lastButton->labelcolor(0);
		lastButton->color(48);
		lastButton->labeltype(FL_EMBOSSED_LABEL);
		lastButton->redraw();
		
		// highlight the prev  button
	        nextButton->labeltype(FL_NORMAL_LABEL);
	        nextButton->color(137);
	        nextButton->labelcolor(49);
		nextButton->redraw();
		
		dpfNav* newNav = (dpfNav*) nextButton->user_data();
	//	if(newNav) dsow* update(newNav);
		if(newNav) update(newNav);

		lastButton = nextButton;
	}
}

// closes the gui
void close_callback(Fl_Button*, dsow *data){
	data->guiWin->hide();
}

// updates the naigation to the next desired one
void update(dpfNav* next_nav) {

	dsow* data = ptrDsow;

	dpfNav *nav = data->display->currentNav;
	dtkMsg.add(DTKMSG_INFO,"dsow::postFrame():"
				" switching navigation from \"%s%s%s\" to ",
	 			dtkMsg.color.vil, nav->getName(), dtkMsg.color.end);
	      			
	// Go to the next naviagtion in the list.
	float response = nav->response();

	if(!next_nav)
		next_nav = data->display->firstNav();

	dtkMsg.append("\"%s%s%s\".\n",dtkMsg.color.vil,
	  		next_nav->getName(), dtkMsg.color.end);

	if(next_nav != nav) {

		// move to where it left off
		next_nav->location.xyz = nav->location.xyz;
		next_nav->location.hpr = nav->location.hpr;
		next_nav->scale = nav->scale;
		next_nav->response(response);
		next_nav->preFrame();

		// pivot is copied only if both are not NULL
		pfCoord *cp = nav->pivot() ;
		pfCoord *ncp = next_nav->pivot() ;
  				
		if (cp && ncp) {
			next_nav->pivot(nav->pivot()) ;
			next_nav->dpfPivot(nav->dpfPivot()) ;
		}

		// velocity is copied only both are not NULL
		pfCoord *vp = nav->velocity() ;
		pfCoord *nvp = next_nav->velocity() ;
		  				
		if (vp && nvp) 
  			*nvp = *vp ;
		data->display->switchNav(next_nav);
	}
}


// creates the gui interface
Fl_Window* make_window(dsow* data) {
	Fl_Window* w;
	
	{ Fl_Window* o = new Fl_Window(300, 320, "Currently loaded Nav DSO's");	//285, 378
		w = o;
		{ Fl_Scroll* o = Listing = new Fl_Scroll(10,10,280,270);
			o->box(FL_EMBOSSED_FRAME);
			o->labeltype(FL_ENGRAVED_LABEL);
			o->end();
			Fl_Group::current()->resizable(o);
		}
		{ Fl_Button* o = Close = new Fl_Button(95, 285, 110, 25, "Close");	//95,285,95,25
			o->labeltype(FL_EMBOSSED_LABEL);
			o->callback((Fl_Callback*) close_callback,data);
		}
		{ Fl_Button* o = Prev = new Fl_Button(10, 285, 80, 25, "@<");	//10,285,80,25
		        o->labeltype(FL_SYMBOL_LABEL);
			o->callback((Fl_Callback*) prev_callback);
		}
		{ Fl_Button* o = Next =new Fl_Button(210, 285, 80, 25, "@>");	//195,285,80,25
			o->labeltype(FL_SYMBOL_LABEL);
			o->callback((Fl_Callback*) next_callback);
		}
		o->end();
	}

	w->show();

	return w;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new dsow(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}


