// The user class for the dedicated server

#ifndef DUSER_H
#define DUSER_H

#include<dtk.h>
#include<string>
#include<iostream>
#include<dpf.h>

using namespace std;

class duser
{
  public:
    duser(string pname,string pavatar);
    ~duser();

    string getName() {return name;}
    string getAvatar() {return avatar;}
    int getPosition(float pos[]);
    pfDCS* getBody(){return body;}
    pfDCS* getMaker(){return marker;}
    int setAvatar(string pavatar) {avatar = pavatar; return 0;} //you need this don't take it out
    
    int initDcs(pfSwitch* pDCSs,float scale);
    int updatePos();
    int updateAvatarMarker();
    int updateAvatarMarker(float basePos[6]);
    int removeDCS(dpf* app, pfGroup* mapSCS);
    int hideAvatar(const bool& ToHide, dpf* app);
    int Highlight(const bool& Value, dpf* app);
/*    
    int setCheckIn(long double newTime) {
      CheckIn = newTime;
      return 0;
    }
    long double getCheckIn() { return CheckIn;}
*/    
    bool operator!=(const duser& Compare) const;
    bool operator == (const duser& Compare) const;
    bool operator < (const duser& Compare) const;
    bool operator > (const duser& Compare) const;
    
  private:
    string name;
    string avatar;
    dtkSharedMem* position;
    pfSwitch* DCSs;
    pfDCS* body; //actual avatar
    pfDCS* marker; //fov marker
    float mapScale;
    pfDCS* Highlighter;
    pfNode* ghost;
//    long double CheckIn; // Last check in time (dtkTime)
};

duser::duser(string pname, string pavatar)
{
  name = pname;
  avatar = pavatar;
  string shmFileName = "user-" + name;

  position = new dtkSharedMem(6*sizeof(float),shmFileName.c_str());
  if(!position || position->isInvalid() || position->getSize() < 6*sizeof(float))
  {
    cout<<"Could not get shared mem for "<<name.c_str()<<endl;
  }

  DCSs = NULL;
  body = NULL;
  marker = NULL;

  ghost =NULL;
  printf("Constructor ended\n");
  //Load a bounding sphere into the world for use with the highlighting concept
}

duser::~duser()
{
  position=NULL;
}
int duser::Highlight(const bool& Value,dpf* app)
{
	if (Value)
	{       
		printf("Beginning highlight\n");
		if (Highlighter == NULL || ghost == NULL)
			return 0;
		 app->display()->world()->addChild(Highlighter);
		 //app->display()->world()->addGhostNode(Highlighter);
		 Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,1,
				  PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);

		 pfSphere sphere = dpf_getBSphere(body);
		 if (sphere.radius > 0.f) 
		 {
			pfMatrix mat ;
			mat.makeTrans(sphere.center[PF_X],

			sphere.center[PF_Y], sphere.center[PF_Z]) ;
			mat.preScale(sphere.radius, sphere.radius, sphere.radius, mat) ;
			Highlighter->setMat(mat) ;
			Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,1,
			PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
		} 
		else 
		{
			Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,0,
			PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
		}
	}
	else
	{
		printf("I am removing myself!\n");
		Highlighter->setTravMask(PFTRAV_APP|PFTRAV_CULL|PFTRAV_DRAW,0,
			PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
		 app->display()->world()->removeChild(Highlighter);
		 //app->display()->world()->removeGhostNode(Highlighter);
	}
	printf("Exiting function!\n");
	return 1;
}

int duser::hideAvatar(const bool& ToHide, dpf* app)
{
	if (ToHide)
	{
		  if (body != NULL)
		    app->display()->world()->removeChild(body);
	}
	else
	{
		  if (body != NULL)
		    app->display()->world()->addChild(body);
	}
	return 1;
}


int duser::removeDCS(dpf* app,pfGroup* mapSCS)
{
  if (body != NULL)
    app->display()->world()->removeChild(body);
  mapSCS->removeChild(DCSs);
  body = NULL;
  marker = NULL;
  return 0;
}

bool duser::operator!=(const duser& Compare) const
{
  if(name != Compare.name)
    return true;
  else
    return false;
}

bool duser::operator==(const duser& Compare) const
{
  if(name == Compare.name)
    return true;
  else
    return false;
}

bool duser::operator < (const duser& Compare) const
{
  if(name < Compare.name)
    return true;
  else
    return false;
}

bool duser::operator > (const duser& Compare) const
{
  if(name > Compare.name)
    return true;
  else
    return false;
}

int duser::getPosition(float pos[])
{
  position->read(pos);

  if(position->differentByteOrder()){
    int j;
    for(j=0;j<6;j++)
		dtk_swapBytes(&pos[j],sizeof(float));
  }

  return 0;
}

int duser::updateAvatarMarker()
{
  float pos[6];
  position->read(pos);
  if(position->differentByteOrder()){
    int j;
    for(j=0;j<6;j++)
		dtk_swapBytes(&pos[j],sizeof(float));
  }
	 
  // The fov pointer needs to have -45 added to the heading and
  // needs to be flipped up by setting the pitch as 90
  marker->setTrans(pos[0]/mapScale, pos[1]/mapScale, 0.020);
  marker->setRot(-45+pos[3], 90, 0);

  if(body == NULL) return 0;

  body->setTrans(pos[0],pos[1],pos[2]);
  body->setRot(pos[3],pos[4],pos[5]);

  DCSs->setVal(PFSWITCH_ON);

  return 0;
}

int duser::updateAvatarMarker(float basePos[6])
{
  pfCoord theirPos;
  position->read(&theirPos);
  if(position->differentByteOrder()){
    int j;
    for(j=0;j<3;j++){
      dtk_swapBytes(&theirPos.xyz[j],sizeof(float));
      dtk_swapBytes(&theirPos.hpr[j],sizeof(float));
    }
  }

  pfCoord myPos;
  myPos.xyz.set(basePos[0], basePos[1], basePos[2]);
  myPos.hpr.set(basePos[3], basePos[4], basePos[5]);

  pfVec3 resultPos = theirPos.xyz - myPos.xyz;
  double length = resultPos.length();

  if(length > (mapScale/2)){  //
    DCSs->setVal(PFSWITCH_OFF); //Diplays only the body which is at index 0.
  }
  else{
    //DCSs->setVal(PFSWITCH_ON); //Displays both body and marker.
    DCSs->setVal(0); //Displays both body and marker.
  }

  // The fov pointer needs to have -45 added to the heading and
  // needs to be flipped up by setting the pitch as 90
  marker->setRot(-45+theirPos.hpr[0], 90, 0);
  marker->setTrans(resultPos[0]/mapScale, resultPos[1]/mapScale, 0.020);

  if(body == NULL) return 0;

  body->setTrans(theirPos.xyz[0],theirPos.xyz[1],theirPos.xyz[2]);
  body->setRot(theirPos.hpr[0],theirPos.hpr[1],theirPos.hpr[2]);

  return 0;
}

int duser::updatePos()
{
  float pos[6];
  position->read(pos);
  if(position->differentByteOrder()){
    int j;
    for(j=0;j<6;j++)
		dtk_swapBytes(&pos[j],sizeof(float));
  }
  position->write(pos,6*sizeof(float));
  return 0;
}

int duser::initDcs(pfSwitch* pDCSs,float scale)
{
  DCSs = pDCSs;
  if (DCSs->getNumChildren() >= 2)
    body = (pfDCS*) DCSs->getChild(1);
  marker = (pfDCS*) DCSs->getChild(0);
  mapScale = scale;
  printf("In initdcs, beginning highlight\n");

  char* baseName = getenv("DPF_BASEDIR");
  printf("My base name is %s\n", baseName);
  if (baseName == NULL)
	  printf("DPF_BASEDIR is not set, do not trust program results\n");
  string Path = ".:./models:../models:" + string(baseName) + "/DataFiles/avatars:" 
	  + string(baseName) + "/DataFiles/radar:" + string(DPF_DATA_DIR);
  printf("Path is %s\n", Path.c_str());
  pfFilePath(Path.c_str());

  Highlighter = new pfDCS();
  if (!Highlighter)
	  return 0;
  printf("Highlighter ok\n");
  ghost = pfdLoadFile("purple.pfb") ;
  if (ghost == NULL)
	  printf("Could not find sphere.pfb in duser.h\n");
  printf("1\n");
  Highlighter->addChild(ghost);
  printf("2\n");
  return 0;
}

#endif
