// This is the client to connect to the dedicated server

#include "dclient.h"
#include <dtk/dtkDSO_loader.h>
#include <sstream>
#include <ctime>

dclient::dclient(dpf *app_in) :
  dtkAugment("dclient")
{
  cout<<"dclient starting"<<endl;
  setDescription("connects to the server");
  app = app_in;
  validate();
}

dclient::~dclient(void)
{
  string ByeBye = compName;
  ByeBye+= "\tserver\t";
  ByeBye+= getTime();
  ByeBye+= "\tdisconnect";

  writeMessage(ByeBye);
  
  for(int i=0;i<users.size();i++){
    delete users[i];
  }

  if(positionShm != NULL)
    delete positionShm;

  if(genMessage!=NULL)
    delete genMessage;
  if(ClientManager) delete ClientManager;
}

int dclient::postConfig(void)
{
  cout<<"in postConfig"<<endl;
  ClientManager = new dtkClient;

  char buff[64];
  unsigned int size = 64;
  gethostname(buff,size);
  compName = buff;

  //Setting up the General Message shared Mem
  
  genMessage = new dtkSharedMem(SHM_SIZE,"genMessage");
  if(!genMessage || genMessage->isInvalid() || genMessage->getSize() < SHM_SIZE)
  {
    cout<<"could not get dclient sharedmem"<<endl;
  }
  else{
    cout<<compName<<" dclient running."<<endl;
    genMessage->flush();
    genMessage->queue();
  }

  //DPF_SHARED_SERVER is the env varible.

  host = getenv("DPF_SHARED_SERVER");
  if(host == NULL){
    cout<<"no host name specified"<<endl;
    return -1;
  }

  if(ClientManager->connectRemoteSharedMem("genMessage",host) == -1)
    cout<<compName<<" could not connect to "<<&host<<endl;
  else
    cout<<compName<<" connected to "<<host<<endl;

  string avatarname = getenv("DPF_AVATAR_NAME");
  string addme = compName;
  addme += "\tserver\t";
  addme += getTime();
  addme += "\taddme\t";
  addme += avatarname;

  writeMessage(addme);

  usleep(300000); //Sleep for 1/3 of a second while server creates shm for it.

  lastTime = Timer.delta();

  string tempname = "user-" + compName;

  positionShm = new dtkSharedMem (6*sizeof(float),tempname.c_str());
  if(!positionShm || positionShm->isInvalid() || positionShm->getSize() < 6*sizeof(float))
  {
    cout<<"Could not get shared mem for "<<tempname<<endl;
  }

  return DTK_CONTINUE;
}

int dclient::postFrame(void)
{
  long double currTime = Timer.get();
  if(currTime - lastTime > .05 ){  //update every 1/20 seconds.
    lastTime = currTime;
    //  Convert 0,0,0 from scene to world units
    coord.xyz.set(0,0,0);
    coord.hpr.set(0,0,0);
    app->display()->dpfToWorld(&coord);

    //Access the position with coord.xyz and coord.hpr.
   
    positionShm->write(&coord,6*sizeof(float));
  
  }

  CheckMessages();
 
  return DTK_CONTINUE; 
}

// Not so Deep Copy
vector<duser*> dclient::getUsers()
{
  /*
  vector<duser*> pass;
  duser* newuser;
  for(int i=0;i<users.size();i++){
    newuser = new duser(users[i]->getName(),users[i]->getAvatar());
    pass.push_back(newuser);
  }
  return pass;
  */
  return users;
}

int dclient::writeMessage(string Message)
{
  cout<<"trying to write message"<<endl;
  
  Message += "\0";
  if(genMessage->write(Message.c_str(),(Message.length()+1)*sizeof(char))){
    cout<<"Error writing to genMessage"<<endl;
    return -1;
  }

  cout<<"done writing messate"<<endl;

  return 0;
}

string dclient::getTime()
{
  //The tm struct;
/*
   typedef struct {
   int tm_hour;    hour (0 - 23) 
   int tm_isdst;   daylight saving time enabled/disabled 
   int tm_mday;    day of month (1 - 31) 
   int tm_min;     minutes (0 - 59) 
   int tm_mon;     month (0 - 11 : 0 = January) 
   int tm_sec;     seconds (0 - 59) 
   int tm_wday;    Day of week (0 - 6 : 0 = Sunday) 
   int tm_yday;    Day of year (0 - 365) 
   int tm_year;    Year less 1900 
   }
*/
  time_t rawtime;
  time(&rawtime);
  tm* currtime = localtime(&rawtime);
  char teatime[10];
  sprintf(teatime,"%d:%d:%d",currtime->tm_hour,currtime->tm_min,currtime->tm_sec);
  
  return teatime;
}

int dclient::CheckMessages()
{
  char buff[1024];
  if(!genMessage->qread(buff)) return -1;

  string from,to,time,message,temp;

  istringstream istr(buff);
  istr>>from>>to>>time>>message;

  if(from == compName) return -1;

  if(to != compName && to != "everyone"){
   cout<<"Message not for me"<<endl;
   return -1;
  }
  
  string avatarname;
  int avatarcounter =0;

  switch(message[0]){
    case 'n': //newuser
      istr>>temp; //temp will be equal to the new user name
      istr>>avatarname;
      addNewUser(temp,avatarname);
      break;
    case 'u': //usernames
      if(message == "usernames"){
        istr>>temp;
        while(istr){
          addNewUser(temp,"");
          istr>>temp;
        }
      }
      else if(message == "useravatars"){
        istr>>temp;
        while(istr){
          if(avatarcounter < users.size()){
            cout<<"setting avatar "<<temp<<endl;
            users[avatarcounter]->setAvatar(temp);
            avatarcounter++;
            istr>>temp;
          }
        }
      }
      break;
    case 'd':
      istr>>temp;
      disconnect(temp);
      break;
    default:
      cout<<compName<<": I did not recognize the command "<<message<<endl;
      break;
  }

  return 0;
}

int dclient::disconnect(string user)
{
  int pos = SearchUsers(user);

  if(pos == -1) return -1; //user not found;

  if(users[pos] == NULL) return -1;
    
  delete users[pos];
  users.erase(users.begin()+pos);
  return 0;
}

int dclient::addNewUser(string newguy,string avatar)
{
  int pos=-1; //position in vectors

  pos = SearchUsers(newguy);

  if(pos != -1){
    if(users[pos]->getName() == newguy){
     cout<<"Already have user"<<endl;
     return -1;
    }
  }

  duser* newUser = new duser(newguy,avatar);
  users.push_back(newUser);

  //Connect Remote Shared Mem
  
  newguy = "user-" + newguy;
  
  while(ClientManager->connectRemoteSharedMem(newguy.c_str(),host) == -1){
    usleep(300000); //Sleep for 1/3 of a second while server creates shm for it.
  }

  return 0;
}

///////////////////////////////////////////
//    Returns where in the clientNameList vector
//    a client is or the closet location
///////////////////////////////////////////

int dclient::SearchUsers(string user)
{
  for(int i=0;i<users.size();i++)
    if(users[i]->getName() == user) return i;

  return -1;
/*
  int size = clientNameList.size();
  int low = 0;
  int high = size;
  string temp;
  int mid = (high+low)/2;

  for(mid = (high+low)/2;low<=high;){
    temp = clientNameList[mid];
    if(temp == user)
      return mid;
    if(user > temp){
        low = mid+1;
    }
    else
      high = mid-1;
  }

  return mid;  
*/
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new dclient(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
