#include <stdio.h>
#include <dpf.h>
#include <dtk.h>
#include <dtk/dtkDSO_loader.h>
#include <string>
#include <vector>
using namespace std;
#include "Messenger.h"
#include "MessageParser.h"
// example generic dtkAugment DSO modified to be a file loader for 
// any set world
// Notice how simple it is to take an example program and modify it
struct Item
{
	int ID; //The items unique id
	string Name; //The description of the item
	pfNode* Node; //The actual pointer to the object in the world
};

class Generic2 : public dtkAugment {
public:

  Generic2(dpf *app) ;
  ~Generic2(void) ;
  //During the post frame step handle the navigation
  int postFrame(void);

private:

  //The application to play with
  dpf* App;
  //Gets the messages sent to it
  MessageParser* Parser;
  //Sends messages to the client
  Messenger* Message;
  Messenger* Global;
  //The list of items
  vector<Item> Items;
  int NextID;
};

// create object and set description.
// be sure to validate it!
Generic2::Generic2(dpf *app) : dtkAugment("FileLoaderDSO")
{
	App = app;
	Parser = new MessageParser("genMessage");
	Message = new Messenger("CollabFile");
	Global = new Messenger("genMessage");
	unsigned int Size = 0;
	char* Hostname = new char[Size];
	gethostname(Hostname, Size);
	Message->setFrom("fileloader");
	Message->setTo(Hostname);
	NextID=1;
	validate();
} 

// destructor
Generic2::~Generic2(void) 
{
 	delete Message;
}



//Every hundred frames output the user list to the collaborative gui
int Generic2::postFrame(void) 
{
	if (Parser->NextMessage() == -1)
		return DTK_CONTINUE;
	if (Parser->getMessage() == "add-file") 
	{
		Item Temp; //Temporary item
		//Get the file to load in the world
		string File2Load = Parser->getString();
		for (unsigned int i=0; i<Items.size(); i++)
		{
			if (Items[i].Name == File2Load)
				return DTK_CONTINUE;
		}
		Temp.Name = File2Load;
		//Get the path to load the file from
		string Path = Parser->getString();
		pfFilePath(Path.c_str());
		//Set my file id
		Temp.ID = NextID;
		//Increment the id
		NextID++;
		//Tell the gui what to add to the list of files loaded
		Message->setFrom("ourfriend");
		Message->setTo("fileguis");
		Message->setMessage("UpdateFileGui");
		Message->addString(File2Load);
		Message->SendMessage();
		//Load the file into a pfNode
		Temp.Node = pfdLoadFile(File2Load.c_str());
		//Add the node to the world
		App->display()->world()->addChild(Temp.Node);
		//Add the item to the item list
		Items.push_back(Temp);
		//Return
		return DTK_CONTINUE;
	}
	if (Parser->getMessage() == "load-file")
	{
		Item Temp;
		string File2Load = Parser->getString();
		for (unsigned int i=0; i<Items.size(); i++)
		{
			if (Items[i].Name == File2Load)
				return DTK_CONTINUE;
		}
		Temp.Name = File2Load;
		//Get the path to load the file from
		string Path = Parser->getString();
		pfFilePath(Path.c_str());
		//Set my file id
		Temp.ID = NextID;
		//Increment the id
		NextID++;
		//Tell the gui what to add to the list of files loaded
		Message->setFrom("ourfriend");
		Message->setTo("fileguis");
		Message->setMessage("UpdateFileGui");
		Message->addString(File2Load);
		Message->SendMessage();
		//Load the file into a pfNode
		Temp.Node = pfdLoadFile(File2Load.c_str());
		//Add the node to the world
		App->display()->world()->addChild(Temp.Node);
		//Add the item to the item list
		Items.push_back(Temp);
		return DTK_CONTINUE;
	}
	if (Parser->getMessage() == "remove-file" )
	{
		//Get the file to remove
		string Temp = Parser->getString();
		printf("Removing item %s\n", Temp.c_str());
		//If it is bad, ignore
		if (Temp == "")
			return DTK_CONTINUE;
		//Try to find the item
		for (unsigned int i=0; i< Items.size(); i++)
		{
			//Item found
			if (Items[i].Name == Temp)
			{
				Message->setFrom("ourfriend");
				Message->setTo("fileguis");
				Message->setMessage("remove-gui");
				Message->addString(Items[i].Name);
				Message->SendMessage();
				//Remove the item
				App->display()->world()->removeChild(Items[i].Node);
				//Remove the item from the item list
				Items.erase(Items.begin()+i);
			}
		}
		return DTK_CONTINUE;
	}
	if (Parser->getMessage() == "get-files-loaded")
	{
		Global->setFrom("ourfriend");
		Global->setTo("fileguis");
		Global->setMessage("files-loaded-are");
		Global->addInt(Items.size());
		for (unsigned int i=0; i<= Items.size(); i++)
			Global->addString(Items[i].Name);

		Global->SendMessage();

	}
	return DTK_CONTINUE;
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new Generic2(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
