// Eric Tester, etester@vt.edu, wrote the original version of this plays
// back the file created by playbackWrite
//
// the file name is specified by the environment variable
// DPF_PLAYBACK_FILE, if set, otherwise the file used is
// "/tmp/playback-$LOGNAME.log"

#include <stdio.h>
#include <string.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Slider.h>
#include <FL/fl_file_chooser.h>
#include <Fl/fl_message.h>
//#include <FL/filename.H>

#define PLAY true

Fl_Window *PlayBackWin=(Fl_Window *)0;
Fl_Button *Rewind=(Fl_Button *)0;
Fl_Button *RewindStep=(Fl_Button *)0;
Fl_Button *Play=(Fl_Button *)0;
Fl_Button *Pause=(Fl_Button *)0;
Fl_Button *FFStep=(Fl_Button *)0;
Fl_Button *FForward=(Fl_Button *)0;
Fl_Button *CloseGUI=(Fl_Button *)0;
Fl_Return_Button *Quit=(Fl_Return_Button *)0;
Fl_Output *Text=(Fl_Output *)0;
Fl_Slider *StatusBar=(Fl_Slider *)0;
Fl_Button *Reset=(Fl_Button *)0;
Fl_Button *Record=(Fl_Button *)0;

#define File (menu_+0)
#define New (menu_+1)
#define Open (menu_+2)
#define Save (menu_+3)
#define SaveAs (menu_+4)
#define Close (menu_+5)
#define Exit (menu_+6)
	
bool PlayStop = PLAY;
bool PlayPause = PLAY;
bool PlayRecord = PLAY;
bool rs	= PLAY;
bool rw = PLAY;
bool fs = PLAY;
bool ff = PLAY;
bool running = true;
int statusBarMax = 10;

#define DESCRIPTION 	"\npbwr Log file is: %s\n" \
			"  %spbwr commands:  %s\n" \
	           	"  \'%sT%s\'  Toggle- toggle on/off the graphical interface\n"


class pbwr : public dpfNav {

	public:

  		pbwr (dpf* in_app);
  		~pbwr (void);
  		int postConfig (void);		// r
  		int postFrame (void);		// r

	private:
  
  		FILE* file;
  		char* logfile;
  
  		dpf* app;
  		dpfDisplay* display;
  		dtkDequeuer* dequeuer;
  		dtkInXKeyboard *keyboard;
  
 		KeyCode xkc_t;
  
  		float reset_location[6];
  		float prevFileLoc[6];
  		float currFileLoc[6];
  	
  		dtkTime t;
  		long double currFileTime;
  		long double prevFileTime;
  		long double resettime;
  		long double prevtime;
  		pfCoord prevCoord;
  		
  		bool playbackflag;
  		bool recordflag;
  		bool guiflag;
  		Fl_Window *guiWin;
  		
  		// button callbacks
		friend void rewind_callback(Fl_Button*, pbwr*);
		friend void rewindstep_callback(Fl_Button*, pbwr*);
		friend void record_callback(Fl_Button*, pbwr*);
		friend void play_callback(Fl_Button*, pbwr*);
		friend void pause_callback(Fl_Button*, pbwr*);
		friend void ffstep_callback(Fl_Button*, pbwr*);
		friend void fforward_callback(Fl_Button*, pbwr*);
		friend void reset_callback(Fl_Button*, pbwr*);
		friend void closegui_callback(Fl_Button*, pbwr*);
		friend void quit_callback();
		
		// menu callbacks
		friend void new_callback(Fl_Menu_*, pbwr*);
		friend void open_callback(Fl_Menu_*, pbwr*);
		friend void save_callback(Fl_Menu_*, pbwr*);
		friend void save_as_callback(Fl_Menu_*, pbwr*);
		friend void close_callback(Fl_Menu_*, pbwr*);
		friend void exit_callback(Fl_Menu_*, void*);
		
		friend Fl_Window* make_window(pbwr*);
};

Fl_Menu_Item menu_[] = {
 {"File", 0,  0, 0, 64, 5, 0, 14, 0},
 {"New", 0x4006e,  (Fl_Callback*)new_callback, 0, 0, 5, 0, 14, 0},
 {"Open", 0x4006f,  (Fl_Callback*)open_callback, 0, 0, 5, 0, 14, 0},
 {"Save", 0x40073, (Fl_Callback*)save_callback, 0, 0, 5, 0, 14, 0},
 {"Save As", 0x40061,  (Fl_Callback*)save_as_callback, 0, 0, 5, 0, 14, 0},
 {"Close", 0x40063,  (Fl_Callback*)close_callback, 0, 0, 5, 0, 14, 0},
 {"Exit", 0x40071,  (Fl_Callback*)exit_callback, 0, 0, 5, 0, 14, 0},
 {0},
 {0}
};

/*******************************************
		Class Methods
*******************************************/

// constructor
pbwr::pbwr (dpf* in_app) :  dpfNav (in_app,"pbwr") {
  	app = in_app;
  	keyboard = NULL;
  	dequeuer = NULL;
  	display = NULL;
  	guiWin = NULL;
  	logfile = getenv("DPF_PLAYBACK_FILE");
 	
  	char *logfile_mem = NULL;
  	playbackflag = recordflag = guiflag = false;
  	
  	if(!logfile) {
		char *format = "/tmp/playback-%s";
   		char *logname = getenv("LOGNAME");
      	
   		if(!logname) logname = "nonknown_user";
	     	logfile = logfile_mem = (char *) dtk_malloc(strlen(format) + strlen(logname) + 1);
      		sprintf(logfile, format, logname);
      		logname = NULL;
	      	format = NULL;
	}
	
  	setDescription(DESCRIPTION,logfile, dtkMsg.color.rvil, dtkMsg.color.end,
		dtkMsg.color.grn, dtkMsg.color.end,
		dtkMsg.color.red, dtkMsg.color.end,
		dtkMsg.color.grn, dtkMsg.color.end);
			
  	if( (file = fopen(logfile,"a+")) == NULL) {
   		dtkMsg.add(DTKMSG_ERROR,1,
			"pbwr::pbwr() failed to open logfile \"%s\".\n",
		 	logfile);
	//	dtkMsg.add(DTKMSG_NOTICE, "No default log file available\n");
	//	fl_message("No default log file available, creating one."); 	
    		if(logfile_mem) free(logfile_mem);
      		return;
	}

  	if(logfile_mem) free(logfile_mem);

  	logfile_mem = NULL;
//	printf("curlocation = %i\n", ftell(file));
//	fseek(file, -69L , SEEK_END);
//	fscanf(file, "%i", &statusBarMax);
//	printf("statusBarMax = %i\n", statusBarMax);
	rewind(file);		/////

  	validate();
}

// destructor
pbwr::~pbwr (void) {
  
  	if (dequeuer) {
    		delete dequeuer;
	    	dequeuer = NULL;
  	}
  
  	if (file) {
    		fclose (file);
	    	file = NULL;
  	}
  	
  	if(logfile) {
  		delete logfile;
  		logfile = NULL;
  	}
  	
  	if (guiWin) {
  		delete guiWin;
  		guiWin = NULL;
  	}
}

int pbwr::postConfig (void) {

  	dtkMsg.add(DTKMSG_NOTICE,getDescription());

  	dequeuer = new dtkDequeuer(app->record());
  	if(!dequeuer || dequeuer->isInvalid())
    	return DTK_CALLBACK_ERROR;
  
  	display = app->display();
  	if(!display) return DTK_CALLBACK_ERROR;
  
  	keyboard = static_cast<dtkInXKeyboard *>(app->get("xkeyboard", DTKINXKEYBOARD_TYPE));
  	if(!keyboard) return DTK_CALLBACK_ERROR;
  	
  	guiWin = make_window(this);
  	if(!guiWin) { 	
  		dtkMsg.add(DTKMSG_NOTICE,"%s Sorry, unable to create grapical interface.\n",
     				 getName());
		return DTK_CALLBACK_ERROR;
  	}
  
  	//////////////// get X key codes from key syms ///////////////////////
  
  	if( !(xkc_t = keyboard->getKeyCode(XK_t)))
		return DTK_CALLBACK_ERROR;
  
  	keyboard->queue();
  
 	// initialize initial location and time;
  	fscanf (file, "%Lg %f %f %f %f %f %f", &currFileTime,
		&reset_location[0], &reset_location[1],
		&reset_location[2], &reset_location[3],
		&reset_location[4], &reset_location[5]);
  
  	// initialize previous time and location		
	for (int i = 0; i < 3; i++) {
		location.xyz[i]  = currFileLoc[i]   = 
   	 	prevFileLoc[i]   = reset_location[i];
    		location.hpr[i]  = currFileLoc[i+3] = 
    		prevFileLoc[i+3] = reset_location[i+3];
  	}
  
  	resettime = prevFileTime = currFileTime;
  
  	t.reset(currFileTime,0);
	
  	return dpfNav::postConfig();
}


int pbwr::postFrame (void) {

  	dtkRecord_event* e;
  	dpfNav* currentNav = display->currentNav;
  
  	while((e = dequeuer->getNextEvent(keyboard))) {
    
		if(currentNav) {
 			KeyCode key;
      	
      			// if key is pressed
    			if((keyboard->read(&key,e))) {
	  
	  			if (key == xkc_t) {
	    	
					if(guiflag)	guiWin->show();	
					else 		guiWin->hide();
					guiflag = !guiflag;
				}
			}
		
			// records to the log file
			if(recordflag) {
	
//				dtkMsg.add(DTKMSG_NOTICE, "in RECORD Mode\n");
	
				static int rflag = 1;	     		// flag to record info
		  		//static int doubleflag = 0; 		// flag to test for duplicates
			  	pfCoord location;
	  
	  			// get info from nav coord.
			  	location.xyz = currentNav->location.xyz;
		  		location.hpr = currentNav->location.hpr;
	  
			 	// checks if previous coord is the same as current coord
			  	if ( (prevCoord.xyz == location.xyz) &&
	    		    	     (prevCoord.hpr == location.hpr) ) {
	    
					// for duplicates				  				  
	    			  	//if (!doubleflag) doubleflag = 1;
				      	rflag = 0;
	    
	    			} //else if (doubleflag) doubleflag = 0;
			  
			  	// if true print value and to file
				if (rflag) {  
	    				fprintf(file, "%.25Lg %.7g %.7g %.7g %.7g %.7g %.7g\n", prevtime = t.get(),
				    		location.xyz[0], location.xyz[1], location.xyz[2], 
		       				location.hpr[0], location.hpr[1], location.hpr[2]);
	    
		    			prevCoord = location;
				}
		  		rflag = 1;
			}
    		}
	}
	
	// reads from the log file
  	if(playbackflag) {
  	
		//dtkMsg.add(DTKMSG_NOTICE, "in PLAYBACK Mode\n");

		long double time_now = t.get();
   		while(time_now >= currFileTime) {

  			for (int i = 0; i < 3; i++) {
    				prevFileLoc[i]   = currFileLoc[i];
    				prevFileLoc[i+3] = currFileLoc[i+3];
    			}	

	  		prevFileTime = currFileTime;
	  	
  			// reads in from file,  if there is nothing there
  			// prints end of file
//			printf("currloc =%i\n", ftell(file));	
  			if (7!=(fscanf (file, "%Lg%f%f%f%f%f%f", &currFileTime,
	     			&currFileLoc[0], &currFileLoc[1],
	      			&currFileLoc[2], &currFileLoc[3],
	      			&currFileLoc[4], &currFileLoc[5]))) {	

      				Text->value("end of playback file");
      				Play->labelcolor(0);
      				PlayStop = PLAY;      				
	      			playbackflag = 0;
	      			guiWin->redraw();
    				return DTK_CONTINUE;
    			}
		}
				
		// linear interpolates data
	    	for (int i = 0; i < 3; i++) {
			location.xyz[i] = prevFileLoc[i] + (time_now-prevFileTime) *
				(currFileLoc[i] - prevFileLoc[i]) / (currFileTime-prevFileTime);
	 
			location.hpr[i] = prevFileLoc[i+3] + (time_now-prevFileTime) *
				(currFileLoc[i+3] - prevFileLoc[i+3])/ (currFileTime-prevFileTime);
		}
		
		StatusBar->value(time_now);
   	} 
   	
   	Fl::check();
  	
   	return DTK_CONTINUE;
}

/*******************************************
		Menu Callbacks
*******************************************/

// creates a new file for recording
void new_callback(Fl_Menu_*, pbwr* ) {

	fl_message("This will eventually allow users to create new log files.");
/*	pbwr* data = (pbwr*) menu_[1].user_data();
	printf("logfile = %s\n", data->logfile);
	data = NULL;

	// get new file name
	char *tmp = fl_file_chooser("Create a new log file as, *.log","*",NULL);
	
	// if not null
	if(tmp) {
		printf("new file %s\n", tmp);

		fclose (data->file);					// close old file
		free(data->logfile);
			
		sprintf(data->logfile,"%s\n" ,tmp);			// copy over filename
		data->logfile = (char*) malloc(strlen(tmp));
		printf("after malloc\n");
		printf("%s\n", data->logfile);
		
		data->file = fopen(data->logfile,"a+");			// open then new file for recording
		data->file = fopen(tmp,"a+");
		if(!data->file){
			printf("file was not opened properly\n");
		}else 
			printf("file opened ok\n");

	}else
		printf("new file not created\n");

	tmp = NULL;
*/
}		


void open_callback(Fl_Menu_*, pbwr* data) {
	fl_message ("This will eventually allow users to open different log files.");
}

void save_callback(Fl_Menu_*, pbwr* data) {
	fl_message("This will eventually allow users to save changes to currently open log files");
}

void save_as_callback(Fl_Menu_*, pbwr* data) {

	fl_message("This will eventually allow users to save changes to currently open log files");
}

void close_callback(Fl_Menu_*, pbwr*) {

	fl_message("This will eventually allow users to close currently open log.");
/*	if(fl_ask("This will close the current loaded logfile. " \
			  "Do you wish to continue?\n")) {
	
		printf("closed file\n");			 
	} else printf("didnot close filee\n");
	
	pbwr* data = (pbwr*) menu_[5].user_data();

	data = NULL;
*/
}

void exit_callback(Fl_Menu_*, void*) {
	running = !running;
}

/*******************************************
		Button Callbacks
*******************************************/

// rewinds through the log file faster then rewindstep
void rewind_callback (Fl_Button*, pbwr* data) {

	if (rw){
		Rewind->labelcolor(1);		
		RewindStep->labelcolor(0);	rs = PLAY;
//		Play->labelcolor(0);		PlayStop = PLAY;	
		FFStep->labelcolor(0); 		fs = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
//		Pause->activate();
//		Record->deactivate();
		StatusBar->value(0);
		
		if (data->file != NULL) {
			rewind(data->file);
			Text->value("Rewind to the beginning of the file");
		        data->t.reset (data->resettime,0);
		        data->prevFileTime = data->currFileTime = data->resettime;

			 for (int i = 0; i < 3; i++) {
		        	 data->display->currentNav->location.xyz[i] = data->reset_location[i];
				 data->display->currentNav->location.hpr[i] = data->reset_location[i+3];
			 }

//			 data->playbackflag = false;
		}
		
	} else {

		Rewind->labelcolor(0);		
		RewindStep->labelcolor(0);	rs = PLAY;
//		Play->labelcolor(0);		PlayStop = PLAY;
		FFStep->labelcolor(0); 		fs = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
//		Pause->deactivate();
//		Record->activate();
	}
	
	PlayBackWin->redraw();
	rw = !rw;
}

// rewinds through the logfile slower then rewind
void rewindstep_callback (Fl_Button*, pbwr* data){
	
	if (rs) {
		RewindStep->labelcolor(60);
		Rewind->labelcolor(0);		
		Play->labelcolor(0);		PlayStop = PLAY;
		FFStep->labelcolor(0);
		FForward->labelcolor(0);
		Record->deactivate();
		Pause->activate();
		

	} else {
		RewindStep->labelcolor(0);
		FFStep->labelcolor(0);
		FForward->labelcolor(0);
		Record->activate();
		Pause->deactivate();
	}
	
	rw = fs = ff = PLAY;
	PlayBackWin->redraw();
	rs = !rs;
}

//  1)  if no log file is loaded then creates a new file and 
//		begins recording from the current location
//				or
//  2)	begins recording and appends to the end of the current
//		loaded log file
void record_callback(Fl_Button*, pbwr* data){

	if (PlayRecord){

		Record->labelcolor(1);
		FFStep->labelcolor(0);
		FForward->labelcolor(0);
		Pause->labelcolor(0);		PlayPause = PLAY;
		
		Play->deactivate();
		FForward->deactivate();
		FFStep->deactivate();
		Rewind->deactivate();
		RewindStep->deactivate();		
		
		//w dso code
		Text->value("Beginning recording...");

	} else {
	
		Record->labelcolor(0);
		FFStep->labelcolor(0);		
		FForward->labelcolor(0);

		Play->activate();
		FForward->activate();
		FFStep->activate();
		Rewind->activate();
		RewindStep->activate();	
		
		// w dso code
		Text->value("Stopped recording...");
	}
	
	fs = ff = PLAY;
	PlayRecord = !PlayRecord;
    data->recordflag = !data->recordflag;
    PlayBackWin->redraw();

	// w dso code    
    data->t.reset(data->prevtime,0);
}

// plays currently loaded log file
void play_callback(Fl_Button*, pbwr* data){

	static dpfNav* prevNav = NULL;
	float response;

	if (PlayStop){
		Play->labelcolor(60);
		RewindStep->labelcolor(0);	rs = PLAY;
		Rewind->labelcolor(0);		rw = PLAY;
		FFStep->labelcolor(0); 		fs = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
		Pause->activate();
		Record->deactivate();
		
		// r dso code
		Text->value("Starting playback...");
		data->t.reset(data->prevFileTime,0);
//////////
		// gets the current navigation and swaps it with pbwr NAV dso	
		prevNav = data->display->currentNav;

		response = prevNav->response();

		dtkMsg.add(DTKMSG_INFO,"dsow::postFrame():"
                              " switching navigation from \"%s%s%s\" to ",
                              dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end);
		
	    	dtkMsg.append("\"%s%s%s\".\n",dtkMsg.color.vil,
	                      data->getName(), dtkMsg.color.end);
		 
		
		if(prevNav) {
			if (prevNav != data){

				// move to where it left off
				data->location.xyz = prevNav->location.xyz;
				data->location.hpr = prevNav->location.hpr;
				data->scale = prevNav->scale;
				data->response(response);
				data->preFrame();
				
				// pivot is copied only if both are not NULL
				pfCoord *cp = prevNav->pivot() ;
				pfCoord *ncp = data->pivot() ;
			
				if (cp && ncp) {
					data->pivot(prevNav->pivot());
				      	data->dpfPivot(prevNav->dpfPivot());
				}
				
				// velocity is copied only both are not NULL
			pfCoord *vp = prevNav->velocity() ;
				pfCoord *nvp = data->velocity() ;
				
				if (vp && nvp)
					*nvp = *vp ;
				
				data->display->switchNav(data);
			}
		}
/////////
		//
	} else {
		Play->labelcolor(0);
		RewindStep->labelcolor(0);	rs = PLAY;
		Rewind->labelcolor(0);		rw = PLAY;
		FFStep->labelcolor(0); 		fs = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
		Pause->deactivate();
		Record->activate();
		
		// r dso code
		Text->value("Stopped playback...");
		data->prevFileTime = data->t.get();

		dtkMsg.add(DTKMSG_INFO, "dsow::postFrame():"
				" switching navigation from \"%s%s%s\" to ",
				dtkMsg.color.vil, data->getName(), dtkMsg.color.end);

		if(prevNav) {
			dtkMsg.append("\"%s%s%s\".\n", dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end);
		
			response = data->response();
			if(prevNav != data){
				
				prevNav->location.xyz = data->location.xyz;
				prevNav->location.hpr = data->location.hpr;
				prevNav->scale = data->scale;
				prevNav->response(response);
				prevNav->preFrame();

				pfCoord *cp = data->pivot();
				pfCoord *ncp = prevNav->pivot();

				if(cp && ncp) {
					prevNav->pivot(data->pivot());
					prevNav->dpfPivot(data->dpfPivot());
				}
				
				pfCoord *vp = data->velocity();
				pfCoord *nvp = prevNav->velocity();

				if(vp && nvp) * nvp = *vp;

				data->display->switchNav(prevNav);
			}	
		}
		
	}
	
	PlayStop = !PlayStop;
	data->playbackflag = !data->playbackflag;
	PlayBackWin->redraw();
}

// pauses current action and holds at current location
void pause_callback(Fl_Button* , pbwr* data){

	//PlayPause = data->playbackflag;
	
	if (PlayPause){
	
		Pause->labelcolor(1);
		Play->deactivate();
		Record->deactivate();
		FForward->deactivate();
		FFStep->deactivate();
		Rewind->deactivate();
		RewindStep->deactivate();
		
		// dso code
		Text->value("Paused playback...");
		data->prevFileTime = data->t.get();		
		data->playbackflag = false;
		
	} else {
	
		Pause->labelcolor(0);
		Play->activate();
//		if(PlayRecord) Record->activate();
		
		Rewind->activate();
		RewindStep->activate();		
		FForward->activate();
		FFStep->activate();
		
		
		// dso code
		Text->value("Resuming playback...");
		data->t.reset(data->prevFileTime,0);
		data->playbackflag = !data->playbackflag;
	}

	PlayPause = !PlayPause;
	PlayStop = !PlayStop;
	//data->playbackflag = !data->playbackflag;
	PlayBackWin->redraw();
}

// steps forward through currently loaded log file, but slower
// then FForward
void ffstep_callback (Fl_Button*, pbwr*){
	
	if (fs) {

		FFStep->labelcolor(60);
		Rewind->labelcolor(0);
		RewindStep->labelcolor(0);
		Record->deactivate();
		Play->labelcolor(0);
		Pause->activate();
		FForward->labelcolor(0);	

		
	} else {
	
		Rewind->labelcolor(0);
		RewindStep->labelcolor(0);
		Record->activate();
		Pause->deactivate();
		FFStep->labelcolor(0);
	}	
	
	rw = rs = ff = PLAY;
	PlayBackWin->redraw();
	fs = !fs;
}

// fastforwards through current loaded log file, faster than
// stepping throught it with FFStep
void fforward_callback (Fl_Button*, pbwr* data) {
	if (ff){
	
		FForward->labelcolor(1);		
		Rewind->labelcolor(0);		rw = PLAY;
		RewindStep->labelcolor(0);	rs = PLAY;
		Play->labelcolor(0);		PlayStop = PLAY;	
		FFStep->labelcolor(0); 		fs = PLAY;
//		Pause->activate();
//		Record->deactivate();
		StatusBar->value(statusBarMax);
		StatusBar->redraw();

		if(data->file != NULL) {
			fseek(data->file, 0L, SEEK_END);
			Text->value("Fast Forward to end of the file");
		}

	} else {

		FForward->labelcolor(0);		
		Rewind->labelcolor(0);		rw = PLAY;
		RewindStep->labelcolor(0);	rs = PLAY;
		Play->labelcolor(0);		PlayStop = PLAY;
		FFStep->labelcolor(0); 		fs = PLAY;
//		Pause->deactivate();
//		Record->activate();
	}
	
	PlayBackWin->redraw();
	ff = !ff;
}

// resets the buttons to original location and if a log file is
// loaded, reset to the head of the that file
void reset_callback(Fl_Button*, pbwr* data) {

///*	
	Rewind->activate();		Rewind->labelcolor(0);
	RewindStep->activate();	RewindStep->labelcolor(0);
	Record->activate();		Record->labelcolor(0);
	Play->activate();		Play->labelcolor(0);
	Pause->deactivate();	Pause->labelcolor(0);
	FFStep->activate();		FFStep->labelcolor(0);
	FForward->activate();	FForward->labelcolor(0);
	rw = rs = PlayRecord = PlayStop = PlayPause = fs = ff = PLAY;
//*/
	Play->activate();		Play->labelcolor(0);
	Pause->deactivate();	Pause->labelcolor(0);
	PlayRecord = PlayStop = !PLAY;
	
	// dso code
	Text->value("Stopping and resetting playback...");
	fflush(stdout);
  
	rewind(data->file);
	data->t.reset (data->resettime,0);
	data->prevFileTime = data->currFileTime = data->resettime;
	  
  	for (int i = 0; i < 3; i++) {
	    data->display->currentNav->location.xyz[i] = data->reset_location[i];
	    data->display->currentNav->location.hpr[i] = data->reset_location[i+3];
	}
 
  	data->playbackflag = false;
	PlayStop = true;
	StatusBar->value(0);
  	PlayBackWin->redraw();
}

// closes the gui but does not close down the entire program
void closegui_callback(Fl_Button*, pbwr* data) {
	PlayBackWin->hide();
	data->guiflag = true;
}

// closes down everthing
void quit_callback() {
	running = !running;
}

/*******************************************
		creates the gui
*******************************************/

Fl_Window* make_window(pbwr* data) {

 	 Fl_Window* w;
	 { Fl_Window* o = PlayBackWin = new Fl_Window(300, 150, "PlayBack");		//300, 120
		w = o;
		{ Fl_Button* o = Rewind = new Fl_Button(0, 25, 50, 30, "@<<");
			o->type(2);
			o->down_box(FL_DOWN_BOX);
     			o->labeltype(FL_SYMBOL_LABEL);
      			o->callback((Fl_Callback*)rewind_callback, data);
			o->when(FL_WHEN_CHANGED);
    		}
		{ Fl_Button* o = RewindStep = new Fl_Button(50, 25, 45, 30, "@|<");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      	//		o->callback((Fl_Callback*)rewindstep_callback, data);
    		}
    		{ Fl_Button* o = Record = new Fl_Button(95, 25, 30, 30, "@circle");
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->labelcolor(0);
//     			o->callback((Fl_Callback*)record_callback, data);
    		}
    		{ Fl_Button* o = Play = new Fl_Button(125, 25, 50, 30, "@>");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->labelcolor(0);
      			o->callback((Fl_Callback*)play_callback, data);
    		}
    		{ Fl_Button* o = Pause = new Fl_Button(175, 25, 30, 30, "@||");
	      		o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->deactivate();
      			o->callback((Fl_Callback*)pause_callback, data);
    		}
    		{ Fl_Button* o = FFStep = new Fl_Button(205, 25, 45, 30, "@>|");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      	//		o->callback((Fl_Callback*)ffstep_callback);
    		}
    		{ Fl_Button* o = FForward = new Fl_Button(250, 25, 50, 30, "@>>");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->callback((Fl_Callback*)fforward_callback,data);
			o->when(FL_WHEN_CHANGED);
			o->type(2);
    		}
    		{ Fl_Menu_Bar* o = new Fl_Menu_Bar(0, 0, 300, 25);
      			o->menu(menu_);
      			o->activate();
    		}
		{ Fl_Output* o = Text = new Fl_Output(5, 60, 290, 25);
      			o->labeltype(FL_NO_LABEL);
      			o->value("Ready");
    		}
		{ Fl_Slider* o = StatusBar = new Fl_Slider(5, 90, 290, 20);
			o->type(3);
		        o->selection_color(81);
			//o->maximum(statusBarMax);
			o->maximum(47);
			o->step(1);
      			o->value(0);
    		}	  	
    		{ Fl_Button* o = Reset = new Fl_Button(5, 120, 95, 25, "Reset");		// 5,90,95,25	
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_EMBOSSED_LABEL);
      			o->callback((Fl_Callback*)reset_callback, data);
    		}
   	 	{ Fl_Button* o = CloseGUI = new Fl_Button(105, 120, 90, 25, "Close GUI");	//105,90,90,25
      			o->labeltype(FL_EMBOSSED_LABEL);
      			o->callback((Fl_Callback*)closegui_callback, data);
    		}
    		{ Fl_Return_Button* o = Quit = new Fl_Return_Button(200, 120, 95, 25, "Quit");	//200,90,95,25
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_EMBOSSED_LABEL);
      			o->callback((Fl_Callback*)quit_callback);
     	 		o->deactivate();
    		}

		o->end();
  	}
 
	w->show();
  
//  	menu_[1].user_data((void*)data);
//  	menu_[5].user_data((void*)data);
  
  	return w;
}



/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new pbwr(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
