#include <dpf.h>
#include <math.h>
#include <stdlib.h>
#include "playbackReadWriteNav.h"

#define PLAY true
#define LOWER log10(.5)
#define UPPER log10(30)

#define File (menu_+0)
#define New (menu_+1)
#define Open (menu_+2)
#define Save (menu_+3)
#define SaveAs (menu_+4)
#define Close (menu_+5)
#define Exit (menu_+6)

Fl_Menu_Item menu_[] = {
	{"File", 0,  0, 0, 64, 5, 0, 14, 0},
	{"New", 0x4006e,  (Fl_Callback*)new_callback, 0, 0, 5, 0, 14, 0},
	{"Open", 0x4006f,  (Fl_Callback*)open_callback, 0, 0, 5, 0, 14, 0},
	{"Save", 0x40073, (Fl_Callback*)save_callback, 0, 0, 5, 0, 14, 0},
	{"Save As", 0x40061,  (Fl_Callback*)save_as_callback, 0, 0, 5, 0, 14, 0},
	{"Close", 0x40063,  (Fl_Callback*)close_callback, 0, 0, 5, 0, 14, 0},
	{0},
	{0}
};


Fl_Window *guiWin=(Fl_Window *)0;
Fl_Button *Play=(Fl_Button *)0;
Fl_Value_Input *FrameLoc=(Fl_Value_Input *)0;
Fl_Output *Text=(Fl_Output *)0;
Fl_Slider *StatusBar=(Fl_Slider *)0;

static Fl_Button *Rewind=(Fl_Button *)0;
static Fl_Button *Pause=(Fl_Button *)0;
static Fl_Button *FForward=(Fl_Button *)0;
static Fl_Button *CloseGUI=(Fl_Button *)0;
static Fl_Output *FileDisplay=(Fl_Output *)0;
static Fl_Slider *SpeedBar=(Fl_Slider *)0;
static Fl_Button *Reset=(Fl_Button *)0;
static Fl_Button *Record=(Fl_Button *)0;
static Fl_Value_Input * SpeedDisplay=(Fl_Value_Input *)0;

static bool PlayPause = PLAY;
static bool PlayRecord = PLAY;
static bool rs = PLAY;
static bool rw = PLAY;
static bool fs = PLAY;
static bool ff = PLAY;

static dpfNav *prevNav = NULL;
static float response;
bool PlayStop = PLAY;
int statusBarMax;


/*******************************************
		Menu Callbacks
*******************************************/

// creates a new file for recording
void new_callback(Fl_Menu_*, playbackReadWriteNav* ) {

	fl_message("This will eventually allow users to create new log files.");
}		


void open_callback(Fl_Menu_*, playbackReadWriteNav* data) {
	fl_message ("This will eventually allow users to open different log files.");
}

void save_callback(Fl_Menu_*, playbackReadWriteNav* data) {
	fl_message("This will eventually allow users to save changes to currently open log files");
}

void save_as_callback(Fl_Menu_*, playbackReadWriteNav* data) {

	fl_message("This will eventually allow users to save changes to currently open log files");
}

void close_callback(Fl_Menu_*, playbackReadWriteNav*) {

	fl_message("This will eventually allow users to close currently open log.");
}

/*******************************************
		Button Callbacks
*******************************************/

// rewinds through the log file faster then rewindstep
void rewind_callback (Fl_Button*, playbackReadWriteNav* data) {

	if (rw){
		Rewind->labelcolor(1);		
		FForward->labelcolor(0);	ff = PLAY;
		StatusBar->value(0);
		FrameLoc->value(0);
		data->currCount = 0;
		
		Text->value("Rewind to the beginning");
		StatusBar->value(0);
	        data->t.reset (data->resettime,0);
	        data->prevFileTime = data->currFileTime = data->resettime;

		for (int i = 0; i < 3; i++) {
	        	data->display->currentNav->location.xyz[i] = data->reset_location[i];
			data->display->currentNav->location.hpr[i] = data->reset_location[i+3];
		}
		
	} else {

		Rewind->labelcolor(0);		
		FForward->labelcolor(0);	ff = PLAY;
	}
		
	guiWin->redraw();
	rw = !rw;
	
}

//  1)  if no log file is loaded then creates a new file and 
//		begins recording from the current location
//				or
//  2)	begins recording and appends to the end of the current
//		loaded log file
void record_callback(Fl_Button*, playbackReadWriteNav* data){

/*	if (PlayRecord){

		Record->labelcolor(1);
		FForward->labelcolor(0);
		Pause->labelcolor(0);		PlayPause = PLAY;
		
		Play->deactivate();
		FForward->deactivate();
		Rewind->deactivate();
		
		//w dso code
		Text->value("Beginning recording...");

	} else {
	
		Record->labelcolor(0);
		FForward->labelcolor(0);

		Play->activate();
		FForward->activate();
		Rewind->activate();
		
		// w dso code
		Text->value("Stopped recording...");
	}
	
	fs = ff = PLAY;
	PlayRecord = !PlayRecord;
	data->recordflag = !data->recordflag;
    	guiWin->redraw();

	// w dso code    
    	data->t.reset(data->prevtime,0);
*/
	fl_message ("This will eventually allow users to record to the current playback "\
		    "file or create a new file if none is open.");
}

// plays currently loaded log file
void play_callback(Fl_Button*, playbackReadWriteNav* data){

	if (PlayStop){
		
		Play->labelcolor(60);
		Rewind->labelcolor(0);		rw = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
		Pause->activate();
		Record->deactivate();
		
		// r dso code
		Text->value("Starting playback...");
		StatusBar->value(data->currCount);
		FrameLoc->value(data->currCount);
		
		int c = data->currCount;
		if(c!=0) c--;
		
		//data->t.reset(data->fileData[c]->t * speedFactor / pow(10,SpeedBar->value()),0);
		//data->t.reset(data->prevFileTime/pow(10, speedFactor),0);
		data->t.reset(data->prevtime,0);
		//data->t.reset(data->prevFileTime * speedFactor / pow(10,SpeedBar->value()),0);
		
		
		// gets the current navigation and swaps it with playbackReadWriteNav	
		prevNav = data->display->currentNav;
		response = prevNav->response();

		dtkMsg.add(DTKMSG_INFO, "dsow::postFrame(): switching navigation from " 
					"\"%s%s%s\" to \"%s%s%s\".\n", 
					dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end,
				        dtkMsg.color.vil, data->getName(), dtkMsg.color.end);
		 
		if(prevNav) {
			if (prevNav != data){

				// move to where it left off
				data->location.xyz = prevNav->location.xyz;
				data->location.hpr = prevNav->location.hpr;
				data->scale = prevNav->scale;
				data->response(response);
				data->preFrame();
				
				// pivot is copied only if both are not NULL
				pfCoord *cp = prevNav->pivot();
				pfCoord *ncp = data->pivot();
			
				if (cp && ncp) {
					data->pivot(prevNav->pivot());
				      	data->dpfPivot(prevNav->dpfPivot());
				}
				
				// velocity is copied only both are not NULL
				pfCoord *vp = prevNav->velocity();
				pfCoord *nvp = data->velocity();
				
				if (vp && nvp)
					*nvp = *vp;
				
				data->display->switchNav(data);
			}
		}
	
	} else {

		// sets the buttons so that they are the correct color and 
		// have the correct values
		Play->labelcolor(0);
		Rewind->labelcolor(0);		rw = PLAY;
		FForward->labelcolor(0);	ff = PLAY;
		Pause->deactivate();
		Record->activate();
		Text->value("Stopped playback...");
		data->prevFileTime = data->t.get();
		data->currCount = (int)StatusBar->value();


		// swap back the previous NAV DSO 
		dtkMsg.add(DTKMSG_INFO, "dsow::postFrame():"
				" switching navigation from \"%s%s%s\" to ",
				dtkMsg.color.vil, data->getName(), dtkMsg.color.end);

		if(prevNav) {
			dtkMsg.append("\"%s%s%s\".\n", dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end);
		
			response = data->response();
			if(prevNav != data){
				
				prevNav->location.xyz = data->location.xyz;
				prevNav->location.hpr = data->location.hpr;
				prevNav->scale = data->scale;
				prevNav->response(response);
				prevNav->preFrame();

				pfCoord *cp = data->pivot();
				pfCoord *ncp = prevNav->pivot();

				if(cp && ncp) {
					prevNav->pivot(data->pivot());
					prevNav->dpfPivot(data->dpfPivot());
				}
				
				pfCoord *vp = data->velocity();
				pfCoord *nvp = prevNav->velocity();

				if(vp && nvp) * nvp = *vp;

				data->display->switchNav(prevNav);
			}	
		}
	}
	
	PlayStop = !PlayStop;
	data->playbackflag = !data->playbackflag;
	guiWin->redraw();
}

// pauses current action and holds at current location
void pause_callback(Fl_Button* , playbackReadWriteNav* data){

	
	if (PlayPause){
	
		Pause->labelcolor(1);
		Play->deactivate();
		Record->deactivate();
		FForward->deactivate();
		Rewind->deactivate();
		
		// dso code
		Text->value("Paused playback...");
		data->prevFileTime = data->t.get();		
		data->playbackflag = false;

///////////////////////////////
		// swap back the previous NAV DSO
		dtkMsg.add(DTKMSG_INFO, "dsow::postFrame():"
					" switching navigation from \"%s%s%s\" to ",
					dtkMsg.color.vil, data->getName(), dtkMsg.color.end);
	
		if(prevNav) {

			dtkMsg.append("\"%s%s%s\".\n", dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end);
			response = data->response();
			
			if(prevNav != data){
				prevNav->location.xyz = data->location.xyz;
				prevNav->location.hpr = data->location.hpr;
				prevNav->scale = data->scale;
				prevNav->response(response);
				prevNav->preFrame();
		
				pfCoord *cp = data->pivot();
				pfCoord *ncp = prevNav->pivot();
	
				if(cp && ncp) {
					prevNav->pivot(data->pivot());
					prevNav->dpfPivot(data->dpfPivot());
				}
		
				pfCoord *vp = data->velocity();
				pfCoord *nvp = prevNav->velocity();
		
				if(vp && nvp) * nvp = *vp;
				data->display->switchNav(prevNav);
			}
		}

///////////////////////////////		

	} else {
	
		Pause->labelcolor(0);
		Play->activate();
		Rewind->activate();
		FForward->activate();
		
		
		// dso code
		Text->value("Resuming playback...");
		data->t.reset(data->prevFileTime,0);
		data->playbackflag = !data->playbackflag;

//////////
	        // gets the current navigation and swaps it with playbackReadWriteNav
		prevNav = data->display->currentNav;
		response = prevNav->response();
		
		dtkMsg.add(DTKMSG_INFO, "dsow::postFrame(): switching navigation from " \
					"\"%s%s%s\" to \"%s%s%s\".\n",
					dtkMsg.color.vil, prevNav->getName(), dtkMsg.color.end,
					dtkMsg.color.vil, data->getName(), dtkMsg.color.end);

		if(prevNav) {
			if (prevNav != data) {
				
				// move to where it left off
				 data->location.xyz = prevNav->location.xyz;
				 data->location.hpr = prevNav->location.hpr;
				 data->scale = prevNav->scale;
				 data->response(response);
				 
				 data->preFrame();
				 
				 // pivot is copied only if both are not NULL
				 
				 pfCoord *cp = prevNav->pivot();
				 pfCoord *ncp = data->pivot();
		 		 
				 if (cp && ncp) {
					 
					 data->pivot(prevNav->pivot());
					 data->dpfPivot(prevNav->dpfPivot());   
				 }

				 // velocity is copied only both are not NULL
				 
				 pfCoord *vp = prevNav->velocity();
				 pfCoord *nvp = data->velocity();
				 
				 if (vp && nvp) *nvp = *vp;
				 
				 data->display->switchNav(data);
			}
		}
/////////
		
	}

	PlayPause = !PlayPause;
	PlayStop = !PlayStop;
	guiWin->redraw();
}

// fastforwards to the end of the currently open log file
void fforward_callback (Fl_Button*, playbackReadWriteNav* data) {
	if (ff){
	
		FForward->labelcolor(1);		
		Rewind->labelcolor(0);		rw = PLAY;
		Play->labelcolor(0);
		StatusBar->value(statusBarMax);
		FrameLoc->value(statusBarMax);
		data->currCount = statusBarMax;
			
		Text->value("Fast Forward to the end");
		StatusBar->value(statusBarMax);
		//data->t.reset(data->t.get(),0);
		data->t.reset(data->t.get() * speedFactor / SpeedDisplay->value(),0);
		data->prevtime = data->prevFileTime = data->currFileTime = data->fileData[statusBarMax-1]->t;

                for (int i = 0; i < 3; i++) {
                	data->display->currentNav->location.xyz[i] = data->fileData[statusBarMax]->xyzhpr[i];
			data->display->currentNav->location.hpr[i] = data->fileData[statusBarMax]->xyzhpr[i+3];
		}
	} else {

		FForward->labelcolor(0);		
		Rewind->labelcolor(0);		rw = PLAY;
	}

	Play->labelcolor(0);
	Play->redraw();
	PlayStop = Play;
	data->playbackflag = false;
	ff = !ff;
}

// updates the StatusBar with values input by the user
void frameloc_callback(Fl_Value_Input*, playbackReadWriteNav* data) {

	if(FrameLoc->value() > statusBarMax) FrameLoc->value(statusBarMax);
	else if (FrameLoc->value() < 0) FrameLoc->value(0);

	StatusBar->value(FrameLoc->value());
	statusbar_callback(NULL, data);
}

void statusbar_callback(Fl_Slider*, playbackReadWriteNav* data) {

	int Index,prevIndex,i;

	data->currCount = (int) StatusBar->value();

	FrameLoc->value(data->currCount);

	Index = data->currCount;

	if(Index == 0) prevIndex = 0;
	else prevIndex = Index-1;
	
	for (i = 0; i < 6; i++) {
		data->prevFileLoc[i] = data->fileData[prevIndex]->xyzhpr[i];
		data->currFileLoc[i] = data->fileData[Index]->xyzhpr[i];
	}
	printf("index = %i\n", Index);

	data->prevtime = data->fileData[prevIndex]->t/speedFactor;
	data->prevFileTime = data->fileData[prevIndex]->t;
	data->currFileTime = data->fileData[Index]->t;
	
	if(data->currCount == statusBarMax) Text->value("end of playback file");
	else Text->value("Ready");
	
	data->t.reset(data->prevtime, 0);
}

//
void speeddisplay_callback(Fl_Value_Input*, playbackReadWriteNav* data) {
	
	float newSpeed = SpeedDisplay->value();
	long double time = data->t.get();

	data->t.reset(time * speedFactor / newSpeed);
	
	speedFactor = newSpeed;
	SpeedBar->value(log10(newSpeed));
}

// changing the speedFactor based on the value of the slider SpeedBar
void speedbar_callback(Fl_Slider*, playbackReadWriteNav* data) {

	float newSpeed = SpeedBar->value();
	long double time = data->t.get();
	
	data->t.reset(time * speedFactor / pow(10,newSpeed), 0);

	speedFactor = pow(10,newSpeed);	
	SpeedDisplay->value(speedFactor);
}

// resets the buttons to original location and if a log file is
// loaded, reset to the head of the that file
void reset_callback(Fl_Button*, playbackReadWriteNav* data) {

	Rewind->activate();		Rewind->labelcolor(0);
	Record->activate();		Record->labelcolor(0);
	Play->activate();		Play->labelcolor(0);
	Pause->deactivate();		Pause->labelcolor(0);
	FForward->activate();		FForward->labelcolor(0);
	
	rw = rs = PlayRecord = PlayStop = PlayPause = fs = ff = PLAY;
	
	Play->activate();		Play->labelcolor(0);
	Pause->deactivate();		Pause->labelcolor(0);
	PlayRecord = PlayStop = !PLAY;
	
	// dso code
	Text->value("Stopping and resetting playback...");
	data->currCount = 0;
	StatusBar->value(0);
	FrameLoc->value(0.0000);
	fflush(stdout);
  
	data->t.reset (data->resettime,0);
	data->prevFileTime = data->currFileTime = data->resettime;
	  
  	for (int i = 0; i < 3; i++) {
	    data->display->currentNav->location.xyz[i] = data->reset_location[i];
	    data->display->currentNav->location.hpr[i] = data->reset_location[i+3];
	}
 
  	data->playbackflag = false;
	PlayStop = true;
	StatusBar->value(0);
	FrameLoc->value(0.0000);
  	guiWin->redraw();
}

// closes the gui but does not close down the entire program
void closegui_callback(Fl_Button*, playbackReadWriteNav* data) {
	guiWin->hide();
	data->guiflag = true;
}

/*******************************************
		creates the gui
*******************************************/

// does all the gui creation and setting callback associated with 
// different widgets 
Fl_Window* make_window(playbackReadWriteNav* data) {

 	 Fl_Window* w;
	 { Fl_Window* o = guiWin = new Fl_Window(420, 167, "PlayBack");
		w = o;
    		{ Fl_Menu_Bar* o = new Fl_Menu_Bar(0, 0, 420, 25);
      			o->menu(menu_);
      			o->activate();
    		}
    		{ Fl_Value_Input* o = FrameLoc = new Fl_Value_Input(55, 30, 50, 20,"Frame: ");
			o->labeltype(FL_EMBOSSED_LABEL);
			o->labelcolor(59);
			o->textcolor(59);
			o->color(49);
      			o->value(0);
			o->step(1);
			o->minimum(0);
			o->maximum(statusBarMax);
			o->callback((Fl_Callback*)frameloc_callback, data);
		}
		{ Fl_Value_Output* o = new Fl_Value_Output(115, 30, 45, 20, "/");
			o->box(FL_FLAT_BOX);
			o->textcolor(59);
			o->labelcolor(59);
			o->value(statusBarMax);
		}			
		{ Fl_Slider* o = StatusBar = new Fl_Slider(165, 30, 250, 20);
			o->type(5);
		        o->selection_color(59);	
		        o->maximum(statusBarMax);
			o->step(1);
      			o->value(0);
      			o->callback((Fl_Callback*) statusbar_callback, data);
      			o->when(FL_WHEN_CHANGED);
    		}
		{ Fl_Value_Input* o = SpeedDisplay = new Fl_Value_Input(55, 55, 50, 20, "Speed: ");
			o->labeltype(FL_EMBOSSED_LABEL);
			o->labelcolor(89);
			o->textcolor(89);
			o->color(49);
			o->value(1);
			o->step(0.001);
			o->minimum(.5);
			o->maximum(30);
			o->callback((Fl_Callback*) speeddisplay_callback,data);
			o->when(FL_WHEN_ENTER_KEY);
		}
		{ Fl_Slider* o = SpeedBar = new Fl_Slider(110, 55, 305, 20);
			o->type(5);
			o->selection_color(89);
			o->step(0.0001);
			o->value(0);
			o->minimum(LOWER);
			o->maximum(UPPER);
			o->align(FL_ALIGN_LEFT|FL_ALIGN_BOTTOM);
			o->callback((Fl_Callback*) speedbar_callback,data);
		}
		{ Fl_Box* o = new Fl_Box(5,80,410,50);
			o->box(FL_DOWN_BOX);
			o->color(41);
			o->labeltype(FL_NO_LABEL);
		}
		{ Fl_Button* o = Rewind = new Fl_Button(15, 90, 50, 30, "@<<");
			o->type(2);
			o->down_box(FL_DOWN_BOX);
     			o->labeltype(FL_SYMBOL_LABEL);
      			o->callback((Fl_Callback*)rewind_callback, data);
			o->when(FL_WHEN_CHANGED);
    		}
    		{ Fl_Button* o = Record = new Fl_Button(65, 90, 30, 30, "@circle");
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->labelcolor(0);
     			o->callback((Fl_Callback*)record_callback, data);
    		}
    		{ Fl_Button* o = Play = new Fl_Button(95, 90, 45, 30, "@>");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->labelcolor(0);
      			o->callback((Fl_Callback*)play_callback, data);
    		}
    		{ Fl_Button* o = Pause = new Fl_Button(140, 90, 30, 30, "@||");
	      		o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->deactivate();
      			o->callback((Fl_Callback*)pause_callback, data);
    		}
    		{ Fl_Button* o = FForward = new Fl_Button(170, 90, 50, 30, "@>>");
			o->type(2);
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_SYMBOL_LABEL);
      			o->callback((Fl_Callback*)fforward_callback,data);
			o->when(FL_WHEN_CHANGED);
    		}
		{ Fl_Output* o = Text = new Fl_Output(225, 90, 180, 25);	
      			o->labeltype(FL_NO_LABEL);
      			o->value("Ready");
    		}
		{ Fl_Output* o = FileDisplay = new Fl_Output(5, 135, 185, 25);
		        o->labeltype(FL_NO_LABEL);
			o->value(data->logfile);
		//	o->value("/tmp/logFileName");
		}
		{ Fl_Button* o = Reset = new Fl_Button(195, 135, 110, 25, "Reset");
      			o->down_box(FL_DOWN_BOX);
      			o->labeltype(FL_EMBOSSED_LABEL);
      			o->callback((Fl_Callback*)reset_callback, data);
    		}
   	 	{ Fl_Button* o = CloseGUI = new Fl_Button(310, 135, 105 , 25, "Close GUI");
      			o->labeltype(FL_EMBOSSED_LABEL);
      			o->callback((Fl_Callback*)closegui_callback, data);
    		}

		o->end();
  	}
 
	w->show();
  
//  	menu_[1].user_data((void*)data);
//  	menu_[5].user_data((void*)data);
  
  	return w;
}
