// Eric Tester, etester@vt.edu, wrote the original version of this plays
// back the file created by playbackWrite
//
// the file name is specified by the environment variable
// DPF_PLAYBACK_FILE, if set, otherwise the file used is
// "/tmp/playback-$LOGNAME.log"
#include <stdlib.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include "playbackReadWriteNav.h"

#define PLAY true

extern Fl_Window *guiWin;
extern Fl_Button *Play;
extern Fl_Value_Input *FrameLoc;
extern Fl_Output* Text;
extern Fl_Slider* StatusBar;

extern bool PlayStop;		// = PLAY
extern int statusBarMax;
double speedFactor;
	 
#define DESCRIPTION 	"\nplaybackReadWriteNav Log file is: %s\n" \
			"  %splaybackReadWriteNav commands:  %s\n" \
	      	 	"  \'%so%s\'  Toggle- toggle on/off the graphical interface\n"
	     
/*******************************************
		Class Methods
*******************************************/

// constructor
playbackReadWriteNav::playbackReadWriteNav (dpf* in_app) :  dpfNav (in_app,"playbackReadWriteNav") {
  
	app = in_app;
  	keyboard = NULL;
  	dequeuer = NULL;
  	display = NULL;
  	guiWin = NULL;
  	fileData = NULL;
	prevtime = 0;
	lastModified = 0;

	// gets the logfile from the environment variable, if it exists
  	logfile = getenv("DPF_PLAYBACK_FILE");
 	
  	playbackflag = recordflag = guiflag = false;
 	
	// if not, then create one	
  	if(!logfile) {
		char *format = "/tmp/playback-%s";
   		char *logname = getenv("LOGNAME");
      	
   		if(!logname) logname = "nonknown_user";
	     	logfile = (char *) dtk_malloc(strlen(format) + strlen(logname) + 1);
      		sprintf(logfile, format, logname);

      		logname = NULL;
	      	format = NULL;
	}

	// prints out the instructions
  	setDescription(DESCRIPTION,logfile, dtkMsg.color.rvil, dtkMsg.color.end,
		dtkMsg.color.grn, dtkMsg.color.end,
		dtkMsg.color.red, dtkMsg.color.end,
		dtkMsg.color.grn, dtkMsg.color.end);
		
	// attempts to open the log file	
  	if( (file = fopen(logfile,"a+")) == NULL) {
   		dtkMsg.add(DTKMSG_ERROR,1,
			"playbackReadWriteNav::playbackReadWriteNav() failed to open/create logfile \"%s\".\n",
		 	logfile);
      		return;
	}

////////// cycles through the file and allocates memory to hold all the elements	

	int maxChunk = 5000;
	totalCount = 0;
	speedFactor = 1;
	
	fileData = (inFileData**) dtk_malloc ((maxChunk) * sizeof(inFileData *));
	
	if(!fileData) {
		printf("Error: could not allocate enough memory\n");
		free(fileData);
		exit(-1);
	}

  int c;
  for (c=0; c<maxChunk; c++)
    fileData[c] = NULL;

  //  Get the modified timestamp of the playback file
  stat(logfile, &fileStat);
  lastModified = fileStat.st_mtime;

  //  Load the playback file
  reloadFile();

  	validate();
}

//  Reload the playback file
void playbackReadWriteNav::reloadFile() {
  printf ("loading file\n");

	int chunkCount = 0;
	int maxChunk = 5000;
  totalCount = 0;
 	if( (file = fopen(logfile,"a+")) == NULL) {
    printf ("file is null\n");
  }
	// rewind to beginning of log file	
	rewind(file);
	// loop through the file and adds elements to the list
	for(;;){

    if (fileData[totalCount] == NULL)
      fileData[totalCount] = new inFileData;
	
		// gets the info from the file	
		if(7 == (fscanf (file, "%Lg %f %f %f %f %f %f", &fileData[totalCount]->t,
				&fileData[totalCount]->xyzhpr[0], &fileData[totalCount]->xyzhpr[1], 
				&fileData[totalCount]->xyzhpr[2], &fileData[totalCount]->xyzhpr[3], 
				&fileData[totalCount]->xyzhpr[4], &fileData[totalCount]->xyzhpr[5])))
		{
			// reallocates more memory for data larger than maxChunk 	
			if(chunkCount++ == maxChunk-1) {
			
				fileData = (inFileData **) dtk_realloc(fileData,
						(totalCount + maxChunk + 1) * sizeof(inFileData *));

				chunkCount = 0;
			}
			
			totalCount++;

		} else break;
	}

	totalCount--;
	statusBarMax = totalCount;
  if (StatusBar)
    StatusBar->maximum(statusBarMax);
	currCount = 0;
	
/////// initialize initial location and time;
	currFileTime = fileData[0]->t;
	
	int i;
	for (i=0; i < 6; i++) {
    printf ("reset[%i] is %f\n", i, fileData[0]->xyzhpr[i]);
		reset_location[i]= fileData[0]->xyzhpr[i];
  }
//////
  
	// initialize previous time and location		
	for (i = 0; i < 3; i++) {
		location.xyz[i]  = currFileLoc[i]   = prevFileLoc[i]   = reset_location[i];
    		location.hpr[i]  = currFileLoc[i+3] = prevFileLoc[i+3] = reset_location[i+3];
  	}
	
  	resettime = prevFileTime = currFileTime;
  
  	t.reset(currFileTime,0);

  printf ("done loading file\n");
}

// destructor
playbackReadWriteNav::~playbackReadWriteNav (void) {
  
  	if (dequeuer) {
    		delete dequeuer;
	    	dequeuer = NULL;
  	}
  
  	if (file) {
    		fclose (file);
	    	file = NULL;
  	}

	int i;
	for ( i=0; i < totalCount; i++){
		if(fileData[i]) {
			delete fileData[i];
			fileData[i] = NULL;
		}
	} 	

	if(fileData) free(fileData);
	
  	if (guiWin) {
  		delete guiWin;
  		guiWin = NULL;
  	}
}

int playbackReadWriteNav::postConfig (void) {

  	dtkMsg.add(DTKMSG_NOTICE,getDescription());

  	dequeuer = new dtkDequeuer(app->record());
  	if(!dequeuer || dequeuer->isInvalid())
    	return DTK_CALLBACK_ERROR;
  
  	display = app->display();
  	if(!display) return DTK_CALLBACK_ERROR;
  
  	keyboard = static_cast<dtkInXKeyboard *>(app->get("xkeyboard", DTKINXKEYBOARD_TYPE));
  	if(!keyboard) return DTK_CALLBACK_ERROR;
  	
  	guiWin = make_window(this);
  	if(!guiWin) { 	
  		dtkMsg.add(DTKMSG_NOTICE,"%s Sorry, unable to create grapical interface.\n",
     				 getName());
		return DTK_CALLBACK_ERROR;
  	}
  
  	//////////////// get X key codes from key syms ///////////////////////
  
  	if( !(xkc_o = keyboard->getKeyCode(XK_o)))
		return DTK_CALLBACK_ERROR;
  
  	keyboard->queue();
  
	
  	return dpfNav::postConfig();
}


int playbackReadWriteNav::postFrame (void) {

  	dtkRecord_event* e;
  	dpfNav* currentNav = display->currentNav;
  
  	while((e = dequeuer->getNextEvent(keyboard))) {
    
		if(currentNav) {
 			KeyCode key;
      	
      			// if key is pressed
    			if((keyboard->read(&key,e))) {
	  
	  			if (key == xkc_o) {
	    	
					if(guiflag)	guiWin->show();	
					else 		guiWin->hide();
					guiflag = !guiflag;
				}
			}
		
			// records to the log file
			if(recordflag) {
	
				static int rflag = 1;	     		// flag to record info
		  		//static int doubleflag = 0; 		// flag to test for duplicates
			  	pfCoord location;
	  
	  			// get info from nav coord.
			  	location.xyz = currentNav->location.xyz;
		  		location.hpr = currentNav->location.hpr;
	  
			 	// checks if previous coord is the same as current coord
			  	if ( (prevCoord.xyz == location.xyz) &&
	    		    	     (prevCoord.hpr == location.hpr) ) {
	    
					// for duplicates				  				  
	    			  	//if (!doubleflag) doubleflag = 1;
				      	rflag = 0;
	    
	    			} //else if (doubleflag) doubleflag = 0;
			  
			  	// if true print value and to file
				if (rflag) {  
	    				fprintf(file, "%.25Lg %.7g %.7g %.7g %.7g %.7g %.7g\n", prevtime = t.get(),
				    		location.xyz[0], location.xyz[1], location.xyz[2], 
		       				location.hpr[0], location.hpr[1], location.hpr[2]);
	    
		    			prevCoord = location;
				}

		  		rflag = 1;
			}
    		}
	}
	
	// reads from the log file
  	if(playbackflag) {
  	
		int i;
		long double time_now = t.get();
		currCount = currCount;

		while(time_now >= (currFileTime / speedFactor)) {

  			for (i = 0; i < 3; i++) {
    				prevFileLoc[i]   = currFileLoc[i];
    				prevFileLoc[i+3] = currFileLoc[i+3];
    			}	

	  		prevFileTime = currFileTime;
				  	
			// reads in data from array, if at end of array, print end of file
			if (currCount < totalCount){
				currFileTime = fileData[currCount]->t;
				for (i = 0; i < 6; i++)
					currFileLoc[i] = fileData[currCount]->xyzhpr[i];
			} else {
				Text->value("end of playback file");
				StatusBar->value(statusBarMax);
      				Play->labelcolor(0);
      				PlayStop = PLAY;      				
	      			playbackflag = 0;
	      			guiWin->redraw();
    				return DTK_CONTINUE;			
			}
		
			// increment current index in fileData	
			currCount++;
		}
				
		// linear interpolates hpr data
		for (i = 0; i< 3; i++) {	
			
			//  If the difference between consecutive angles is more than 180, 
      			//  add 360 to the lower angle
      			if (fabs (currFileLoc[i+3] - prevFileLoc[i+3]) > 180) {
        		
        			if (currFileLoc[i+3] < prevFileLoc[i+3]) 
        				currFileLoc[i+3] += 360;
        			else
					prevFileLoc[i+3] += 360;
      			}

			location.hpr[i] = prevFileLoc[i+3] + (time_now - (prevFileTime / speedFactor)) *
				(currFileLoc[i+3] - prevFileLoc[i+3]) / 
				((currFileTime / speedFactor) - (prevFileTime / speedFactor));  
			
    		}
		
		// linear interpolates xyz data		
	 	for (i = 0; i < 3; i++) {
			
			location.xyz[i] = prevFileLoc[i] + ((time_now )- (prevFileTime / speedFactor)) *
				(currFileLoc[i] - prevFileLoc[i]) / 
				((currFileTime / speedFactor) - (prevFileTime / speedFactor));
				
		}

		// puts the values into the widgets	
		StatusBar->value(currCount);
		FrameLoc->value(currCount);
   	} 
   	
   	Fl::check();
  	
    //  Check to see if the playback file has been modified.
    //  If so, reload it
    stat(logfile, &fileStat);
    if (fileStat.st_mtime > lastModified) {
      printf ("logfile is %s, mtime is %i\n", logfile, fileStat.st_mtime);
      lastModified = fileStat.st_mtime;
      reloadFile();
    }

   	return DTK_CONTINUE;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new playbackReadWriteNav(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
