#ifndef PLAYBACKREADWRITENAV_H
#define PLAYBACKREADWRITENAV_H

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Slider.h>
#include <FL/fl_file_chooser.h>
#include <Fl/fl_message.h>
#include <FL/Fl_Value_Input.h>
#include <FL/Fl_Value_Output.h>
#include <Fl/Fl_Box.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

extern double speedFactor;
struct inFileData{
        long double t;
        float   xyzhpr[6];
};

class playbackReadWriteNav : public dpfNav {

        public:
                playbackReadWriteNav (dpf* in_app);
                ~playbackReadWriteNav (void);
                int postConfig (void);
               	int postFrame (void);
	private:
     		FILE* file;
		char* logfile;
		inFileData** fileData;
		int totalCount;
		int timeCount;
		int currCount;
		
		dpf* app;
		dpfDisplay* display;
		dtkDequeuer* dequeuer;
		dtkInXKeyboard *keyboard;
		KeyCode xkc_o;
		
                float reset_location[6];
                float prevFileLoc[6];
		float currFileLoc[6];
		dtkTime t;
		long double currFileTime;
		long double prevFileTime;
		long double resettime;
		long double prevtime;
		pfCoord prevCoord;
		bool playbackflag;
		bool recordflag;
		bool guiflag;

    //  Stat record for the playback file
    struct stat fileStat;

    //  Modified time of the playback file when it was last read in
    time_t lastModified;

    //  Reload the playback file if it was modified
    void reloadFile();

		// button callbacks
		friend void rewind_callback(Fl_Button*, playbackReadWriteNav*);
		friend void record_callback(Fl_Button*, playbackReadWriteNav*);
		friend void play_callback(Fl_Button*, playbackReadWriteNav*);
		friend void pause_callback(Fl_Button*, playbackReadWriteNav*);
		friend void fforward_callback(Fl_Button*, playbackReadWriteNav*);
		friend void frameloc_callback(Fl_Value_Input*, playbackReadWriteNav*);
		friend void statusbar_callback(Fl_Slider*, playbackReadWriteNav*);
		friend void speeddisplay_callback(Fl_Value_Input*, playbackReadWriteNav*);
		friend void speedbar_callback(Fl_Slider*, playbackReadWriteNav*);
		friend void reset_callback(Fl_Button*, playbackReadWriteNav*);
		friend void closegui_callback(Fl_Button*, playbackReadWriteNav*);

		// menu callbacks
		friend void new_callback(Fl_Menu_*, playbackReadWriteNav*);
		friend void open_callback(Fl_Menu_*, playbackReadWriteNav*);
		friend void save_callback(Fl_Menu_*, playbackReadWriteNav*);
		friend void save_as_callback(Fl_Menu_*, playbackReadWriteNav*);
		friend void close_callback(Fl_Menu_*, playbackReadWriteNav*);
		friend Fl_Window* make_window(playbackReadWriteNav*);
};

#endif
