//
// DIVERSE graphics interface -- Performer ---- version 2.0.7
// released: Fri May 24 16:57:38 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//

//  Simple class that sets the background to be blue instead of solid black.
//  This use the performer class "pfEarthSky"

#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>


class SkyGroundPerformer : public dtkAugment {
public:

  SkyGroundPerformer(dpf *app) ;
  ~SkyGroundPerformer(void) ;
  int postConfig(void);

private:

  dpf* App;
  pfEarthSky* eSky;
} ;

// create object and set description.
// be sure to validate it!
SkyGroundPerformer::SkyGroundPerformer(dpf *app) :
  dtkAugment("skyGroundPerformer") 
{
  setDescription("Set background to blue") ;
  App = app;
  validate();
} ;

// destructor
SkyGroundPerformer::~SkyGroundPerformer(void) {
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int SkyGroundPerformer::postConfig(void) {
  eSky = new pfEarthSky();
  eSky->setMode(PFES_BUFFER_CLEAR, PFES_FAST);
  eSky->setColor(PFES_CLEAR, 0.772, 0.807, 1.0, 1);
  
  dpfPipe *p;
  dpfPwin *w;
  dpfScreen *s;

  for (p = App->display()->first(); p; p = p->next()) {
    for (w = p->first(); w; w = w->next()) {
      for (s = w->first(); s; s = s->next()) {

        if (s->pfchan(0)) {
          s->pfchan(0)->setESky(eSky);
        }
        if (s->pfchan(1))
          s->pfchan(1)->setESky(eSky);
      }
    }
  }
  return DTK_REMOVE_CALLBACK ; 
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new SkyGroundPerformer(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
