#include <stdlib.h>
#include <stdio.h>
#include <dtk.h>
#include "tasc-callback.h"

extern dtkSharedMem *shm, *shmZ;
extern float positions[4];
extern float z;

Fl_Window *guiWin;

static Fl_Slider *z_Slider 	 = NULL;
static Fl_Value_Input *X1 	 = NULL,
		      *Y1 	 = NULL, 
		      *X2  	 = NULL, 
		      *Y2 	 = NULL, 
		      *z_Val 	 = NULL;

// updates the values of the start/end positions
void update (void *) {

	if(shm->read(&positions)) { isRunning = 0; return; }
	
	if(shm->differentByteOrder())
		for (int i=0; i < 4; i++)
			dtk_swapBytes (&positions[i],sizeof(float));
	
	X1->value(positions[0]); Y1->value(positions[1]);
	X2->value(positions[2]); Y2->value(positions[3]);			
	guiWin->redraw();
}

// puts the z (height) value to shm
static void put_value(){
	if(shm) 
		if(shmZ->write(&z)) { isRunning = 0; return; } 	////
}

// updates he z (height) value with the slider
static void deltaZ_callback(Fl_Widget*, void*) {	
	z_Val->value(z_Slider->value()); 
	z = z_Slider->value();
	put_value();
}

// submitting means dumping the (x,y) coordinates for the start/end
// points to a file called 'select.txt'
static void submit_callback(Fl_Widget*, void*) {

	FILE *out;
	char* selectFile = getenv("DPF_SELECT_FILE");
	bool freeFlag;

	if(selectFile) freeFlag = false;
	else freeFlag = true;
		
	if (!selectFile) {
		selectFile = (char *) dtk_malloc(strlen("select.txt") + 1);
		sprintf(selectFile,"%s", "select.txt");
	}

	if( (out = fopen(selectFile,"w")) == NULL) {
		printf("Error: could not create \"%s\"\n", selectFile);
		if (selectFile) free(selectFile);
		return;
	}

	fprintf(out, "%f %f\n",	X1->value(), Y1->value());
	fprintf(out, "%f %f\n",	X2->value(), Y2->value());	
									
	fclose(out);
	
	if(freeFlag) free (selectFile);
	
	out = NULL;
}
// resets the values of the start/end positions, and z (height) slider values
static void reset_callback(Fl_Widget*, void*) {
	X1->value(0);		Y1->value(0);
	X2->value(0);		Y2->value(0);
	z_Val->value(2);	z_Slider->value(2);
}

static void quit_callback (Fl_Widget*, void*) {	isRunning = 0; }

// creates the window
void make_window () {
  Fl_Window* w;
  { Fl_Window* o = new Fl_Window(300, 290, "tasc_ControlPanel");	// 361, 240
    w = o;
    { Fl_Tile* o = new Fl_Tile(5, 10, 290, 145, "Path");	
      o->box(FL_EMBOSSED_FRAME);
      o->labeltype(FL_EMBOSSED_LABEL);
      o->labelfont(1);
      o->labelcolor(72);
      o->align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);
      { Fl_Box* o = new Fl_Box(10, 45, 95, 25, "Start Position:");	// 15,45,95,25
        o->labeltype(FL_EMBOSSED_LABEL);
        o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
      }
      { Fl_Value_Input* o = X1 = new Fl_Value_Input(130, 45, 60, 25, "X:  ");	// 145, 45, 80, 25
        o->labeltype(FL_EMBOSSED_LABEL);
      }
      { Fl_Value_Input* o = Y1 = new Fl_Value_Input(225, 45, 60, 25, "Y:  ");
        o->labeltype(FL_EMBOSSED_LABEL);
      }
      { Fl_Box* o = new Fl_Box(10, 80, 95, 25, "Stop Position:");
        o->labeltype(FL_EMBOSSED_LABEL);
        o->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
      }
      { Fl_Value_Input* o = X2 = new Fl_Value_Input(130, 80, 60, 25, "X:  ");
        o->labeltype(FL_EMBOSSED_LABEL);
      }
      { Fl_Value_Input* o = Y2 = new Fl_Value_Input(225, 80, 60, 25, "Y:  ");
        o->labeltype(FL_EMBOSSED_LABEL);
      }
      { Fl_Button* o = new Fl_Button(15, 120, 130, 25, "Submit");
        o->labeltype(FL_EMBOSSED_LABEL);
        o->callback(submit_callback, NULL);
      }
      { Fl_Button* o = new Fl_Button(155, 120, 130, 25, "Reset");
      	o->labeltype(FL_EMBOSSED_LABEL);
      	o->callback(reset_callback, NULL);
      }
      o->end();
    }
    
    { Fl_Tile* o = new Fl_Tile(5, 165, 290, 75, "Height from terrain");
      o->box(FL_EMBOSSED_FRAME);
      o->labeltype(FL_EMBOSSED_LABEL);
      o->labelfont(1);
      o->labelcolor(72);
      o->align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);
      { Fl_Value_Input* o = z_Val = new Fl_Value_Input(35, 200, 60, 25, "Z:  ");
        o->labeltype(FL_EMBOSSED_LABEL);
        o->minimum(0);
        o->maximum(20);
        o->value(2);
        o->step(0.0001);
        
      }
      { Fl_Slider* o = z_Slider = new Fl_Slider(100, 200, 185, 25);
        o->type(1);
        o->minimum(2.00);
        o->maximum(100);
        o->step(0.0001);
        o->value(2);
        o->callback(deltaZ_callback, NULL);
      }
      o->end();
    }
    { Fl_Button* o = new Fl_Button(5, 250, 290, 30, "Quit");
      o->labeltype(FL_EMBOSSED_LABEL);
      o->callback((Fl_Callback*)quit_callback);
    }
    o->end();
  }
  guiWin = w;
  w->show();
}
