#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dtk.h>

#include "tasc-callback.h"

int isRunning = 1;

dtkSharedMem *shm  = NULL,
   	     *shmZ = NULL;
			 
float positions[4];
float z = 2;

extern void update (void *);

static void catcher(int sig) {
	dtkMsg.add(DTKMSG_INFO,"caught signal %d: exiting.\n",sig);
	isRunning = 0;
}

// main
int main (int argc, char** argv) {

	// get shared mem segments
	shm  = new dtkSharedMem(sizeof(float)*4, "select", &positions);
	shmZ = new dtkSharedMem(sizeof(float),   "walk", &z);
	
	// checks to see if they are valid
	if (shm==NULL) { printf("shm\n"); return 1;}
	if (shmZ==NULL){ printf("shmz\n"); return 1;}
		
	if (shm->isInvalid()) { 
		printf("shm invalid\n");
		delete shm;
		shm = NULL;
		return 1;	
	}
	
	if(shmZ->isInvalid()) {
		printf("shmz invalid\n");
		delete shmZ;
		shmZ = NULL;
		return 1;
	}

	size_t size = shm->getSize();
	if(size < (sizeof(float)*4))
		return dtkMsg.add(DTKMSG_ERROR,0,1,
			    "DTK shared memory named \"%s\" is smaller"
			    " than a %s.\n", "select", "float");
			  
	size = shmZ->getSize();  
   	if(size < sizeof(float))
		return dtkMsg.add(DTKMSG_ERROR,0,1,
			    "DTK shared memory named \"%s\" is smaller"
			    " than a %s.\n", "walk", "float");
			    
	// signal catchers
	signal(SIGINT, catcher);
	signal(SIGTERM, catcher);
  	signal(SIGQUIT, catcher);	
  	
  	// creates the gui window
	make_window();
  
	// if did not create window, clean up memory and exit	
	if(!guiWin) {

		if (shm) {
			delete shm;
			shm = NULL;
		}
		
		if(shmZ) {
			delete shmZ;
			shmZ = NULL;
		}
		
		return 1;
	}

	// set the initial height above the terrain to be 2 units
	if(shm) if(shm->write(&z)) { isRunning = 0; }

	// loop through
	while (isRunning){
	
		if(guiWin->visible_r()) {
        		while (isRunning && Fl::wait(.75) > 0);
	  			update(NULL);
	  		
	        } else {
        		while(isRunning && ((Fl::wait()>0) || !guiWin->visible_r()));
		}
	}

	// clean up memory
	if(shm){
		delete shm;
		shm = NULL;
	}
	
	if(shmZ) {
		delete shmZ;
		shmZ = NULL;
	}
	
	// exit
	return 0;	
}
