#ifndef ITEM_H
#define ITEM_H

//Data container for an item
struct Item
{
	string Name;                 //Every item has a name
	string Description;          //Description of a name
	string URL;                  //URL where it can be downloaded
	bool Control;                //Is it being controlled by someone?
	string Controller;           //Who is controlling the item?
	vector<Item> Children;       //The children that this item has
	float LastKnownPosition[6];
};

class ItemServer
{
	public:

	ItemServer(const string& Name);
	~ItemServer();

	//Start the server loop
	void Serve();

	//Add or remove items from the world
	void AddItem(const string& Name, const string& URL);
	void RemoveItem(const string& Name);

	//Request or release control of objects
	void RequestControl(const string& From, const string& Name);
	void ReleaseControl(const string& Name);

	//List objects in the world
	void ListObjects(const string& To, const string& TheMessage);
	//Get the url of an object
	void GetURL(const string& To, const string& Object);

	//Get informatoin about the object
	void GetInfo(const string& To, const string& From);
	//Set information about the object
	void SetInfo(const string& Object, const string& Info);

	//Associate two items together
	void Associate(const string& Parent, const string& Child);
	//Disassociate two item
	void DisAssociate(const string& Child);
	//Replicate an object in the world
	void Replicate(const string& Object);

	//Register a user in the world
	void Register(const string& Computer, const string& Segment);
	//Unregister a user from the world
	void UnRegister(const string& Computer, const string& Segment);
	
	//Send the url of the file to download to the user
	void SendURL(const string& From, const string& Object);
	//Update the url for the computer
	void UpdateURL(const string& Object, const string& URL);
	//Update the last known position of the object
	void UpdatePosition(const string& Object, float* Position);
	//Send the last known positions of the objects to the client
	void SendPosition(const string& To);

	private:
	Communicator*  CLine;           //Communication gateway
	MultiCast      MCast;           //Multicasting object for messages
	//vector<dtkSharedMem*> Positions;//Positions of each user
	vector<Item>   Objects;         //List of objects I am taking care of
	//vector<Person> People;          //List of people that are in my world
	string	       BaseURL;         //The base url where you can get
	//string	       HostName;        //The hostname of the server
	//dtkClient*     RemoteConnector; //Used to connect the remote shared memory segements
	/*
	vector<string> Objects;         //List of objects
	vector<string> Descriptions;    //List of objects
	vector<string> URLS;            //List of urls
	vector<bool>   Control;         //List of controllers
					// the files if you don't have them
	*/
};
#endif
