///////////// MessageParser.h //////////////////////
//  Used to parse commands, or messages.
////////////////////////////////////////////////////

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//
// All Messages should be in the following form.
// 
// A string of >= 1024 characters
//
// Formatting as follows.
//
// ______ <tab> ____ <tab> ____________ <tab> _______ <tab> ___________________
// (From)       (To)       (Time Stamp)       (see *)       (parameters if any)
//
// * the actual message or command
//
// Examples:
//
// server <tab> pitr.sv.vt.edu <tab> 08:30:00 <tab> disconnect
//
// Class1 <tab> Class2 <tab> 09:30:00 <tab> Set_Speed <tab> 20
// 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef MESSAGEPARSER_H
#define MESSAGEPARSER_H

#include <string>
#include <sstream>

using namespace std;

class MessageParser
{
  private:
    string CurrMessage;
    istringstream* istr;
    char buff[1024];
    dtkSharedMem* Receive;

    string from;
    string to;
    string time;
    string message;

  public:
    MessageParser(string ShmName = "");
    ~MessageParser();
    
    int NextMessage();
    
    string getFrom() {return from;}
    string getTo() {return to;}
    string getTime() {return time;}

//  purpose:    To return only the actual message/command of the  entire message.
    string getMessage() {return message;}

    int getInt();
    // Returns everything after the message until the delimiter
    string getLine(char delimiter = '\0');
    string getString();
    char getChar();
    double getDouble();
    
};

#endif
