#ifndef Rule_h
#define Rule_h

struct ForwardContainer
{
	string AMessage;
	string Destination;
	string Segment;
	int Port;
};

struct TransformationContainer
{
	string AMessage;
	string NewMessage;
	string Destination;
	string Segment;
	string Data;
	int Port;
};

class TransitMessage
{
	public:
		bool AddMessage(const string& AMessage){this->AMessage = AMessage;};
		bool RemoveMessage(){AMessage = "";};
		string GetMessage() {return AMessage;};
	private:
		string AMessage;
};

//Add a rule for a message
class TransformationRule : public TransitMessage
{
	public:
	TransformationRule();
	TransformationRule(TransformationContainer* Transform);
	~TransformationRule();

	bool AddTransformation(TransformationContainer* Transformation);
	bool RemoveTransformation();
	TransformationContainer* GetTransformation();

	//Actually perform the transformation 
	bool DoTransformation(Message& AMessage);

	private:
	TransformationContainer* Container;
	Communicator* Sender;
};

class ForwardRule : public TransitMessage
{
	public:
	ForwardRule();
	ForwardRule(ForwardContainer* Forward);
	~ForwardRule();
	bool AddForward(ForwardContainer* Forward);
	ForwardContainer* GetForward();
	bool RemoveForward();
	bool DoForward(Message& AMessage);

	private:
	ForwardContainer* Container;
	Communicator* Sender;
};

class RuleManager
{

	public:

	RuleManager();
	~RuleManager();

	/**** Actual usage functions ****/

	//Check to see if a rule is valid
	bool Check(const string& MessageName);
	//Handle the rule
	bool Handle(Message& AMessage);

	/**** End actual usage functions ****/


	/**** Begin Forwarding setup features ****/

	//Add a message to be forwarded to the system
	//Will create a forward or will add another destination
	bool AddForward(ForwardContainer* Forward);

	//Will remove a specific forward
	bool RemoveForward(ForwardContainer* Forward);
	//Will remove all forwards for a message
	bool RemoveAllForwards(const string& Message);
	vector<ForwardRule*> GetAllForwards(const string& AMessage);
	vector<ForwardRule*> GetAllForwards();

	/**** End Forwarding setup features ****/

	/**** Begin Transaction functions ****/

	//Will add a transformation to an object
	bool AddTransformation(TransformationContainer* Transform);
	//Will remove a transformation from the system
	bool RemoveTransformation(TransformationContainer* Transform);
	//Will remove all transformations for a message
	bool RemoveAllTransformations(const string& AMessage);
	vector<TransformationRule*> GetAllTransformations(const string& AMessage);
	vector<TransformationRule*> GetAllTransformations();

	/**** End Forward Transaction functions ****/


	/**** Begin utility function ****/
	bool ClearEverything();
	//Clear out all forwards / transforms on this message
	bool ClearMessage(const string& AMessage);


	private:
	vector<TransformationRule*> TList;
	vector<ForwardRule*>        FList;
};

#endif
