#ifndef base_server_h
#define base_server_h

struct Segment
{
	bool                    Listen;
	bool                    BroadCast;
	bool 			BroadCastAll;
	vector<string>          SegmentName;
	vector<communicator>	Data;
	string		        Name;
	vector<string>          Messages;
};

class Base_Server
{

	public:
	Base_Server();
	~Base_Server();

	//The basic serve function that handles messages once
	void Serve();

	//Segment adding functions
	
	//Add a segment that the server listens to
	void AddListenSegment(const string& Segment, const string& HostName);
	//Add a segment that the server broadcasts to
	void AddBroadCastSegment(const string& Segment);  


	//Set up basic rules for listening to messages
	void AddListenForMessage(const string& Segment, const string& Message, string* Targets, const int& NumTargets);
	void BroadCastMessageFromListener(const string& Message, const string& LSegment, const string& BSegment);
	void BroadCastListenToBroadcast(const string& LSegment, const string& BSegment);

	//Segment removal functions
	//Remove a listen segment
	void RemoveListenSegment(const string& Segment);
	//Remove a broadcast segment
	void RemoveBroadcastSegment(const string& Segment);

	//Size and access functions
	
	//Listener functions
	int NumListenSegments();
	string* GetListenSegments();

	//Broadcast functions
	int NumBroadCastSegments();
	string* GetBroadCastSegments();

	private:
	vector<Segment*> ListenSegment;
	vector<Segment*> BroadCastSegment;
};
#endif
