////////////////// Messenger.h ////////////////////////////
//
// this class will hold the shared mem Send so you don't have to
// define it 10 million times.
// ////////////////////////////////

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//
// All Messages should be in the following form.
// 
// A string of >= 1024 characters
//
// Formatting as follows.
//
// ______ <tab> ____ <tab> ____________ <tab> _______ <tab> ___________________
// (From)       (To)       (Time Stamp)       (see *)       (parameters if any)
//
// * the actual message or command
//
// Examples:
//
// server <tab> pitr.sv.vt.edu <tab> 08:30:00 <tab> disconnect
//
// Class1 <tab> Class2 <tab> 09:30:00 <tab> Set_Speed <tab> 20
// 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef MESSENGER_H
#define MESSENGER_H

#include <string>
#include <iostream>
//#include <ctime>

using namespace std;

#define SHM_SIZE 1024*sizeof(char)

class Messenger
{
  public:
    Messenger(string ShmName = "");
    ~Messenger();

    // Connect the shared mem to another computer
    int connectRemoteSharedMem(const string& Name, const string& host);

    // First set all of these. So the complete message can have structure.
    int setFrom(string From) {from = From; return 0;} //Should only have to do this once.
    int setTo(string To) {to = To;return 0;}
    int setMessage(string Message) {message = Message;return 0;}
    int setTime(string pTime) {Time = pTime;return 0;}
    
    // Use these to add parameters
    int addInt(int add);
    int addDouble(double add);
    int addString(string add);
    int addChar(char add);
    
    // Finally use this to send the message off.
    // this function will construct the message.
    int SendMessage();
    
  private:
    // New Message stuff.
    dtkSharedMem* Send;
    dtkClient* ClientManager;
    dtkTime Timer;

    string from;
    string to;
    string message;
    string parameters;
    string Time;

    string getTime();
    
};

#endif
