#include <dtk.h>                         //For dtk stuff
#include <FL/fl_file_chooser.H>          //For the browse button
#include <vector>                        //For vectors
#include <iostream>                      //For strings
#include <string>                        //For strings
using namespace std;
#include "Communicator.h"                //For message passing
#include "ItemServer.h"                  //For the fltk gui
#include "func_declarations.h"           //For function declarations

//Basically a struct that contains the parent and children
//information about associating objects together

//Global variables
vector<Info> Parents;         //Parent info
string HostName;              //The hostname for this computer
string TempItem;              //Temporary storage string
string ControlledItem;        //The item the gui has control over
Communicator* MCenter;        //The message center
dtkSharedMem* ObjectPosition; //The objects position in shared memory
float XYZIncrementValue;          //The xyz increment
float HPRIncrementValue;          //The hpr increment
//The x,y,z, and hpr values for an item
float X;
float Y;
float Z;
float H;
float P;
float R;
bool   Controlled;            //Are we controlling an item?

//Fl window variables
Fl_Window* MainWindow;            //Main fltk window
Fl_Window* AssociateWindow;       //The association  window
Fl_Window* QueryWindow;           //The query window
Fl_Window* InfoWindow;            //The add info window
Fl_Window* ItemManipulationWindow;//The item manipulation window
Fl_Window* URLWindow;             //The url window


//Basically setup the gui for the user 
void SetupWindows()
{
	Controlled = false;
	//Initalize variables, this function is much like 
	//a constructor for a class
	ControlledItem="";
	
	//Figure out what server I connect to

	//Figure out who I am
	char buff[64];
	unsigned int size = 64;
	gethostname(buff,size);
	HostName = string(buff);
	printf("My hostname is %s\n", HostName.c_str());

	string Name = HostName + "-ITEM_POSITION";
	ObjectPosition = new dtkSharedMem(sizeof(float)*6, Name.c_str());

	//Figure out what server I am supposed to connect to
	//If I am the server or the env is not set, do nothing
	string Server;
	char* Test = getenv("DPF_SHARED_SERVER");
	if (Test !=NULL)
		Server = string(Test);
	else
		Server = HostName;
	printf("Shared server is %s\n", Server.c_str());

	//Create the message center
	MCenter = new Communicator("ITEM_MESSAGE");
	//Hook the message server up to the server
	if (Server != HostName)
	MCenter->SetRemote(Server);
	//Make the main gui window
	MainWindow = make_main_window();
	//Make the associate window
	AssociateWindow = make_associate_window();
	//Make the query window
	QueryWindow = make_querywindow();
	//Make the information window
	InfoWindow = make_addwindow();
	//Hide the release button
	ReleaseButton->hide();

	ItemManipulationWindow = make_control_window();//The item manipulation window
	URLWindow = make_url_window();             //The url window
	printf("before callbacks\n");

	//Add callbacks for all of the windows
	AddCallbacks();
	Register();
	//Show the maindow 
	MainWindow->show();
	XYZIncrementValue = .01;
	HPRIncrementValue = 1;
	printf("Got this far?\n");
}

//Basic driver function
int main()
{
	//Setup the gui
	SetupWindows();
	//Infinite main loop that runs while the gui is active
	while (1)
	{
		//Handle messages sent to this program
		HandleMessages();
		//If the user closed the window, die
		if (!MainWindow->visible())
		{
			SendDeathMessage();
			return 0;
		}
		//Sleep so that the cpu is not eaten alive
		Fl::wait(.100000000);
		usleep(100000);
	}
	return 0;
}

void FakePopulate()
{
	char* A = new char[20];
	strcpy(A, "pitr.sv.vt.edu");
	char* B = new char[20];
	strcpy(B, "glyph.sv.vt.edu");
	char* C = new char[20];
	strcpy(C,"ok.sv.vt.edu");
	MainItemBrowser->add(A);
	MainItemBrowser->data(1, (void*)A);
	MainItemBrowser->show(1);
	MainItemBrowser->add(B);
	MainItemBrowser->data(2, (void*)B);
	MainItemBrowser->show(2);
	MainItemBrowser->add(C);
	MainItemBrowser->data(3, (void*)C);
	MainItemBrowser->show(3);
}

static void QueryButtonCallback(Fl_Widget*, void*)
{
	for (int i=1; i<= AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->selected(i))
		{
			printf("Associated item found at %d\n", i);
			Message Temp;
			Temp.SetTo("ITEM_SERVER");
			Temp.SetFrom(HostName);
			Temp.SetMessage("GET_INFO");
			Temp.AddString((char*)AssociatedItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			printf("Message sent\n");
			return;
		}
	}	
	//Add the item that has query information
	for (int i=1; i<= MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->selected(i))
		{
			printf("Deassociated item found at %d\n", i);
			Message Temp;
			Temp.SetTo("ITEM_SERVER");
			Temp.SetFrom(HostName);
			Temp.SetMessage("GET_INFO");
			Temp.AddString((char*)MainItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			printf("Message sent\n");
		}
	}	
	//Now the gui waits for the server to send a message which is 
	//caught by a function in Message.cpp
}

//Query window callbacks
static void QueryReturnCallback(Fl_Widget*, void*)
{
	//Hide the query window
	QueryWindow->hide();
}

//Main window callbacks
static void AddButtonCallback(Fl_Widget*, void*)
{
	printf("Add button called\n");
	//Hopefully people don't use files over 1024 characters long
	char* Temp = new char[1024];
	Temp = fl_file_chooser("Pick a file to load", NULL, NULL);
	if (Temp == NULL)
		return;
	char* Dir = NULL;
	char* App = NULL;
	if (Temp[0] == '/')
	{
		int Location = 0;
		for (unsigned int i=0; i< strlen(Temp); i++)
		{
			if (Temp[i] == '/')
				Location = i;
		}
		App = new char[strlen(Temp) - Location +1];
		//Copy the app name 
		for (unsigned int i=1; i< strlen(Temp) - Location+ 1; i++)
		{
			App[i-1] = Temp[Location+i];
		}
		//Get the path to the app
		Dir = new char[200];
		for (int i=0; i<Location; i++)
		{
			Dir[i] = Temp[i];
		}
		Dir[Location] = '\0';
	}
	else
	{
		App = strdup(Temp);
		Dir = new char[200];
		strcpy(Dir , getenv("PWD"));
	}
	//Now that we have the file, send it off
	printf("Application is %s, in directory %s\n", App, Dir);
	Message Temp1;
	Temp1.SetTo("ITEM_SERVER");
	Temp1.SetFrom(HostName);
	Temp1.SetMessage("ADD_ITEM");
	Temp1.AddString(App);
	Temp1.AddString(Dir);
	MCenter->SendMessage(Temp1);
}

static void RemoveButtonCallback(Fl_Widget*, void*)
{
	//Remove from browser, send message
	printf("Remove button called\n");
	Message Temp;
	Temp.SetTo("ITEM_SERVER");
	Temp.SetFrom(HostName);
	Temp.SetMessage("REMOVE_ITEM");
	for (int i=1; i<= AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
		{
			char* Temp2 = (char*)AssociatedItemBrowser->data(i);
			Temp.AddString(string(Temp2));
			MCenter->SendMessage(Temp);
			return;
		}
	}
	for (int i=1; i<= MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() == i)
		{
			char* Temp2 = (char*)MainItemBrowser->data(i);
			Temp.AddString(string(Temp2));
			MCenter->SendMessage(Temp);
			return;
		}
	}
}

static void HighlightButtonCallback(Fl_Widget*, void*)
{
	//Send message to dso containing the name
	printf("Highlight button called\n");
	Message Temp;
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.SetMessage("HIGHLIGHT_ITEM");
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
		{
			Temp.AddString((char*)AssociatedItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			return;
		}
	}
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() ==i)
		{
			Temp.AddString((char*)MainItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			return;
		}
	}
}

static void JumpButtonCallback(Fl_Widget*, void*)
{
	//Send message to dso and then move to a certain position
	printf("Jump button called\n");
	Message Temp;
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.SetMessage("JUMP_TO_ITEM");
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
		{
			Temp.AddString((char*)AssociatedItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			return;
		}
	}
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() ==i)
		{
			Temp.AddString((char*)MainItemBrowser->data(i));
			MCenter->SendMessage(Temp);
			return;
		}
	}
}

static void AssociateButtonCallback(Fl_Widget*, void*)
{
	bool Found = false;
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() == i)
			Found = true;
	}
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
			Found = true;
	}
	if (!Found)
		return;
		
	//Check to see if one item is selected in the main window
	//If not, exit
	//Else, bring up association window 
	printf("Associate button called\n");
	//Build up a list of the items not selected in the main browser
	MainItemBrowser2->clear();
	bool DoNotDo = false;
	//Go through each entry in the main browser
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		//If I am the parent put my info in the associate window
		if (MainItemBrowser->value() == i)
		{
			ParentItem->label((char*)MainItemBrowser->data(i));
		}
		//If I am not the parent and I have no children add me to the list
		else
		{
			char* Temp = (char*)MainItemBrowser->data(i);
			//Get a pointer out of the widget for easy comparison
			for (unsigned int j=0; j<Parents.size(); j++)
			{
				//If an entry is already a parent you cannot associate it
				if (!strcmp(Parents[j].Parent.c_str(), Temp) )
						DoNotDo=true;
				//Otherwise add it to the association window
			}
			//If I am not a parent then add me to the potential list of children
			if (!DoNotDo)
			{
				MainItemBrowser2->add(Temp);
				MainItemBrowser2->data(MainItemBrowser2->size(), (char*)MainItemBrowser->data(i));
				MainItemBrowser2->show(MainItemBrowser2->size());
			}
			else
				DoNotDo = false;
		}
	}
	AssociateWindow->show();
}

static void DeassociateButtonCallback(Fl_Widget*, void*)
{
	//Make sure one or several items are selected in the
	//association browser and then deassociate it
	printf("Deassociate button called\n");
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		//If the item at i is highlighted do some work
		if (AssociatedItemBrowser->value() == i)
		{
			//Get the highlighted items data (name)
			char* Temp = (char*)AssociatedItemBrowser->data(i);
			Message AMessage;
			AMessage.SetTo("ITEM_SERVER");
			AMessage.SetFrom(HostName);
			AMessage.SetMessage("DISASSOCIATE");
			AMessage.AddString(Temp);
			MCenter->SendMessage(AMessage);
		}
	}
}

static void RequestButtonCallback(Fl_Widget*, void*)
{
	//Request control for one object in the browswer windows
	printf("Request button called\n");
	for (int i=1; i<= AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->selected(i))
		{
			ControlledItem = string((char*)AssociatedItemBrowser->data(i));
			break;
		}
	}
	if (ControlledItem == "")
	{
	for (int i=1; i<= MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->selected(i))
		{
			ControlledItem = string((char*)MainItemBrowser->data(i));
			break;
		}
	}
	}
	if (ControlledItem == "")
	{
		printf("Could not find the controlled item \n");
		return;
	}
	
	Message Temp;
	Temp.SetTo("ITEM_SERVER");
	Temp.SetFrom(HostName);
	Temp.SetMessage("REQUEST_ITEM_CONTROL");
	Temp.AddString(ControlledItem);
	MCenter->SendMessage(Temp);
}

static void ReleaseButtonCallback(Fl_Widget*, void*)
{
	//Release the item you have control over
	printf("Release button called\n");
	Message Temp;
	Temp.SetTo("ITEM_SERVER");
	Temp.SetFrom(HostName);
	Temp.SetMessage("RELEASE_ITEM_CONTROL");
	Temp.AddString(ControlledItem);
	MCenter->SendMessage(Temp);
}
static void ReplicateButtonCallback(Fl_Widget*, void*)
{
	//Send message to server saying replicate
	printf("Replicate button called\n");
	//Print out the parent child information
	bool Found = false;
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() == i)
			Found = true;
	}
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
			Found = true;
		
	}
	if (Found)
	{
		AddURLBox->value("");
		URLWindow->show();
	}
}

//Associate Window callbacks
static void OkButtonCallback(Fl_Widget*, void*)
{
	//Send message to server saying replicate
	for (int i=1; i<=MainItemBrowser2->size(); i++)
	{
		if (MainItemBrowser2->selected(i) )
		{
				
			string A = ParentItem->label();
			string B = (char*)MainItemBrowser2->data(i);
			printf("Parent is %s , child is %s\n", A.c_str(), B.c_str());
			Message Test;
			Test.SetTo("ITEM_SERVER");
			Test.SetFrom(HostName);
			Test.SetMessage("ASSOCIATE");
			Test.AddString(A);
			Test.AddString(B);
			MCenter->SendMessage(Test);
		}
	}
	AssociateWindow->hide();
}

static void MainItemBrowserCallback(Fl_Widget*, void*)
{
	AssociatedItemBrowser->clear();
	printf("Call back for the main item browser!!\n");
	//Display all the children of the highlighted parent
	for (int i=1; i<= MainItemBrowser->size(); i++)
	{
		//Item is selected
		if (MainItemBrowser->value() == i)
		{
			//Clear out old list
			//Add in correct child info
			char* Temp = (char*)MainItemBrowser->data(i);
			printf("Parent is %s\n", Temp);
			for (unsigned int j=0; j<Parents.size(); j++)
			{
				//Found the right parent
				if (Parents[j].Parent == string(Temp))
				{
					//printf("Found parent matching, parent has %d children\n", Parents[j].Children.size());
					//Add in the children for the parent
					for (unsigned int k=0; k< Parents[j].Children.size(); k++)
					{
						AssociatedItemBrowser->add(strdup(Parents[j].Children[k].c_str()));
						AssociatedItemBrowser->data(AssociatedItemBrowser->size(),strdup(Parents[j].Children[k].c_str()));
					}
				}
			}
		}
	}
}


//Basically when the user wants to add information to an item
//Show the infowindow and wait for the callback to be registered
static void AddQInfoCallback(Fl_Widget*, void*)
{
	bool Set = false;
	for (int i=1; i<=AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
		{
			TempItem = string((char*)AssociatedItemBrowser->data(i));
			Set = true;
		}
	}
	if (!Set)
	{
	for (int i=1; i<=MainItemBrowser->size(); i++)
	{
		if (MainItemBrowser->value() == i)
		{
			TempItem = string((char*)MainItemBrowser->data(i));
			Set = true;
		}
	}
	}
	if (!Set)
	{
		printf("Nothing selected, wtf\n");
		return;
	}
	AddInputBox->value("");
	InfoWindow->show();
}

//Send a message saying add information to a specific object
//Hide the info window when the function is finished
static void AddInfoButtonCallback(Fl_Widget*, void*)
{
	Message Temp;
	Temp.SetTo("ITEM_SERVER");
	Temp.SetFrom(HostName);
	Temp.SetMessage("ADD_INFO");
	printf("Setting information for object %s\n", TempItem.c_str());
	Temp.AddString(TempItem);
	string Testing = AddInputBox->value();
	if (Testing == "")
		return;
	Temp.AddString(string(AddInputBox->value()));
	AddInputBox->value("");
	MCenter->SendMessage(Temp);
	InfoWindow->hide();
}


//Send an update message to my dso saying to move the position of 
//the item
void SendUpdateMessage(const string& Variable, const float& Value)
{
	Message Temp;
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.AddString(Variable);
	Temp.AddFloat(Value);
	Temp.SetMessage("UPDATE_POSITION");
	MCenter->SendMessage(Temp);
}

//Handle the users pressing the + or - key
static void XIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("X", X+XYZIncrementValue);
}
static void YIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("Y", Y + XYZIncrementValue);
}
static void ZIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("Z", Z + XYZIncrementValue);
}
static void HIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("H", H + HPRIncrementValue);
}

static void PIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("P", P + HPRIncrementValue);
}

static void RIncrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("R", R + HPRIncrementValue);
}

static void XDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("X", X - XYZIncrementValue);
}
static void YDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("Y", Y - XYZIncrementValue);
}
static void ZDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("Z", Z - XYZIncrementValue);
}
static void HDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("H", H - HPRIncrementValue);
}
static void PDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("P", P - HPRIncrementValue);
}
static void RDecrementCallback(Fl_Widget*, void*)
{
	SendUpdateMessage("R", R - HPRIncrementValue);
}
//Handle users typing in a value for the position of an item
static void X_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(X_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &X);
	SendUpdateMessage("X", X);
}

static void Y_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(Y_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &Y);
	SendUpdateMessage("Y", Y);
}

static void Z_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(Z_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &Z);
	SendUpdateMessage("Z", Z);
}

static void H_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(H_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &H);
	SendUpdateMessage("H", H);
}

static void P_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(P_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &P);
	SendUpdateMessage("P", P);
}

static void R_InputCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(R_Input->value());
	if (Temp !=NULL)
		sscanf(Temp, "%f\n", &R);
	SendUpdateMessage("R", R);
}

//When the user specifies the url to add for an object do this
static void AddURLButtonCallback(Fl_Widget*, void*)
{
	URLWindow->hide();
	string URL = AddURLBox->value();
	string Item = "";
	bool Found = false;
	for (int i=1; i<= AssociatedItemBrowser->size(); i++)
	{
		if (AssociatedItemBrowser->value() == i)
		{
			Item = string((char*)AssociatedItemBrowser->data(i));
			Found = true;
		}
	}
	if (!Found)
	{
		for (int j=1; j<= MainItemBrowser->size(); j++)
		{
			if (MainItemBrowser->value() == j)
			{
				Item = string((char*)MainItemBrowser->data(j));
				Found = true;
			}
		}
	}
	if (!Found || URL == "")
		return;
	Message Temp;
	Temp.SetTo("ITEM_SERVER");
	Temp.SetFrom(HostName);
	Temp.AddString(Item);
	Temp.AddString(URL);
	Temp.SetMessage("UPDATE_URL");
	MCenter->SendMessage(Temp);
}

//Change the xyz or hpr increment that the sliders use
static void XYZIncrementCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(XYZIncrement->value());
	sscanf(Temp, "%f", &XYZIncrementValue);
	free(Temp);
	Message Temp1;
	Temp1.SetTo(HostName);
	Temp1.SetFrom(HostName);
	Temp1.SetMessage("XYZ");
	Temp1.AddFloat(XYZIncrementValue);
	MCenter->SendMessage(Temp1);
}

static void HPRIncrementCallback(Fl_Widget*, void*)
{
	char* Temp = strdup(HPRIncrement->value());
	sscanf(Temp, "%f", &HPRIncrementValue);
	free(Temp);
	Message Temp1;
	Temp1.SetTo(HostName);
	Temp1.SetFrom(HostName);
	Temp1.SetMessage("HPR");
	Temp1.AddFloat(HPRIncrementValue);
	MCenter->SendMessage(Temp1);
}

static void LoadButtonCallback(Fl_Widget*, void*)
{
	//C input, yay!
	FILE* InputFile;
	InputFile = fopen("temp-position", "r");
	if (InputFile == NULL)
		return;

	char* Buff = new char[800];
	fgets(Buff, 255, InputFile);
	//sscanf(Buff, "%f %f %f %f %f %f %f %f", &X, &Y, &Z, &H &P, &R, &XYZIncrementValue, &HPRIncrementValue);
	delete [] Buff;

	SendUpdateMessage("X", X);
	SendUpdateMessage("Y", Y);
	SendUpdateMessage("Z", Z);
	SendUpdateMessage("H", H);
	SendUpdateMessage("P", P);
	SendUpdateMessage("R", R);
}

//Save the gui's data values
static void SaveButtonCallback(Fl_Widget*, void*)
{
	FILE* OutputFile;
	OutputFile = fopen("temp-position", "w");
	fprintf(OutputFile, "%s %s %s %s %s %s %f %f\n", X_Input->value(), Y_Input->value(), Z_Input->value(), H_Input->value(), P_Input->value(), R_Input->value(), XYZIncrementValue, HPRIncrementValue);
}

//Send a message to mainpulate the item mode in the dso
void SendManipulationMessage(const string& AName, const string& AStatus, const string& MessageMe)
{
	Message Temp;
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.SetMessage(MessageMe);
	Temp.AddString(AName);
	Temp.AddString(AStatus);
	MCenter->SendMessage(Temp);
}

static void AllCallback(Fl_Widget*, void*)
{
	if (AllButton->value() == 0)
	{
		AllButton->value(1);
		return;
	}
	//Turn all the other buttons off
	Navigate_Button->value(0);
	X_Button->value(0);
	Y_Button->value(0);
	Z_Button->value(0);
	Zoom_Button->value(0);
	Rotate_Button->value(0);
	Pan_Button->value(0);
	//Return to normal mainpulation
	Message Temp; 
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.SetMessage("RETURN_TO_NORMAL_MANIPULATION");
	MCenter->SendMessage(Temp);
}

//Used for zooming in / out
static void ZoomButtonCallback(Fl_Widget*, void*)
{
	//Button is turned off
	if (Zoom_Button->value() == 0)
	{
		printf("Zoom button turned off\n");
		SendManipulationMessage("ZOOM", "OFF", "SET_MANIPULATION_MODE");
		AllCallback(NULL, NULL);
		return;
	}
	AllButton->value(0);
	Pan_Button->value(0);
	Rotate_Button->value(0);
	Navigate_Button->value(0);
	X_Button->value(0);
	Y_Button->value(0);
	Z_Button->value(0);
	SendManipulationMessage("ZOOM", "ON", "SET_MANIPULATION_MODE");
}

//Used for rotating
static void RotateButtonCallback(Fl_Widget*, void*)
{
	if (Rotate_Button->value() == 0)
	{
		printf("Rotate button turned off\n");
		X_Button->value(0);
		Y_Button->value(0);
		Z_Button->value(0);
		SendManipulationMessage("ROTATE", "OFF", "SET_MANIPULATION_MODE");
		AllCallback(NULL, NULL);
		return;
	}
	AllButton->value(0);
	Zoom_Button->value(0);
	Pan_Button->value(0);
	Navigate_Button->value(0);
	SendManipulationMessage("ROTATE", "ON", "SET_MANIPULATION_MODE");
}

//Used for panning
static void PanButtonCallback(Fl_Widget*, void*)
{
	if (Pan_Button->value() == 0)
	{
		printf("Pan button turned off\n");
		SendManipulationMessage("PAN", "OFF", "SET_MANIPULATION_MODE");
		AllCallback(NULL, NULL);
		return;
	}
	AllButton->value(0);
	Zoom_Button->value(0);
	Rotate_Button->value(0);
	Navigate_Button->value(0);
	X_Button->value(0);
	Y_Button->value(0);
	Z_Button->value(0);
	SendManipulationMessage("PAN", "ON", "SET_MANIPULATION_MODE");
}

//Did I hit the x,y, or z button?
static void XButtonCallback(Fl_Widget*, void*)
{
	//Parent not turned on , halt
	if (Rotate_Button->value() == 0)
	{
		X_Button->value(0);
		Y_Button->value(0);
		Z_Button->value(0);
		return;
	}
	//Turning myself off
	if (X_Button->value() == 0)
	{
		SendManipulationMessage("X", "OFF", "SET_ROT_MODE");
		return;
	}
	//Turn siblings off
	Y_Button->value(0);
	Z_Button->value(0);
	//Tell dso what axis to play with
	SendManipulationMessage("X", "ON", "SET_ROT_MODE");
}
static void YButtonCallback(Fl_Widget*, void*)
{
	//Parent not turned on , halt
	if (Rotate_Button->value() == 0)
	{
		X_Button->value(0);
		Y_Button->value(0);
		Z_Button->value(0);
		return;
	}
	//Turning myself off
	if (Y_Button->value() == 0)
	{
		SendManipulationMessage("Y", "OFF", "SET_ROT_MODE");
		return;
	}
	//Turn siblings off
	X_Button->value(0);
	Z_Button->value(0);
	//Tell dso what axis to play with
	SendManipulationMessage("Y", "ON", "SET_ROT_MODE");
}
static void ZButtonCallback(Fl_Widget*, void*)
{
	//Parent not turned on , halt
	if (Rotate_Button->value() == 0)
	{
		X_Button->value(0);
		Y_Button->value(0);
		Z_Button->value(0);
		return;
	}
	//Turning myself off
	if (Z_Button->value() == 0)
	{
		SendManipulationMessage("Z", "OFF", "SET_ROT_MODE");
		return;
	}
	//Turn siblings off
	X_Button->value(0);
	Y_Button->value(0);
	//Tell dso what axis to play with
	SendManipulationMessage("Z", "ON", "SET_ROT_MODE");
}

static void NavigateCallback(Fl_Widget*, void*)
{
	Message Temp;
	Temp.SetTo(HostName);
	Temp.SetFrom(HostName);
	Temp.SetMessage("NAVIGATION");
	if (Navigate_Button->value())
	{
		Temp.AddString("ON");
		AllButton->value(0);
		X_Button->value(0);
		Y_Button->value(0);
		Z_Button->value(0);
		Zoom_Button->value(0);
		Rotate_Button->value(0);
		Pan_Button->value(0);
	}
	else
	{
		AllCallback(NULL, NULL);
		Temp.AddString("OFF");
	}
	MCenter->SendMessage(Temp);
}


//Add in all of the callbacks for all of the objects
void AddCallbacks()
{
	AllButton->callback(AllCallback, NULL);
	Navigate_Button->callback(NavigateCallback, NULL);
	X_Button->callback(XButtonCallback);
	Y_Button->callback(YButtonCallback);
	Z_Button->callback(ZButtonCallback);
	Zoom_Button->callback(ZoomButtonCallback, NULL);
	Rotate_Button->callback(RotateButtonCallback, NULL);
	Pan_Button->callback(PanButtonCallback, NULL);
	Load_Button->callback(LoadButtonCallback, NULL);
	Save_Button->callback(SaveButtonCallback, NULL);
	AddURLButton->callback(AddURLButtonCallback, NULL);
	XYZIncrement->callback(XYZIncrementCallback, NULL);
	HPRIncrement->callback(HPRIncrementCallback, NULL);
	X_Input->callback(X_InputCallback, NULL);
	Y_Input->callback(Y_InputCallback, NULL);
	Z_Input->callback(Z_InputCallback, NULL);
	H_Input->callback(H_InputCallback, NULL);
	P_Input->callback(P_InputCallback, NULL);
	R_Input->callback(R_InputCallback, NULL);
	XIncrement->callback(XIncrementCallback, NULL);
	YIncrement->callback(YIncrementCallback, NULL);
	ZIncrement->callback(ZIncrementCallback, NULL);
	HIncrement->callback(HIncrementCallback, NULL);
	PIncrement->callback(PIncrementCallback, NULL);
	RIncrement->callback(RIncrementCallback, NULL);
	XDecrement->callback(XDecrementCallback, NULL);
	YDecrement->callback(YDecrementCallback, NULL);
	ZDecrement->callback(ZDecrementCallback, NULL);
	HDecrement->callback(HDecrementCallback, NULL);
	PDecrement->callback(PDecrementCallback, NULL);
	RDecrement->callback(RDecrementCallback, NULL);
	AddInfoButton->callback(AddInfoButtonCallback, NULL);
	AddQInfo->callback(AddQInfoCallback, NULL);
	MainItemBrowser->callback(MainItemBrowserCallback, NULL);
	OkButton->callback(OkButtonCallback, NULL);
	ReplicateButton->callback(ReplicateButtonCallback, NULL);
	ReleaseButton->callback(ReleaseButtonCallback, NULL);
	RequestButton->callback(RequestButtonCallback, NULL);
	DeassociateButton->callback(DeassociateButtonCallback, NULL);
	AssociateButton->callback(AssociateButtonCallback, NULL);
	JumpButton->callback(JumpButtonCallback, NULL);
	HighlightButton->callback(HighlightButtonCallback, NULL);
	RemoveButton->callback(RemoveButtonCallback, NULL);
	AddButton->callback(AddButtonCallback, NULL);
 	QueryButton->callback(QueryButtonCallback, NULL);	
	QueryReturn->callback(QueryReturnCallback, NULL);
}
