///////////// MessageParser.h //////////////////////
//  Used to parse commands, or messages.
////////////////////////////////////////////////////


#ifndef MESSAGEPARSER_H
#define MESSAGEPARSER_H

#include <string>
#include <sstream>
#include <iostream>

using namespace std;

class MessageParser
{
  private:
    string CurrMessage;
    istringstream* istr;
    char buff[1024];
    dtkSharedMem* Receive;

  public:
    MessageParser(string ShmName = "");
    ~MessageParser();
    
    int getMessage(string& Message, string TargetClass);
    
    int getInt();
    string getString();
    string getLine();
    double getDouble();
    char getChar();
    
};

MessageParser::MessageParser(string ShmName = "")
{
  CurrMessage = "";
  istr = NULL;
  
  if(ShmName.length() == 0){
    char* charShmName = getenv("DTK_COMMAND_SHM_NAME");
    if(charShmName == NULL) ShmName = "CommandMessage";
    else ShmName = charShmName;

    if(ShmName.length() == 0) ShmName = "CommandMessage";
  }

  Receive = new dtkSharedMem(1024*sizeof(char),ShmName.c_str());
  if(!Receive || Receive->isInvalid() || Receive->getSize() < (1024*sizeof(char))) {
		printf ("could not get shared mem in MessageParser\n");
  } else {
    printf("M:connected to sharedMem = %s\n", ShmName.c_str());
    Receive->flush();
    Receive->queue();
  }
}

MessageParser::~MessageParser()
{
  delete istr;
  istr = NULL;
}

int MessageParser::getMessage(string& Message, string targetClass)
{
  Message = "BOGUS";
  if(!Receive->qread(buff)) {
    printf("M: did not read\n");
    return 0;
  }
   
  // Eric 07/18
  if(!buff){
    printf("M:buff NULL\n");
    return 0;
  }
  
  if(istr != NULL){
    delete istr;
    istr = NULL;
  }
  
  string newMessage = buff;
  
//  printf("M:message = %s\n", buff);   //Eric 07/18

  istr = new istringstream(newMessage);
  string classname;
  (*istr)>>classname;
  if(classname == "NONE" || classname == targetClass){
    //(*istr)>>Message;
    Message = buff;	// Eric 07/18
    return 1;
  }

  return 0;
}

int MessageParser::getInt()
{
  int nextInt=-1;
  if(istr)  (*istr)>>nextInt;	  //E
  return nextInt;
}

string MessageParser::getString()
{
  string nextString = "BOGUS";
  if(istr) (*istr)>>nextString;	  //E
  return nextString;
}

string MessageParser::getLine() {
  string nextString = "BOGUS";
  char buf[1024];
  
  if(istr) {
    char c = (*istr).get();
    (*istr).getline(buf, 1024, '\n');
  }
  nextString = buf;
  return nextString;
}

double MessageParser::getDouble()
{
  double nextD = -1;
  if(istr) (*istr)>>nextD;	  //E
  return nextD;
}

char MessageParser::getChar()
{
  char nextChar = '';
  if(istr) (*istr)>>nextChar;	  //E
  return nextChar;
}

#endif
