////////////////// Messenger.h ////////////////////////////
//
// this class will hold the shared mem Send so you don't have to
// define it 10 million times.
// ////////////////////////////////

#ifndef MESSENGER_H
#define MESSENGER_H

#include <string>

using namespace std;

#define SHM_SIZE 1024*sizeof(char)

class Messenger
{
  public:

    // Use this if you don't know your targetClass
    int SendMessage(string Message);
    // Use this if you do.
    int SendMessage(string Message, string targetClass);   
    Messenger(string ShmName = "");
    ~Messenger();
    
  private:
    // New Message stuff.
    dtkSharedMem* Send;
    
};

Messenger::Messenger(string ShmName = "")
{
  // New Message Stuff.
  
  //If user passed a string to the constructor it has presedence.
  if(ShmName.length() == 0){
    char* charShmName = getenv("DTK_COMMAND_SHM_NAME");
    if(charShmName == NULL) ShmName = "CommandMessage";
    else ShmName = charShmName;
    //If the length is still 0 then it's default
    if(ShmName.length() == 0) ShmName = "CommandMessage";
  }
    
  
  Send = new dtkSharedMem (SHM_SIZE, ShmName.c_str());
  if(!Send || Send->isInvalid() || Send->getSize() < SHM_SIZE){
    cout<<"could not get shared mem for Messenger"<<endl;
  }
 
  // End Message Stuff 
}

Messenger::~Messenger()
{
  delete Send;
}

//////////////////////////////////////////////////////////////////////////
//  Function: SendMessage
//
//  Purpose:  Writes to shared memory a string in the following format.
//            _______________ <tab> ______________ <tab> ______________ 
//            target class name     actual command       any parameters
//
//            and then the string is terminated with a null char. "\0"
//
//  Parameters:  A string (command) that will be passed into shared memory.
//
//  Returns: An integer according to how it executed.
//
//////////////////////////////////////////////////////////////////////////

int Messenger::SendMessage(string Message)
{
//  string NONE = "NONE";
//  string pass = NONE + "\t" + Message + "\0";
  string pass = Message + "\0";
  return Send->write(pass.c_str(),(pass.length()+1)*sizeof(char));
}

int Messenger::SendMessage(string Message, string targetClass)
{
  string pass = targetClass + "\t" + Message + "\0";
  return Send->write(pass.c_str(),(pass.length()+1)*sizeof(char));
}

#endif
