// This is the client to connect to the dedicated server

#include "dclient.h"
#include <dtk/dtkDSO_loader.h>

dclient::dclient(dpf *app_in) :
  dtkAugment("dclient")
{
  cout<<"dclient starting"<<endl;
  setDescription("connects to the server");
  app = app_in;
  mapSCS = NULL;
  validate();
}

dclient::~dclient(void)
{
  messenger->setTo("server");
  messenger->setMessage("disconnect");
  messenger->SendMessage();

  for(unsigned int i=0;i<users.size();i++){
    delete users[i];
  }

  if(positionShm != NULL)
    delete positionShm;

  if(parser) delete parser;

  if(messenger) delete messenger;
}

int dclient::preConfig(void)
{
  //Load the radar

  dtkAugment* radar = app->check("radar");
  if(radar == NULL){
    printf ("loading the radar\n");
    if(app->load("radar") == -1){
        dtkMsg.add(DTKMSG_ERROR,
          "dclient preConfig unable to load radar %s",
          __FILE__);
    }
  }       
  
  return DTK_CONTINUE;
}

int dclient::postConfig(void)
{
  cout<<"in postConfig"<<endl;
  
  char buff[64];
  unsigned int size = 64;
  gethostname(buff,size);
  compName = buff;
  ClientManager = new dtkClient;
  parser = new MessageParser("genMessage");
  messenger = new Messenger("genMessage");

  //DPF_SHARED_SERVER is the env varible.

  host = getenv("DPF_SHARED_SERVER");
  if(host == NULL){
    cout<<"no host name specified"<<endl;
    return -1;
  }

  if(messenger->connectRemoteSharedMem(host) == -1){
    cout<<compName.c_str()<<" could not connect to "<<&host<<endl;
  }

  string avatarname = getenv("DPF_AVATAR_NAME");
  messenger->setFrom(compName);
  messenger->setTo("server");
  messenger->setMessage("addme");
  messenger->addString(avatarname);

  messenger->SendMessage();

  usleep(300000); //Sleep for 1/3 of a second while server creates shm for it.

//  lastFrameTime = Timer.delta();
//  lastPingTime = -20;

  string tempname = "user-" + compName;

  positionShm = new dtkSharedMem (6*sizeof(float),tempname.c_str());
  if(!positionShm || positionShm->isInvalid() || positionShm->getSize() < 6*sizeof(float))
  {
    cout<<"Could not get shared mem for "<<tempname.c_str()<<endl;
  }

  return DTK_CONTINUE;
}

int dclient::postFrame(void)
{
/*  long double currTime = Timer.get();

  if(currTime - lastPingTime > 10){ //Check in w/the server every 10 sec.
    lastPingTime = currTime;
    messenger->setTo("server");
    messenger->setMessage("checkin");
    messenger->SendMessage();
  }
*/
//  if(currTime - lastFrameTime > .05 ){  //update every 1/20 seconds.
//    lastFrameTime = currTime;
    //  Convert 0,0,0 from scene to world units
    coord.xyz.set(0,0,0);
    coord.hpr.set(0,0,0);
    app->display()->dpfToWorld(&coord);

    //Access the position with coord.xyz and coord.hpr.
   
    positionShm->write(&coord,6*sizeof(float));
  
//  }

  CheckMessages();
 
  return DTK_CONTINUE; 
}

// Not so Deep Copy
vector<duser*> dclient::getUsers()
{
  /*
  vector<duser*> pass;
  duser* newuser;
  for(int i=0;i<users.size();i++){
    newuser = new duser(users[i]->getName(),users[i]->getAvatar());
    pass.push_back(newuser);
  }
  return pass;
  */
  return users;
}

int dclient::CheckMessages()
{
//  string* doh = NULL;
//  usleep(1000);
//  string doh2 = *doh;
  if(parser->NextMessage() == -1) return -1;

  if(parser->getFrom() == compName) return -1;

  if(parser->getTo() != compName && parser->getTo() != "everyone"){
   cout<<"Message not for me"<<endl;
   return -1;
  }
  
  string avatarname;
  string message = parser->getMessage();
  string temp;
  unsigned int avatarcounter =0;

  switch(message[0]){
    case 'n': //newuser
      temp = parser->getString(); //temp will be equal to the new user name
      avatarname = parser->getString();
      addNewUser(temp,avatarname);
      break;
    case 'u': //usernames
      if(message == "usernames"){
        temp = parser->getString();
        while(temp != ""){
          addNewUser(temp,"");
          temp = parser->getString();
        }
      }
      else if(message == "useravatars"){
        temp = parser->getString();
        while(temp!=""){
          if(avatarcounter < users.size()){
            cout<<"setting avatar "<<temp.c_str()<<endl;
            users[avatarcounter]->setAvatar(temp);
            avatarcounter++;
            temp = parser->getString();
          }
        }
      }
      break;
    case 'd':
      temp = parser->getString();
      disconnect(temp);
      break;
    default:
      //cout<<compName<<": I did not recognize the command "<<message<<endl;
      break;
  }

  return 0;
}

int dclient::disconnect(string user)
{
  int pos = SearchUsers(user);

  if(pos == -1) return -1; //user not found;

  if(users[pos] == NULL) return -1;
  cout<<"dropping user "<<users[pos]->getName().c_str()<<endl;
  users[pos]->removeDCS(app,mapSCS);
    
  delete users[pos];
  users.erase(users.begin()+pos);
  cout<<"Leaving disconnect"<<endl;
  return 0;
}

int dclient::addNewUser(string newguy,string avatar)
{
  int pos=-1; //position in vectors

  pos = SearchUsers(newguy);

  if(pos != -1){
    if(users[pos]->getName() == newguy){
     cout<<"Already have user: "<<users[pos]->getName().c_str()<<endl;
     return -1;
    }
  }

  duser* newUser = new duser(newguy,avatar);
  users.push_back(newUser);

  //Connect Remote Shared Mem
  
  newguy = "user-" + newguy;

  if(ClientManager->connectRemoteSharedMem(newguy.c_str(),host) == -1){
      cout<<"Could not connect shm for the newguy"<<endl;
      return -1;
  }
  
  return 0;
}

///////////////////////////////////////////
//    Returns where in the clientNameList vector
//    a client is or the closet location
///////////////////////////////////////////

int dclient::SearchUsers(string user)
{
  for(unsigned int i=0;i<users.size();i++)
    if(users[i]->getName() == user) return i;

  return -1;
/*
  int size = clientNameList.size();
  int low = 0;
  int high = size;
  string temp;
  int mid = (high+low)/2;

  for(mid = (high+low)/2;low<=high;){
    temp = clientNameList[mid];
    if(temp == user)
      return mid;
    if(user > temp){
        low = mid+1;
    }
    else
      high = mid-1;
  }

  return mid;  
*/
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new dclient(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
