#include "dclient.h"
#include "firepower.h"
#include <Performer/pr/pfTexture.h>
#include <dtk/dtkDSO_loader.h>

using namespace std;

#define LASER_KEY XK_L
#define SCUD_KEY XK_F12 
#define BOMB_KEY XK_F11

class firepower : public dtkAugment
{
  public:
    firepower(dpf *app);
    ~firepower(void);
    
    int preConfig(void);
    int postConfig(void);
    int postFrame(void);

  private:

    dpf* app;
    dtkTime Timer;
    long double timeCounter;
    long double lastLaser;
    long double lastBomb;
    long double lastScud;
    dtkSharedMem* myPos;
    string compName;

    void* arena;  
    pfGeode* laserGeode;
    pfGeode* bombGeode;
    pfNode* scudModel;
    
    Messenger* messenger;
    MessageParser* parser;
    KeyCode laser_key;        //  Key code to shoot lasers 
    KeyCode scud_key;         //  Key code to launch a scud
    KeyCode bomb_key;         //  Key code to drop bombs
    dtkInXKeyboard *keyboard; //  Keyboard to read from 
    dtkDequeuer *dq;          //  Dequeuer for the keyboard
    dclient* client;
    vector<duser*> users;
    vector<ammunition*> projectiles;

    int checkMessages();
    int checkUsers();
    int searchUsers(string name);
    int createBomb(float pos[]);
    int createLaser(float pos[]);
    int createScud(float pos[]);
    int updateProjectiles();
    int checkShoot();
    int createLaserGeode();
    int createBombGeode();
    int createScudModel();
    int addCrossHair();
};

// Constructor

firepower::firepower(dpf *app_in) : dtkAugment("firepower")
{
  setDescription("shoots firepowers");
  app = app_in;
  messenger = new Messenger("genMessage");
  parser = new MessageParser("genMessage");
  keyboard = NULL;
  dq = NULL;
  myPos = NULL;
  lastLaser = lastBomb = lastScud = 0;

  validate();
}

firepower::~firepower(void)
{
  delete messenger;
  delete parser;
}

int firepower::preConfig(void)
{
  //  Start the dequeuer
  dq = new dtkDequeuer(app->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;
  
  return DTK_REMOVE_CALLBACK;
}

int firepower::postConfig(void)
{
  char buf[128];
  gethostname(buf,128);
  compName = buf;
  messenger->setTo("everyone");
  messenger->setFrom(compName);
  messenger->setMessage("shoot");

  client = (dclient*) app->check("dclient");
  arena = pfGetSharedArena();
  Timer.delta();

  // Create the laser geode
  createLaserGeode();

  // Create scud
  createScudModel(); 

  // Create bomb geode
  createBombGeode();

  //  addCrossHair;
  addCrossHair();

  //  Get the keyboard
  keyboard = (dtkInXKeyboard *) app->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard){
    cout<<"couldn't get keyboard"<<endl; 
    return DTK_CALLBACK_ERROR;
  }
  keyboard->queue();

  //  Get the X key codes for shooting a laser
  if(!(laser_key = keyboard->getKeyCode(LASER_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"firepower::postConfig() failed" 
                            " to get x key code for LASER_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }

  //  Get the X key codes for shooting the map
  if(!(scud_key = keyboard->getKeyCode(SCUD_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"firepower::postConfig() failed" 
                            " to get x key code for SCUD_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }
  
  //  Get the X key codes for shooting the map
  if(!(bomb_key = keyboard->getKeyCode(BOMB_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"firepower::postConfig() failed" 
                            " to get x key code for BOMB_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }
 
  myPos = new dtkSharedMem(6 * sizeof(float), ("user-" + compName).c_str());
  if (!myPos || myPos->isInvalid() || myPos->getSize() < (6*sizeof(float))) {
    printf ("ERROR: could not get shared mem for the local machine\n");
  }
  
  return DTK_CONTINUE;
}

int firepower::addCrossHair()
{
 //  Load the 2D map in front of the user
  void* arena = pfGetSharedArena();

  //  Create a geoset with one polygon
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType (PFGS_QUADS);
  gSet->setNumPrims(1);

  pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

  //  Create a polygon to draw the map on
  verts[0].set(0,-1/8.f, -1/8.f);
  verts[1].set(0,1/8.f, -1/8.f);
  verts[2].set(0,1/8.f, 1/8.f);
  verts[3].set(0,-1/8.f, 1/8.f);
  gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

  /*
  //  Make the map slightly transparent
  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 1, 1, .8);
  gSet->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, color, NULL);
  */

  //  Set the texture to cover the entire polygon
  pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
  textcoords[0].set (0,0);
  textcoords[1].set (1, 0);
  textcoords[2].set (1, 1);
  textcoords[3].set (0, 1);

  gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

  pfGeode* eode = new pfGeode();
  eode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  gState->setMode(PFSTATE_TRANSPARENCY, PFTR_ON);

  //  Load the map texture onto the polygon
  pfTexture* tex = new pfTexture();

  pfFilePath("../:./models:../models:"DPF_DATA_DIR);

  int ret = tex->loadFile("crosshair.rgb");
  if (ret == 0) {
    return -1;
  }

  gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gState->setAttr(PFSTATE_TEXTURE, tex);
  
  gSet->setGState(gState);
  pfDCS* cross = new pfDCS;
  cross->setRot(90,0,0);
  cross->setTrans(0,2,0);

  cross->addChild(eode);
  
  app->display()->scene()->addChild(cross);
  
  return 0;
}

int firepower::createLaserGeode()
{
  pfGeoSet* gSet = pfdNewCylinder(100,arena);

  laserGeode = new pfGeode();
  laserGeode->ref();
  laserGeode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);

  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);

  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 0, 0, 1);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  gSet->setGState(gState);
  app->display()->addGhostNode(laserGeode);

  return 0;
}

int firepower::createBombGeode()
{
  pfGeoSet* gSet = pfdNewCylinder(100,arena);

  bombGeode = new pfGeode();
  bombGeode->ref();
  bombGeode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);

  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);

  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(0, 0, 0, 1);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  gSet->setGState(gState);
  app->display()->addGhostNode(bombGeode);

  return 0;
}

int firepower::createScudModel()
{
  pfFilePath(".:./models:../models:"DPF_DATA_DIR);

  scudModel = pfdLoadFile("scud.pfb");
  app->display()->addGhostNode(scudModel);

  return 0;
}

int firepower::checkUsers()
{
  users = client->getUsers();
  return 0;
}

int firepower::searchUsers(string name)
{
  for(unsigned int i=0; i<users.size();i++){
    if(users[i]->getName() == name) return i;
  }

  return -1;
}

int firepower::checkMessages()
{
  if(parser->NextMessage() == -1) return -1;

  if(parser->getTo() != "everyone") return -1; //Only messages addressed to everyone
                                             // will be processed.
  
  if(parser->getFrom() == compName) return -1;

  string message = parser->getMessage();

  switch(message[0]){
    case 'l':
      if (message == "laser"){
        string whoShot = parser->getFrom();
        int pos = searchUsers(whoShot); 
        if(pos == -1) return -1;
        float shooterPos[6];
        users[pos]->getPosition(shooterPos);
        createLaser(shooterPos);
      }
      else return -1;
      break;
    case 's':
      if (message == "scud"){
        string whoShot = parser->getFrom();
        int pos = searchUsers(whoShot); 
        if(pos == -1) return -1;
        float shooterPos[6];
        users[pos]->getPosition(shooterPos);
        createScud(shooterPos);
      }
      else return -1;
      break;
    case 'b':
      if (message == "bomb"){
        string whoShot = parser->getFrom();
        int pos = searchUsers(whoShot); 
        if(pos == -1) return -1;
        float shooterPos[6];
        users[pos]->getPosition(shooterPos);
        createBomb(shooterPos);
      }
      else return -1;
      break;
    default:
      return -1;
      break;
  }

  return 0;
}

int firepower::createLaser(float pos[])
{  
  laser* newLaser = new laser(pos,Timer.get(),laserGeode);
  app->display()->world()->addChild(newLaser->getrotDCS());
  projectiles.push_back(newLaser);

  return 0;
}

int firepower::createBomb(float pos[])
{  
  bomb* newBomb = new bomb(pos,Timer.get(),bombGeode);
  app->display()->world()->addChild(newBomb->getrotDCS());
  projectiles.push_back(newBomb);

  return 0;
}

int firepower::createScud(float pos[])
{
  if(scudModel==NULL) cout<<"NULL"<<endl;
  scud* newScud = new scud(pos,Timer.get(),scudModel);
  app->display()->world()->addChild(newScud->getrotDCS()); 
  projectiles.push_back(newScud);
  
  return 0;
}

int firepower::updateProjectiles()
{
  for(unsigned int i=0;i<projectiles.size();i++){
    if (projectiles[i]->getKillTime() < Timer.get()) {
      pfNode* ammo2Bkilled = (pfNode*) projectiles[i]->getrotDCS();
      app->display()->world()->removeChild(ammo2Bkilled);
      projectiles.erase(projectiles.begin() +i);
      i--;
      continue;
    }
    projectiles[i]->update(Timer.get());
  }
  
  return 0;
} 

int firepower::checkShoot()
{
  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  //  Check for keyboard events
  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    //  shoot a laser
    if (pressed && keyCode == laser_key) {
      // Limit the laser fire to 1 per second.
      if(lastLaser + .1 < Timer.get()){
        lastLaser = Timer.get();
        messenger->setMessage("laser");
        messenger->SendMessage();
        float pos[6];
        myPos->read(pos);
        createLaser(pos);
      }
    }

    // drop bomb 
    else if (pressed && keyCode == bomb_key){
      // Limit the bomb dropping to 1 per second.
      if(lastBomb + .1 < Timer.get()){
        lastBomb = Timer.get();
        messenger->setMessage("bomb");
        messenger->SendMessage();
        float pos[6];
        myPos->read(pos);
        createBomb(pos);
      }
    }
    
    // launch scud
    else if (pressed && keyCode == scud_key){
      // Limit the scud fire to one per 1 sec.
      if(lastScud + 1 < Timer.get()){
        lastScud = Timer.get();
        messenger->setMessage("scud");
        messenger->SendMessage();
        float pos[6];
        myPos->read(pos);
        createScud(pos);
      }
    }
    
  }
  return 0;
}

int firepower::postFrame(void)
{
  checkUsers();
  checkMessages();
  updateProjectiles();
  checkShoot();
  return DTK_CONTINUE;
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new firepower(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}

