// Firepower.h contains all the types of weaponry

#ifndef FIREPOWER_H
#define FIREPOWER_H

#include <dpf.h>
#include <cmath>

using namespace std;

///////////////////////////////////////////////////////////
//  Ammunition is something to be fired, dropped,
//  or launched from an avatar
///////////////////////////////////////////////////////////

class ammunition
{
  public:
    ammunition(float pos[6], long double p_createTime);
    pfDCS* getfirepowerDCS() {return firepowerDCS;}
    pfDCS* getrotDCS() {return rotDCS;}
    long double getKillTime() {return killTime;}
    virtual int update(long double currTime)=0;

  protected:
    pfDCS* firepowerDCS;
    pfDCS* rotDCS;
    long double createTime;
    long double killTime;
    double speed;
    float originalPos[6];

};

ammunition::ammunition(float pos[6],long double p_createTime)
{
  rotDCS = new pfDCS;

  rotDCS->setRot(pos[3],pos[4],pos[5]);  
  rotDCS->setTrans(pos[0],pos[1],pos[2]);
  for(int i=0;i<6;i++)
    originalPos[i] = pos[i];

  firepowerDCS = new pfDCS;

  createTime = p_createTime;
}

///////////////////////////////////////////////////////////
//  Bomb is a free falling ammuntion
///////////////////////////////////////////////////////////

class bomb : public ammunition
{
  public:
    bomb(float pos[], long double createTime, pfGeode* bombGeode);

  private:
    int update(long double currTime); 
};

bomb::bomb(float pos[], long double createTime, pfGeode* bombGeode)
  : ammunition(pos,createTime)
{
  firepowerDCS->setRot(0,-90,0);
  firepowerDCS->setTrans(0,0,-4);
  firepowerDCS->setScale(.2,.2,2);

  firepowerDCS->addChild(bombGeode);
  rotDCS->addChild(firepowerDCS);

  killTime = createTime+15;
  speed = 0;

}

int bomb::update(long double currTime)
{
  speed = pow(9.8*(currTime-createTime),2);
  if(speed > 89)
    speed = 89;

  float distance = float((currTime-createTime) * speed); 

  float newZ = originalPos[2] - distance;

  rotDCS->setTrans(originalPos[0],originalPos[1],newZ);
  
  return 0;
}

///////////////////////////////////////////////////////////
//  Laser is a shot ammuntion
///////////////////////////////////////////////////////////

class laser : public ammunition
{
  public:
    laser(float pos[], long double createTime, pfGeode* bombGeode);
    int update(long double currTime);

  private:
    pfMatrix position;
    pfCoord coord;
    long double lastTime;

};

laser::laser(float pos[], long double createTime, pfGeode* laserGeode)
  : ammunition(pos,createTime)
{
  rotDCS->setRot(pos[3],pos[4],pos[5]);
  rotDCS->setTrans(pos[0],pos[1],pos[2]);

  firepowerDCS->setRot(0,-90,0);
  firepowerDCS->setTrans(0,4,0);
  firepowerDCS->setScale(.2,.2,2);

  firepowerDCS->addChild(laserGeode);
  rotDCS->addChild(firepowerDCS);

  killTime = createTime+10;
  lastTime = createTime;

  speed = 300;
}

int laser::update(long double currTime)
{
  float distance = (float)((currTime-lastTime) * speed); 

  firepowerDCS->getMat(position);
  position.getOrthoCoord(&coord);
  float newY = coord.xyz[1] + distance;
  firepowerDCS->setTrans(coord.xyz[0],newY,coord.xyz[2]);

  lastTime = currTime;
  
  return 0;
}

///////////////////////////////////////////////////////////
//  scud is a launched ammunition
///////////////////////////////////////////////////////////

class scud : public ammunition 
{
  public:
    scud(float pos[], long double createTime, pfNode* scudModel); 
    int update(long double currTime); 

  private:
    double speed;

};

scud::scud(float pos[], long double createTime, pfNode* scudModel) : ammunition(pos,createTime) 
{
  firepowerDCS->setRot(0,90,0);
  firepowerDCS->setTrans(0,0,4);
  firepowerDCS->setScale(.5);

  if(scudModel==NULL) cout<<"NULL"<<endl;
  firepowerDCS->addChild(scudModel);
  rotDCS->addChild(firepowerDCS);
  rotDCS->setRot(0,0,0);
  speed = 0;

  killTime = createTime+15;
}

int scud::update(long double currTime)
{
  //For right now the scud missle acceleration is the same as gravity.
  //but the terminal velocity is different.
  speed = pow(9.8*(currTime-createTime),2);
  if(speed > 1000)
    speed = 1000;
  
  float distance = float((currTime-createTime) * speed); 

  float newZ = originalPos[2] + distance;

  rotDCS->setTrans(originalPos[0],originalPos[1],newZ);
  
  return 0;
}

#endif
