#include <dpf.h>
#include <Performer/pr/pfTexture.h>
#include <dtk/dtkDSO_loader.h>
#include "dclient.h"

using namespace std;

#define SHOOT_KEY XK_L

class laser : public dtkAugment
{
  public:
    laser(dpf *app);
    ~laser(void);
    
    int preConfig(void);
    int postConfig(void);
    int postFrame(void);

  private:

    dpf* app;
    dtkTime Timer;
    long double timeCounter;
    long double laserDeleter;
    int speed;
    dtkSharedMem* myPos;
    string compName;

    void* arena;  
    pfGeode* geode;
    
    Messenger* messenger;
    MessageParser* parser;
    KeyCode shoot_key;       //  Key code to toggle map
    dtkInXKeyboard *keyboard; //  Keyboard to read from 
    dtkDequeuer *dq;          //  Dequeuer for the keyboard
    dclient* client;
    vector<duser*> users;
    vector<pfDCS*> Lasers;
    vector<long double> deleteTimes;

    int checkMessages();
    int checkUsers();
    int searchUsers(string name);
    int createLaser(float pos[]);
    int updateLasers();
    int checkShoot();
    int createGeode();
    int addCrossHair();
};

laser::laser(dpf *app_in) : dtkAugment("laser")
{
  setDescription("shoots lasers");
  app = app_in;
  messenger = new Messenger("genMessage");
  parser = new MessageParser("genMessage");
  keyboard = NULL;
  dq = NULL;
  speed = 300;
  myPos = NULL;

  validate();
}

laser::~laser(void)
{
  delete messenger;
  delete parser;
}

int laser::preConfig(void)
{
  //  Start the dequeuer
  dq = new dtkDequeuer(app->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;
  
  return DTK_REMOVE_CALLBACK;
}

int laser::postConfig(void)
{
  char buf[128];
  gethostname(buf,128);
  compName = buf;
  messenger->setTo("everyone");
  messenger->setFrom(compName);
  messenger->setMessage("shoot");

  client = (dclient*) app->check("dclient");
  arena = pfGetSharedArena();
  Timer.delta();

  // Create the geode
  createGeode();

  //  addCrossHair;
  addCrossHair();

  //  Get the keyboard
  keyboard = (dtkInXKeyboard *) app->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard){
    cout<<"couldn't get keyboard"<<endl; 
    return DTK_CALLBACK_ERROR;
  }
  keyboard->queue();

  //  Get the X key codes for toggling the map
  if(!(shoot_key = keyboard->getKeyCode(SHOOT_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"laser::postConfig() failed" 
                            " to get x key code for TOGGLE_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }

  myPos = new dtkSharedMem(6 * sizeof(float), ("user-" + compName).c_str());
  if (!myPos || myPos->isInvalid() || myPos->getSize() < (6*sizeof(float))) {
    printf ("ERROR: could not get shared mem for the local machine\n");
  }
  
  return DTK_CONTINUE;
}

int laser::addCrossHair()
{
 //  Load the 2D map in front of the user
  void* arena = pfGetSharedArena();

  //  Create a geoset with one polygon
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType (PFGS_QUADS);
  gSet->setNumPrims(1);

  pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

  //  Create a polygon to draw the map on
  verts[0].set(0,-1/8.f, -1/8.f);
  verts[1].set(0,1/8.f, -1/8.f);
  verts[2].set(0,1/8.f, 1/8.f);
  verts[3].set(0,-1/8.f, 1/8.f);
  gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

  /*
  //  Make the map slightly transparent
  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 1, 1, .8);
  gSet->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, color, NULL);
  */

  //  Set the texture to cover the entire polygon
  pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
  textcoords[0].set (0,0);
  textcoords[1].set (1, 0);
  textcoords[2].set (1, 1);
  textcoords[3].set (0, 1);

  gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

  pfGeode* eode = new pfGeode();
  eode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  gState->setMode(PFSTATE_TRANSPARENCY, PFTR_ON);

  //  Load the map texture onto the polygon
  pfTexture* tex = new pfTexture();

  pfFilePath("../:./models:../models:"DPF_DATA_DIR);

  int ret = tex->loadFile("crosshair.rgb");
  if (ret == 0) {
    return -1;
  }

  gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gState->setAttr(PFSTATE_TEXTURE, tex);
  
  gSet->setGState(gState);
  pfDCS* cross = new pfDCS;
  cross->setRot(90,0,0);
  cross->setTrans(0,2,0);

  cross->addChild(eode);
  
  app->display()->scene()->addChild(cross);
  
  return 0;
}

int laser::createGeode()
{
  pfGeoSet* gSet = pfdNewCylinder(100,arena);

  geode = new pfGeode();
  geode->ref();
  geode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);

  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);

  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 0, 0, 1);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  gSet->setGState(gState);

  return 0;
}

int laser::checkUsers()
{
  users = client->getUsers();
  return 0;
}

int laser::searchUsers(string name)
{
  for(unsigned int i=0; i<users.size();i++){
    if(users[i]->getName() == name) return i;
  }

  return -1;
}

int laser::checkMessages()
{
  if(parser->NextMessage() == -1) return -1;

  if(parser->getTo() != "everyone") return -1; //Only messages addressed to everyone
                                             // will be processed.
  
  if(parser->getFrom() == compName) return -1;

  string message = parser->getMessage();

  if(message != "shoot") return -1;

  string whoShot = parser->getFrom();
 
  int pos = searchUsers(whoShot); 

  if(pos == -1) return -1;
  
  float shooterPos[6];
  users[pos]->getPosition(shooterPos);

  createLaser(shooterPos);

  return 0;
}

int laser::createLaser(float pos[])
{  
  pfDCS* rotDCS = new pfDCS;

  rotDCS->setRot(pos[3],pos[4],pos[5]);  
  rotDCS->setTrans(pos[0],pos[1],pos[2]);

  pfDCS* laserDCS = new pfDCS;

//  laserDCS->setRot(pos[3],pos[4]-90,pos[5]);
  laserDCS->setRot(0,-90,0);
//  laserDCS->setTrans(pos[0],pos[1],pos[2]);
  laserDCS->setTrans(0,4,0);
  laserDCS->setScale(.2,.2,2);
//  laserDCS->setScale(8,8,24);

  laserDCS->addChild(geode);
  rotDCS->addChild(laserDCS);
  app->display()->world()->addChild(rotDCS);
  Lasers.push_back(laserDCS);
  long double killtime = Timer.get() + 10;
  deleteTimes.push_back(killtime);

  return 0;
}

int laser::updateLasers()
{
  pfMatrix position;
  pfCoord coord;
  float stinky;
  long double oldTime = timeCounter;
  timeCounter = Timer.get();
  long double distance = timeCounter - oldTime;
  distance = distance * speed; 

  for(unsigned int i=0;i<Lasers.size();i++){
    Lasers[i]->getMat(position);
    position.getOrthoCoord(&coord);
    stinky = coord.xyz[1] + distance;
    Lasers[i]->setTrans(coord.xyz[0],stinky,coord.xyz[2]);
  }
  
  int pos = Lasers.size() -1;

  if(pos == -1) return 0;

  if(deleteTimes[pos] < Timer.get()){
    pfNode* laser = (pfNode*) Lasers[pos];
    app->display()->world()->removeChild(laser->getParent(0));
    Lasers[pos] = NULL;
    Lasers.erase(Lasers.end() -1);
    deleteTimes.erase(deleteTimes.end() -1);
  }

  return 0;
} 

int laser::checkShoot()
{
  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  //  Check for keyboard events
  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    //  Toggle the map position
    if (pressed && keyCode == shoot_key) {
      messenger->SendMessage();
      float pos[6];
      myPos->read(pos);
      createLaser(pos);
    }
    
  }
  return 0;
}

int laser::postFrame(void)
{
  checkUsers();
  checkMessages();
  updateLasers();
  checkShoot();
  return DTK_CONTINUE;
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new laser(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}

