#include <dpf.h>
#include <Performer/pr/pfTexture.h>
#include <dtk/dtkDSO_loader.h>
#include <vector>
#include <iostream>

using namespace std;

const int GRIDX = 3; 
const int GRIDY = 3;
const int RADIUS = 2000;
const int HEIGHT = 2000;
const int MapX = 22000;
const int MapY = 27500;

class probes : public dtkAugment
{
  public:
    probes(dpf *app);
    
    int postConfig(void);
    int postFrame(void);

  private:

    dpf* app;
    dtkTime Timer;

    pfNode* probeModel;
    vector<pfDCS*> Probes;

};

// Constructor

probes::probes(dpf *app_in) : dtkAugment("probes")
{
  setDescription("make a bunch of probes as stationary objects");
  app = app_in;

  validate();
}

int probes::postConfig(void)
{
  pfFilePath(".:./models:../models:"DPF_DATA_DIR);

  probeModel = pfdLoadFile("probe.pfb");
  app->display()->addGhostNode(probeModel);
  
  int X = MapX;
  int Y = MapY - 2000;

  double distApartX = X/GRIDX;
  double distApartY = Y/GRIDY;
  double distX = distApartX/2;
  double distY = distApartY/2;

  pfDCS* probeDCS;

  for(int x=0;x<GRIDX;x++){
    for(int y=0;y<GRIDY;y++){
      probeDCS = new pfDCS();
      probeDCS->setRot(0,0,0);
      probeDCS->setTrans(distX,distY,HEIGHT);
      probeDCS->addChild(probeModel);
      app->display()->world()->addChild(probeDCS);
      Probes.push_back(probeDCS);

      distY += distApartY;
    }
    distY = distApartY;
    distX += distApartX;
  }

  return 0;
}

int probes::postFrame(void)
{
  bool good2go = false;
  dpfIntersection* intersect;
  intersect = new dpfIntersection();
  intersect->enableSegment(0);
  intersect->setSegmentLength(0, 30000);
  intersect->setSegmentDir(0, 0, 0, -1);
  intersect->setRoot(app->display()->world());

  pfMatrix pos;
  pfCoord coord;

  for(unsigned int i=0;i<Probes.size();i++){
    Probes[i]->getMat(pos);
    pos.getOrthoCoord(&coord);
    intersect->setSegmentPos(0,coord.xyz[0],coord.xyz[1],20000);
    intersect->doIntersection();

    if(intersect->segmentHit(0)){
      pfVec3 point;
      intersect->getXFormPoint(0,&point);
      Probes[i]->setTrans(coord.xyz[0], coord.xyz[1], point[2]+30);
      good2go = true;
    }
  }

  if(good2go)
      return DTK_REMOVE_CALLBACK;
  else
    return DTK_CONTINUE;
  
}

//////////////////////////////////////////////
//
//  Generic Stuff
//
//////////////////////////////////////////////

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new probes (app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}

