//
// DIVERSE graphics interface -- Performer ---- version 2.0.7
// released: Fri May 24 16:57:38 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//
#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pr/pfTexture.h>
#include "radar.h"

using namespace std;

#define SHM_PREFIX "user-"
//  Key to toggle map
#define TOGGLE_KEY XK_E
#define MOVE_KEY XK_F1

//  Shows a 2D radar with users' positions plotted on the map.
//  
//  Environment variables to set:
//    DPF_MAP_FILE:  The name of the image file to read from (string)
//    DPF_MAP_SCALE:  The scale of the map (float).  This is the width (east-west
//                    distance) of the model that the map represents, in model
//                    units.
//    DPF_MAP_ASPECT:  Ratio of width/height for the map (float)

// create object and set description.
// be sure to validate it!
Radar::Radar(dpf *app) :
  dtkAugment("radar") 
{
  App = app;
  getEnvVars();
  validate();

  keyboard = NULL;
  dq = NULL;
  dqButton = NULL;
  radarSwitcher = NULL;
  button = NULL;
  group2D = NULL;
} 

// destructor
Radar::~Radar(void) {

}

//  Get the environment variables
void Radar::getEnvVars() {

  //  Get the hostname
  char buf[128];
  gethostname(buf, 128);
  myName = buf;

  int good;

  //  Get the map file name
  mapFile = getenv("DPF_MAP_FILE");
  if (mapFile == NULL) {
    printf ("WARNING:  DPF_MAP_FILE not set\n");
  }

  //  Get the circle radar file
  circleRadarFile = getenv("DPF_CIRCLE_RADAR_FILE");
  if(circleRadarFile == NULL){
    cout<<"WARNING DPF_CIRCLE_RADAR_FILE not set"<<endl;
  }

  //  Get the scale
  char* scale = getenv("DPF_MAP_SCALE");
  if (scale == NULL) {
    printf ("WARNING: DPF_MAP_SCALE not set\n");
    mapScale = 22000;  // Default to Ft. Irwin scae
  }
  else {
    good = sscanf(scale, "%f", &mapScale);
    if (!good) {
      printf ("WARNING: DPF_MAP_SCALE not a valid float\n");
      mapScale = 22000;  // Default to Ft. Irwin scae
    }
  }

  //  Get the aspect ratio
  char* aspect = getenv("DPF_MAP_ASPECT");
  if (aspect == NULL) {
    printf ("WARNING: DPF_MAP_ASPECT not set\n");
    mapAspect = 0.8;  // Default to Ft. Irwin scae
  }
  else {
    good = sscanf(aspect, "%f", &mapAspect);
    if (!good) {
      printf ("WARNING: DPF_MAP_ASPECT not a valid float\n");
      mapAspect = 0.8;  // Default to Ft. Irwin scae
    }
  }

}

//  Add a mini-map to the scenegraph
void Radar::addMiniMap() {
  //  Load the 2D map in front of the user
  void* arena = pfGetSharedArena();

  //  Create a geoset with one polygon
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType (PFGS_QUADS);
  gSet->setNumPrims(1);

  pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

  //  Create a polygon to draw the map on
  float w = 1;
  float h = w/mapAspect;
  verts[0].set(0, 0, 0);
  verts[1].set(w, 0, 0);
  verts[2].set(w, h, 0);
  verts[3].set(0, h, 0);
  gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

  //  Make the map slightly transparent
  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 1, 1, .8);
  gSet->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, color, NULL);

  //  Set the texture to cover the entire polygon
  pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
  textcoords[0].set (0,0);
  textcoords[1].set (1, 0);
  textcoords[2].set (1, 1);
  textcoords[3].set (0, 1);

  gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

  pfGeode* geode = new pfGeode();
  geode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  gState->setMode(PFSTATE_TRANSPARENCY, PFTR_ON);

  //  Load the map texture onto the polygon
  pfTexture* tex = new pfTexture();

  int ret = tex->loadFile(mapFile);
  if (ret == 0) {
    printf ("could not load file the map file %s\n", mapFile);
    return;
  }

  gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gState->setAttr(PFSTATE_TEXTURE, tex);
  
  gSet->setGState(gState);
  
  group2D = new pfGroup();
  group2D->addChild(geode);

  radarSwitcher->addChild(group2D);
}

int Radar::addCircleRadar()
{
 //  Load the 2D map in front of the user
  void* arena = pfGetSharedArena();

  //  Create a geoset with one polygon
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType (PFGS_QUADS);
  gSet->setNumPrims(1);

  pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

  //  Create a polygon to draw the map on
  float w = 1;
  float h = w;
  verts[0].set(-w/2.f, -h/2.f, 0);
  verts[1].set(w/2.f, -h/2.f, 0);
  verts[2].set(w/2.f, h/2.f, 0);
  verts[3].set(-w/2.f, h/2.f, 0);
  gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

  //  Make the map slightly transparent
  pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
  color->set(1, 1, 1, .8);
  gSet->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, color, NULL);

  //  Set the texture to cover the entire polygon
  pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
  textcoords[0].set (0,0);
  textcoords[1].set (1, 0);
  textcoords[2].set (1, 1);
  textcoords[3].set (0, 1);

  gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

  pfGeode* geode = new pfGeode();
  geode->addGSet(gSet);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);
  gState->setMode(PFSTATE_TRANSPARENCY, PFTR_ON);

  //  Load the map texture onto the polygon
  pfTexture* tex = new pfTexture();

  int ret = tex->loadFile(circleRadarFile);
  if (ret == 0) {
    printf ("could not load file the map file %s\n", circleRadarFile);
    return -1;
  }

  gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gState->setAttr(PFSTATE_TEXTURE, tex);
  
  gSet->setGState(gState);
  
  radarSwitcher->addChild(geode);

  return 0;
}

//  Get the list of users from the client
vector<duser*> Radar::getUsers() {
  return client->getUsers();
}

//  Check for users connecting and disconnecting
int Radar::checkUsers() {

  vector<duser*> userNames = getUsers();

  if(userNames.size() < users.size()){
    remove(userNames);
    return 0;
  }

  //If this is the first time this has run the avatars will not be set yet
  //so you must wait another frame or two until the avatars are set.

  if(userNames.size() == 0) return -1;
  if(userNames[0]->getAvatar() == "" || userNames[0]->getAvatar() == "BOGUS")
    return -1;
  
  //  Add user records for users that have just connected;
  unsigned int c;

  pfFilePath(".:./models:../models:"DPF_DATA_DIR);

  for (c=0; c<userNames.size(); c++) {

    if (userIndex(userNames[c]->getName()) < 0 ) {

      printf ("adding user %s\n", userNames[c]->getName().c_str());

      pfSwitch* Holder = new pfSwitch;
      //  Load the field of view marker
      pfDCS* marker = new pfDCS();
      pfNode* node = pfdLoadFile("fov.pfb");
      App->display()->addGhostNode(node);
      marker->setScale(0.003);
      marker->addChild(node);
      //marker->setRot(-45, 90, 0);
      Holder->addChild(marker); //At index 0;
      rotDCS->addChild(Holder);
      
      //  Load the avatar
      pfDCS* body = new pfDCS();
      if(userNames[c]->getName() == myName) 
        body = NULL;
      else{
        cout<<"userNames[c]->getAvatar()"<<endl;
        pfNode* model = pfdLoadFile(userNames[c]->getAvatar().c_str());
        App->display()->addGhostNode(model);
        body->setScale(.5);
        body->setRot(0,0,0);
        body->addChild(model);
        Holder->addChild(body);
        App->display()->world()->addChild(body);
      }
      userNames[c]->initDcs(Holder,mapScale);
      body = NULL;
      marker = NULL;

      users.push_back(userNames[c]);
    }

  }
  return 0;
}

int Radar::remove(vector<duser*> UserList)
{
  bool gotHim = false;
  for(unsigned int i=0;i<UserList.size();i++){
    if(users[i] != UserList[i]){
      gotHim = true;
      users.erase(users.begin() + i);
      i--;
    }
  }
  if(!gotHim){
    users.erase(users.end() -1);
  }
  return 0;
}
//  Index in the user list of the given user, -1 if not found
int Radar::userIndex(string name) {
  unsigned int c;

  for (c=0; c<users.size(); c++) {
    if (users[c]->getName() == name)
      return c;
  }

  return -1;

}

//  Update the positions of the markers on the map
void Radar::updatePositions() {

  //  Read positions of the users from shared mem and update the dcs nodes
  unsigned int c;

  float pos[6];
  if(radarMode == circleRadar){ // radar == Circle){
    myPos->read(pos);

    if(myPos->differentByteOrder()){
      int j;
      for(j=0;j<6;j++)
    		dtk_swapBytes(&pos[j],sizeof(float));
    }

    rotDCS->setRot(-pos[3],0,0); //Set the map rotation to your heading.
    for (c=0; c<users.size(); c++) {
      users[c]->updateAvatarMarker(pos);
    }
  }
  else{ //radarMode == miniMap
    for (c=0; c<users.size(); c++) {
      users[c]->updateAvatarMarker();
    }
  }

}

// called only once by dpf, after pfInit() and dpfDisplay:preConfig()
// but before pfConfig()
int Radar::preConfig(void) 
{
  //  Start the dequeuer
  dq = new dtkDequeuer(App->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;

  //  Initialize the object to read from the wand buttons
  button = static_cast<dtkInButton *> (App->get("buttons", DTKINBUTTON_TYPE));

  if(!button) return DTK_CALLBACK_ERROR;

  if(button->number() < 4) {
    printf ("Need 4 buttons\n");
    return DTK_CALLBACK_ERROR;
  }

  button->queue();

  return DTK_REMOVE_CALLBACK ;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int Radar::postConfig(void) {

  //  Get the keyboard
  keyboard = (dtkInXKeyboard *) App->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard) 
    return DTK_CALLBACK_ERROR;
  keyboard->queue();

  //  Get the X key codes for toggling the map
  if(!(toggle_key = keyboard->getKeyCode(TOGGLE_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                            " to get x key code for TOGGLE_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }

  //  Get the X key codes for moving the map
  if(!(move_key = keyboard->getKeyCode(MOVE_KEY)))
  {
    dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                            " to get x key code for MOVE_KEY.\n");
    return DTK_CALLBACK_ERROR;
  }

  radarSwitcher = new pfSwitch();
  radarMode = miniMap;
  radarPos = Center;
  
  //  Set it to be in front of the user, in scene coordinates

  mapDCS = new pfDCS();
  mapDCS->setTrans(-1, 2.5, -1);
  mapDCS->setRot(0, 90, 0);

  addMiniMap();
  addCircleRadar();

  rotDCS = new pfDCS() ;
  radarSwitcher->setVal(0);
  rotDCS->addChild(radarSwitcher);
  rotDCS->setTrans(-.5,-.5,-.5);
  mapDCS->addChild(rotDCS) ;  
  
  App->display()->scene()->addChild(mapDCS);
  
  myPos = new dtkSharedMem(6 * sizeof(float), (SHM_PREFIX + myName).c_str());
  if (!myPos || myPos->isInvalid() || myPos->getSize() < (6*sizeof(float))) {
    printf ("ERROR: could not get shared mem for the local machine\n");
  }

  client = (dclient*) App->check("dclient");

  if (client == NULL) {
    printf ("could not get client\n");
    return DTK_CALLBACK_ERROR;
  }

  client->setMapGroup(rotDCS);

  return DTK_CONTINUE ; 
}

// called every display loop, as a pfFrame callback, before pfDraw()
// this return value causes it to only be called once
int Radar::preFrame(void) {
  return DTK_REMOVE_CALLBACK ; 
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int Radar::postFrame(void) 
{
  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  //  Check for keyboard events
  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    //  Toggle the map position
    if (pressed && keyCode == toggle_key) {

      if (radarMode == miniMap) {
        rotDCS->setTrans(0,0,0);
        radarMode = circleRadar;
        radarSwitcher->setVal(1);
      }
      else { // radarMode == circleRadar
        rotDCS->setRot(0,0,0);
        rotDCS->setTrans(-.5,-.5,-.5);
        radarMode = miniMap;
        radarSwitcher->setVal(0);
      }
    }

    //  Move the map
    if (pressed && keyCode == move_key) {
      if(radarPos == Right){
        mapDCS->setTrans(-1, 2.5, -1);
        mapDCS->setRot(0, 90, 0);
        //mapDCS->setRot(0, 90, 0);
        //mapDCS->setTrans(-.5, 2.5, -1);
        radarPos = Center;
      }
      else{ //(radarPos == Center){
        mapDCS->setRot(-90, 90, 0);
        mapDCS->setTrans(1.5, -.8, -.9);
        radarPos = Right;
      }
    }
  }

  if( checkUsers() == 0)
    updatePositions();

/*
  //  First frame flush any remaining button events
  if (firstFrame) {
    firstFrame = 0;

    dtkRecord_event *event;
    while((event = dqButton->getNextEvent(button))) {
    }
  }
*/
  return DTK_CONTINUE ; 
}


//  Get the mini map group, called by selectPoints
pfGroup* Radar::getMap(){
  return group2D;
}

/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new Radar(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
