//  Data structures used for the radar program
#include <dpf.h>
#include <string>
#include <vector>
#include "duser.h"
#include "dclient.h"

#ifndef _RADAR_H
#define _RADAR_H
enum radarModeType {miniMap,circleRadar};
enum radarPosType {Center,Right};

class Radar : public dtkAugment {

public:

  Radar(dpf *app) ;
  ~Radar(void) ;

  int preConfig(void);
  int postConfig(void);
  int preFrame(void);
  int postFrame(void);

  //  Get the mini map group, called by selectPoints
  pfGroup* getMap();

  //  Return the radar mode
  radarModeType getMode() {return radarMode; }

private:

  //  **Functions:
  radarModeType radarMode;

  radarPosType radarPos;

  //  The Radar Switcher
  pfSwitch* radarSwitcher;

  //  Key code to toggle map
  KeyCode toggle_key;

  // Key code to move map
  KeyCode move_key;

  //  Keyboard to read from
  dtkInXKeyboard *keyboard;

  //  Dequeuer for the keyboard
  dtkDequeuer *dq;

  //  Dequeuer for the buttons
  dtkDequeuer *dqButton;
  
  //  Wand buttons
  dtkInButton *button;

  //  Get the environment variables
  void getEnvVars();

  //  Add the circle radar
  int addCircleRadar();

  //  Add the map to the scene graph
  void addMiniMap();

  //  Get the list of users from the client
  vector<duser*> getUsers();

  //  Check for users connecting and disconnecting
  int checkUsers();

  //  Remove a user
  int remove(vector<duser*> UserList);

  //  Index in the user list of the given user, -1 if not found
  int userIndex(string name);

  //  Update the positions of the markers on the map
  void updatePositions();

  //  **Variables:

  dpf* App;

  //  Pointer to the client program
  dclient* client;
  
  //  Name of the map image file
  char* mapFile;

  //  Name of circle radar file
  char* circleRadarFile;
  
  //  Scale of the map
  float mapScale;

  //  Aspect ratio of the map
  float mapAspect;

  //  A pfSwitch which holds both of the radars.
  pfSwitch* radars;
  //  DCS node for the map
  pfDCS* mapDCS;
  pfDCS* rotDCS;

  //  List of clients connected
  vector<duser*> users;

  //  Hostname of this machine
  string myName;

  //  Shared memory segment for my position
  dtkSharedMem* myPos;

  //  Group node for the 2d map
  pfGroup* group2D;
} ;

#endif
