//
// DIVERSE graphics interface -- Performer ---- version 2.0.5
// released: Sat May 18 10:54:26 EDT 2002
// The DIVERSE HOMPAGE is at: www.diverse.vt.edu
// 
// Copyright (C) 2000, 2001, 2002  John Kelso, Lance Arsenault and Ron Kriz
// 
// This software, DIVERSE graphics interface -- Performer (DPF),
// is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License (LGPL) as
// published by the Free Software Foundation; either version 2.1 of
// the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this software, in the top level source directory in
// a file named "COPYING"; if not, see it at:
// http://www.gnu.org/copyleft/lesser.html or write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//
#include <stdio.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>
#include <Performer/pr/pfTexture.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include "radar.h"

const char* mapTexture = "2Dmap.rgb";
const float scale = 22000.0;
#define MAXPOINTS 10000

//  Key to toggle map
#define TOGGLE_KEY XK_E

//  Key to select a point
#define SELECT_KEY XK_S

// example generic dtkAugment DSO

class SelectPoints : public dtkAugment {
public:

  SelectPoints(dpf *app) ;
  ~SelectPoints(void) ;
  int preConfig(void);
  int postConfig(void);
  int preFrame(void);
  int postFrame(void);

private:

  dtkColor color;
  int count;
  dpf* App;

  //  Node that has the mini-map under it
  pfDCS* mapDCS;

  //  Group node under the dcs
  pfGroup* mapGroup;
  
  //  DCS for the marker that shows your pointer position on the ful map
  pfDCS* markerDCS;

  //  "You are here" indicator on the mini map
  pfDCS* youAreHereDCS;

  //  Shared mem for intersection point
  dtkSharedMem* shm;

  //  Shared mem to write the two selected points
  dtkSharedMem* selectShm;

  //  File to read the path from
  char* file;

  //  Initialize objects in the scene graph (mini map, etc)
  void initSceneGraph();

  //  Show the mini map
  void addMiniMap();

  //  Construct a geode containing to hold the path
  void makeGeode (pfGroup* group, pfVec3*& vertices, pfGeode*& geode);

  //  Check to see if the plot file has been modified.  If it has been, read
  //  the new path from the file
  void checkModified();

  //  Read in a path from the playback file and populate vertices
  void readPath (pfVec3* vertices, float scale, int constantZ);

  //  True if the map is in front of the user, false if it is to the right
  //  Used to switch the map
  int inFront;

  //  Return true if the given point is on the mini-map
  int goodPoint(float* point);

  //  Check the keyboard for key presses
  void checkKey();

  //  Set the mini map to be in front of or to the right of the user
  void setFront();
  void setRight();

  //  Add a line to show the user's selection
  void addSelectLine();

  //  Handle a button/keyboard press to select a point on the map
  void doSelect(float point[]);

  //  Get the z value of the terrain at the given x,y
  float terrainZ(float x, float y);

  //  Points that the user has selected (in the coordinate system of the
  //  mini map)
  pfVec3* selectPoints;

  //  Show the selected points on the large map
  pfVec3* selectPointsLarge;

  //  Number of points selected so far
  int numSelected;

  //  Markers for the selected points on the mini map
  pfDCS* selectDCS[2];

  //  Markers for the selected points on the full map
  pfDCS* selectLargeDCS[2];

  //  Geode of the select segment on the full map
  pfGeode* selectLargeGeode;

  //  Vertices of the path on the large map
  pfVec3* largeVertices;

  //  Geode for the path plot on the large map
  pfGeode* largeGeode;

  //  Vertices of the path on the mini map
  pfVec3* miniVertices;

  //  Geode for the path plot on the mini map
  pfGeode* miniGeode;

  //  Number of vertices in the path (same for large and mini maps)
  int* numVertices;

  //  Key code to toggle map
  KeyCode toggle_key;

  //  Key code to select points
  KeyCode select_key;

  //  Keyboard to read from
  dtkInXKeyboard *keyboard;

  //  Dequeuer for the keyboard
  dtkDequeuer *dq;

  //  Dequeuer for the buttons
  dtkDequeuer *dqButton;

  //  Wand buttons
  dtkInButton *button;

  //  Stat record for the plot file
  struct stat fileStat;

  //  Modified time of the plot file when it was last read in
  time_t lastModified;

  //  Intersect object to plot the path above the ground
  dpfIntersection* intersect;

  //  Flag for the first frame
  int firstFrame;

  //  Nonzero if the radar dso was already loaded
  int radarLoaded;

  //  Pointer to the radar dso
  Radar* radar;
} ;

// create object and set description.
// be sure to validate it!
SelectPoints::SelectPoints(dpf *app) :
  dtkAugment("selectPoints"), color(stdout)
{
  setDescription("generic dtkAugment DSO") ;
  App = app;
  count = 0;
  
  //  Get the file name
  file = getenv("DPF_PLAYBACK_FILE");
  
  largeVertices = NULL;
  miniVertices = NULL;
  selectLargeGeode = NULL;
  largeGeode = NULL;
  miniGeode = NULL;
  numVertices = NULL;
  keyboard = NULL;
  button = NULL;
  dq = NULL;
  dqButton = NULL;
  selectPoints = NULL;
  intersect = NULL;
  mapGroup = NULL;
  radar = NULL;

  lastModified = 0;
  numSelected = 0;
  firstFrame = 1;

  validate();
} 

// destructor
SelectPoints::~SelectPoints(void) {
}

// called only once by dpf, after pfInit() and dpfDisplay:preConfig()
// but before pfConfig()
int SelectPoints::preConfig(void) {

  //  Start button the dequeuer
  dqButton = new dtkDequeuer(App->record());
  if(!dqButton || dqButton->isInvalid()) 
    return  DTK_CALLBACK_ERROR;

  //  Initialize the object to read from the wand buttons
  button = static_cast<dtkInButton *> (App->get("buttons", DTKINBUTTON_TYPE));

  if(!button) return DTK_CALLBACK_ERROR;

  if(button->number() < 4) {
    printf ("Need 4 buttons\n");
    return DTK_CALLBACK_ERROR;
  }

  button->queue();

  return DTK_REMOVE_CALLBACK ;
}

// called only once by dpf, after pfConfig() and dpfDisplay:postConfig()
// pfConfig()
int SelectPoints::postConfig(void) {

  //  Set up an Intersection object to find terrain collisions
  intersect = new dpfIntersection();
  intersect->enableSegment(0);
  intersect->setSegmentLength(0, 30000);
  intersect->setSegmentDir(0, 0, 0, -1);
  intersect->setRoot(App->display()->world());

    //  Get the keyboard
  keyboard = (dtkInXKeyboard *) App->get("xkeyboard", DTKINXKEYBOARD_TYPE);
  if(!keyboard) 
    return DTK_CALLBACK_ERROR;
  keyboard->queue();

  //  Get the X key codes for toggling the map
  if(!(toggle_key = keyboard->getKeyCode(TOGGLE_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                              " to get x key code for TOGGLE_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }

  
  //  Get the X key codes for selecting
  if(!(select_key = keyboard->getKeyCode(SELECT_KEY)))
    {
      dtkMsg.add(DTKMSG_ERROR,"SelectPoints::postConfig() failed" 
                              " to get x key code for SELECT_KEY.\n");
      return DTK_CALLBACK_ERROR;
    }

  //  Start the dequeuer
  dq = new dtkDequeuer(App->record());
  if(!dq || dq->isInvalid()) 
    return  DTK_CALLBACK_ERROR;

  //  Get the shared memory for the point that the user has selected
  shm = new dtkSharedMem (3 * sizeof(float), "localPoint");
  if (!shm || shm->isInvalid() || shm->getSize() < (3*sizeof(float))) {
    printf ("could not get shared mem for segment localPoint\n");
  }

  //  Get the shared memory to write the two points the user selects
  selectShm = new dtkSharedMem (4 * sizeof(float), "select");
  if (!selectShm || selectShm->isInvalid() || selectShm->getSize() < (4*sizeof(float))) {
    printf ("could not get shared mem for segment select\n");
  }


  initSceneGraph();

  return DTK_CONTINUE ; 
}

//  Initialize objects in the scene graph (mini map, etc)
void SelectPoints::initSceneGraph() {

  pfFilePath(".:./ftirwin_roads:"DPF_DATA_DIR);

  //  Load a marker to show our position on the full map
  pfNode* marker = pfdLoadFile("red_arrow.pfb");
  App->display()->addGhostNode(marker);
  markerDCS = new pfDCS();
  markerDCS->addChild(marker);
  markerDCS->setRot(0, 180, 0);
  markerDCS->setTrans(1000, 1000, 1500);
  markerDCS->setScale(400);
  App->display()->world()->addChild(markerDCS);

  //  Add the mini map, starts out in front
  addMiniMap();

  //  Plot the path on the large map and mini map
  makeGeode(mapGroup, miniVertices, miniGeode);
  makeGeode(App->display()->world(), largeVertices, largeGeode);

  checkModified();

  if (!radarLoaded) {
    //  Load a smaller marker ("You are here") to show the navigation position
    //  on the mini map
    pfNode* sphere = pfdLoadFile("fov.pfb");
    App->display()->addGhostNode(sphere);
    youAreHereDCS = new pfDCS();
    youAreHereDCS->setScale(0.003);
    youAreHereDCS->addChild(sphere);
    youAreHereDCS->setRot(-45, 90, 0);
    mapGroup->addChild(youAreHereDCS);
  }

  addSelectLine();

  //  Increase the navigation response to a good level
  App->display()->currentNav->response(512);

}

//  Create a line segment to show where the user is selecting
//  Create a small segment on the mini map and a large one on the full map
void SelectPoints::addSelectLine() {

  void* arena = pfGetSharedArena();

  //  2 vertices for the selected points
  selectPoints = (pfVec3 *) pfMalloc (2 * sizeof(pfVec3), arena);
  selectPointsLarge = (pfVec3 *) pfMalloc (2 * sizeof(pfVec3), arena);

  //  Initialize the vertices to be behind the map
  selectPoints[0].set(0.5, 0.5, -1000);
  selectPoints[1].set(0.5, 0.5, -1000);
  selectPointsLarge[0].set(10000, 10000, -1);
  selectPointsLarge[1].set(10000, 10000, -1);

  //  Length of the vertex list
  int* len = static_cast<int*> (pfMalloc(sizeof(int), arena));
  *len = 2;

  //  Create the geo set for the small select segment
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType(PFGS_LINESTRIPS);
  gSet->setNumPrims(1);  // one line segment
  gSet->setPrimLengths(len);
  gSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, selectPoints, NULL);

  //  Disable lighting
  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gSet->setGState(gState);

  //  Set the color to blue
  pfVec4* color = static_cast<pfVec4*> (pfMalloc(sizeof(pfVec4), arena));
  color->set(0.3, 0.3, 1, 0);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  pfGeode* geode = new pfGeode();
  geode->addGSet(gSet);

  mapGroup->addChild(geode);
  App->display()->addGhostNode(geode);

  //  Create the geoset and geode for the select segment on the large map
  gSet = new pfGeoSet();
  gSet->setPrimType(PFGS_LINESTRIPS);
  gSet->setNumPrims(1);  // one line segment
  len = static_cast<int*> (pfMalloc(sizeof(int), arena));
  *len = 2;
  gSet->setPrimLengths(len);
  gSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, selectPointsLarge, NULL);
  gSet->setGState(gState);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);
  
  selectLargeGeode = new pfGeode();
  selectLargeGeode->addGSet(gSet);
  App->display()->addGhostNode(selectLargeGeode);

  App->display()->world()->addChild(selectLargeGeode);
}

//  Add a mini-map to the scenegraph
void SelectPoints::addMiniMap() {

  //  Load the 2D map of Ft. Irwin in front of the user
  void* arena = pfGetSharedArena();

  //  Check to see if the radar is already loaded
  radar = (Radar*) App->check("radar");

  mapDCS = new pfDCS();

  if (radar == NULL) {
    printf ("radar returned null\n");
    radarLoaded = 0;
  
    //  Create a geoset with one polygon
    pfGeoSet* gSet = new pfGeoSet();
    gSet->setPrimType (PFGS_QUADS);
    gSet->setNumPrims(1);

    pfVec3* verts = (pfVec3 *) pfMalloc (4 * sizeof(pfVec3), arena);

    //  Create a polygon 1 unit wide and 1.25 units high
    float w = 1;
    float h = 1.25;
    verts[0].set(0, 0, 0);
    verts[1].set(w, 0, 0);
    verts[2].set(w, h, 0);
    verts[3].set(0, h, 0);
    gSet->setAttr (PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);

    pfVec4* color = (pfVec4*) pfMalloc( sizeof(pfVec4), arena);
    color->set(1, 1, 1, .8);
    gSet->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, color, NULL);

    //  Set the texture to cover the entire polygon
    pfVec2* textcoords = (pfVec2*) pfMalloc(4 * sizeof(pfVec2), arena);
    textcoords[0].set (0,0);
    textcoords[1].set (1, 0);
    textcoords[2].set (1, 1);
    textcoords[3].set (0, 1);

    gSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, textcoords, NULL);

    pfGeode* geode = new pfGeode();
    geode->addGSet(gSet);

    pfGeoState* gState = new pfGeoState();
    gState->setMode(PFSTATE_CULLFACE, PFCF_OFF);
    gState->setMode(PFSTATE_TRANSPARENCY, PFTR_ON);

    //  Load the map texture onto the polygon
    pfTexture* tex = new pfTexture();

    int ret = tex->loadFile(mapTexture);
    if (ret == 0) {
      printf ("could not load file\n");
      return;
    }

    gState->setMode(PFSTATE_ENTEXTURE, PF_ON);
    gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
    gState->setAttr(PFSTATE_TEXTURE, tex);
    
    gSet->setGState(gState);
    
    //  Set it to be in front of the user, in scene coordinates
    mapDCS->addChild(geode);

    mapGroup = new pfGroup();
    mapDCS->addChild(mapGroup);

    App->display()->scene()->addChild(mapDCS);
    setFront();
  }
  else {
    //  Get the map dcs that was loaded in the radar dso
    radarLoaded = 1;

    mapGroup = radar->getMap();
    mapDCS->addChild(mapGroup);
  }

  inFront = 1;

  //  Create markers for the selected points (blue dots)
  for (int c=0; c<2; c++) {
    selectDCS[c] = new pfDCS();
    selectLargeDCS[c] = new pfDCS();

    //  Create blue dots on the mini map
    pfNode* node = pfdLoadFile("blueSphere.pfb");
    App->display()->addGhostNode(node);
    if (node)
      selectDCS[c]->addChild(node);
    else
      break;

    selectDCS[c]->setScale(0.03);

    //  Put the markers behind the map so they are not seen
    selectDCS[c]->setTrans(0.5, -1000, -1);
    
    mapGroup->addChild(selectDCS[c]);

    //  Create blue arrows on the full map
    node = pfdLoadFile("blueArrow.pfb");
    App->display()->addGhostNode(node);
    if (node)
      selectLargeDCS[c]->addChild(node);
    else
      break;

    selectLargeDCS[c]->setScale(400);
    selectLargeDCS[c]->setRot(0, 180, 0);

    //  Put the markers behind the map so they are not seen
    selectLargeDCS[c]->setTrans(10000, 10000, -1000);
    
    App->display()->world()->addChild(selectLargeDCS[c]);
  }


  
  /*
  //  Create an arrow
  gSet = pfdNewArrow(200, arena);
  geode = new pfGeode();

  pfVec4* color = (pfVec4*) pfMalloc(sizeof(pfVec4), arena);
  color->set(0, 0, 1, 1);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  geode->addGSet(gSet);
  //gSet->setGState(gState);
  pfdStoreFile(geode, "blueArrow.pfb");
  */
}

//  Return true if the given point is on the mini-map
int SelectPoints::goodPoint(float* point) {

    //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
    //  Otherwise discard it
    return (fabs(point[2]) < 0.001 && point[0] >= 0 && point[0] <= 1.0 &&
        point[1] >= 0 && point[1] <= 1.25);
}

//  Check to see if the 'e' key was pressed.  If so, toggle the mini-map
//  between being in front of the user and to the right of the user
//  Also, check the 's' key to select points and check for button presses
//  on the wand.
void SelectPoints::checkKey() {
  
  //  Check shared memory to get the selection point
  //  (where the wand/mouse is pointed at on the mini map)
  float point[3];
  shm->read(point);

  //  Update the selected line with the current pointer position
  if (numSelected == 1) {

    if (goodPoint(point)) {
      selectPoints[1].set(point[0], point[1], 0.02);

      float z = terrainZ(point[0]*scale, point[1]*scale);
      selectPointsLarge[1].set(point[0]*scale, point[1]*scale, z+250);

      //  Update the bounding volume
      selectLargeGeode->getGSet(0)->setBound(NULL, PFBOUND_DYNAMIC);
      selectLargeGeode->setBound(NULL, PFBOUND_DYNAMIC);
    }
    else {
      selectPoints[1].set(0.5, 0.5, -1000);
      selectPointsLarge[1].set(10000, 10000, -1);
    }
  }

  //  Check the keyboard for toggle keys
  dtkRecord_event *event;
  KeyCode keyCode;

  //  Check for keyboard events
  while ((event = dq->getNextEvent(keyboard))) {
    int pressed = keyboard->read(&keyCode, event);

    //  Toggle the map position
    if (pressed && keyCode == toggle_key) {

      //  If radar dso is loaded, that will handle map toggling instead
      if (radarLoaded)
        continue;

      if (inFront) {
        inFront = 0;
        setRight();
      }
      else {
        inFront = 1;
        setFront();
      }

    }

    //  Select points
    if (pressed && keyCode == select_key) {
      if (!radarLoaded || radar->getMode() != circleRadar)
        doSelect(point);
    }
  }

  //  First frame flush any remaining button events
  if (firstFrame) {
    firstFrame = 0;

    dtkRecord_event *event;
    while((event = dqButton->getNextEvent(button))) {
    }
  }

  //  Check for button events
  while((event = dqButton->getNextEvent(button))) {
	  int whichButton;
	  int buttonState = button->read(&whichButton, event);

    if (buttonState && (whichButton == 3)) {
      doSelect(point);
    }
  }
  
}

//  Handle a button/keyboard press to select a point on the map
void SelectPoints::doSelect(float point[]) {
  //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
  //  Otherwise discard it
  if (goodPoint(point)) {
   
    //  If we have already selected two points, discard them
    if (numSelected == 2) {
      numSelected = 0;
      selectDCS[0]->setTrans(0.5, -1000, -1);
      selectDCS[1]->setTrans(0.5, -1000, -1);
      selectLargeDCS[0]->setTrans(10000, 10000, -1000);
      selectLargeDCS[1]->setTrans(10000, 10000, -1000);

      selectPoints[0].set(0.5, 0.5, -1000);
      selectPoints[1].set(0.5, 0.5, -1000);
      selectPointsLarge[0].set(10000, 10000, -1);
      selectPointsLarge[1].set(10000, 10000, -1);
    }

    selectDCS[numSelected]->setTrans(point[0], point[1], 0);
    selectPoints[numSelected].set(point[0], point[1], 0.02);

    float z = terrainZ(point[0]*scale, point[1]*scale);
    selectPointsLarge[numSelected].set(point[0]*scale, point[1]*scale, z+250);
    selectLargeDCS[numSelected]->setTrans(point[0]*scale, point[1]*scale, z+400);

    numSelected++;

    if (numSelected == 2) {
      //  After a valid selection, write them to shared memory
      float select[4];
      select[0] = selectPoints[0][0] * scale;
      select[1] = selectPoints[0][1] * scale;
      select[2] = selectPoints[1][0] * scale;
      select[3] = selectPoints[1][1] * scale;
      selectShm->write(select, 4*sizeof(float));
    }
  }
}

//  Set the map to be in front of the user
void SelectPoints::setFront() {
  //mapDCS->setRot(0, 90, 0);
  //mapDCS->setTrans(-1.2, 2.5, 0);
}

//  Set the map to be to the right of the user (right cave wall)
void SelectPoints::setRight() {
  //mapDCS->setRot(-90, 90, 0);
  //mapDCS->setTrans(1.5, 0, 0);
}

//  Create a geode for the path plot.  This creates a segment list with
//  zero vertices and adds it to group.
void SelectPoints::makeGeode (pfGroup* group, pfVec3*& vertices, pfGeode*& geode) {
  void* arena = pfGetSharedArena();

  //  Create the vertex list
  vertices = static_cast<pfVec3*> (pfMalloc(MAXPOINTS * sizeof(pfVec3), arena));

  //  Length of the vertex list
  if (numVertices == NULL)
    numVertices = static_cast<int*> (pfMalloc(sizeof(int), arena));
  *numVertices = 0;

  //  Create the geo set
  pfGeoSet* gSet = new pfGeoSet();
  gSet->setPrimType(PFGS_LINESTRIPS);
  gSet->setNumPrims(1);
  gSet->setPrimLengths(numVertices);
  gSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, vertices, NULL);

  pfGeoState* gState = new pfGeoState();
  gState->setMode(PFSTATE_ENLIGHTING, PF_OFF);
  gSet->setGState(gState);

  //  Set the color to green
  pfVec4* color = static_cast<pfVec4*> (pfMalloc(sizeof(pfVec4), arena));
  color->set(0.3, 1, 0.3, 0);
  gSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);

  geode = new pfGeode();
  geode->addGSet(gSet);
  App->display()->addGhostNode(geode);

  group->addChild(geode);
}

//  Check to see if the plot file has been modified.  If it has been, read
//  the new path from the file
void SelectPoints::checkModified() {
  if (file == NULL) {
    printf ("Environment variable DPF_PLAYBACK_FILE is not set\n");
    return;
  }

  stat(file, &fileStat);

  if (fileStat.st_mtime > lastModified) {
    lastModified = fileStat.st_mtime;

    readPath(miniVertices, 1/scale, 1);
    readPath(largeVertices, 1, 0);

    //  Update the bounding volumes of the geodes and geosets
    miniGeode->getGSet(0)->setBound(NULL, PFBOUND_DYNAMIC);
    miniGeode->setBound(NULL, PFBOUND_DYNAMIC);
    largeGeode->getGSet(0)->setBound(NULL, PFBOUND_DYNAMIC);
    largeGeode->setBound(NULL, PFBOUND_DYNAMIC);
  }

}

//  Get the z value of the terrain at the given x,y
//  Return -10000 if the segment does not hit the terrain
float SelectPoints::terrainZ(float x, float y) {

  intersect->setSegmentPos(0, x, y, 20000);
  intersect->doIntersection();

  if (intersect->segmentHit(0)) {
    pfVec3 point;
    intersect->getXFormPoint(0, &point);
    return point[2];
  }
  else {
    return -10000;
  }
}

//  Read a path from the playback file and populate vertices.  If constantZ is
//  non-zero, use a constant Z value instead of the Z values from the file
void SelectPoints::readPath (pfVec3* vertices, float scale, int constantZ) {

  //  Read in the map points
  int numPoints = 0;

  if (file == NULL) {
    printf ("Environment variable DPF_PLAYBACK_FILE is not set\n");
    return;
  }

  FILE* in = fopen(file, "r");
  if (in == NULL) {
    printf ("could not load file %s\n", file);
    return;
  }
  float t,x,y,z,h,p,r;
  while (fscanf (in, "%f %f %f %f %f %f %f", &t, &x, &y, &z, &h, &p, &r) >= 7) {
    vertices[numPoints][0] = x*scale;
    vertices[numPoints][1] = y*scale;

    if (constantZ)
      vertices[numPoints][2] = 0.02;

    else {
      float zHit = terrainZ(vertices[numPoints][0], vertices[numPoints][1]);
      if (zHit > -1000)
        vertices[numPoints][2] = zHit + 2.0;
      else {
        lastModified=0;
        vertices[numPoints][2] = z*scale;
      }

    }

    numPoints++;

    if (numPoints >= MAXPOINTS) {
      printf ("Exceeded %i points in the playback file, ignoring the rest\n", numPoints);
      break;
    }
  }

  *numVertices = numPoints;
}

// called every display loop, as a pfFrame callback, before pfDraw()
// this return value causes it to only be called once
int SelectPoints::preFrame(void) {
  //  Check shared memory to get the selection point
  float point[3];
  shm->read(point);

  //  If this point is on the mini-map, it will have z=0, 0<=x<=1 and 0<=y<=1.25
  //  Otherwise discard it
  if (goodPoint(point)) {
    float z = terrainZ(point[0]*scale, point[1]*scale);
    markerDCS->setTrans(point[0]*scale, point[1]*scale, z);
  }

  //  Plot the current navigation point on the mini-map
  
  //  Convert 0,0,0 from scene to world units
  pfCoord coord;
  coord.xyz.set(0,0,0);
  coord.hpr.set(0,0,0);
  App->display()->dpfToWorld(&coord);

  if (!radarLoaded) {
    youAreHereDCS->setRot(-45+coord.hpr[0], 90, 0);
    youAreHereDCS->setTrans(coord.xyz[0]/scale, coord.xyz[1]/scale, 0.02);
  }

  checkKey();

  checkModified();

  return DTK_CONTINUE ; 
}

// called every display loop, as a pfFrame callback, after pfDraw()
// this return value causes it to only be called once, and the object removed
int SelectPoints::postFrame(void) {
  return DTK_CONTINUE ; 
}


/*************** DTK C++ DSO loader/unloader functions ***********/
/*
 * All DTK DSO files are required to declare these two functions.
 * These function are called by the loading program to get your
 * C++ objects loaded.
 *
 *****************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager);
  return new SelectPoints(app);
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment;
  return DTKDSO_UNLOAD_CONTINUE;
}
